/* File appa.c, Program TESTVECTOR */

/*
                            Appendix A. 
                       Utility Procedures and 
           Types, Functions and Procedures from Chapter 3
                         Program TESTVECTOR



     Once, when I did a presentation on neural network technology on 
satellite TV, I commented that I would be wary of commercial network 
packages for three main reasons. First, neural network code is not 
difficult to write.  Second, the field has few algorithms that are 
well understood enough and stable enough to be "canned."  If you don't 
write the code yourself, and don't have access to the source code, 
then it is very difficult to make the modifications that will probably 
be necessary.  Third, a point that is made repeatedly throughout this 
book, commercial packages rarely help you with the most important part 
of the application: the data representation fitted to the problem.

     These comments caused a heated comment from one viewer on an
evaluation form about the arrogance of academics who don't appreciate 
the needs of practicing engineers.  However, at this time any neural 
net application is a research application.  Anyone who uses a new 
algorithm that they do not fully understand, whose workings are
intrinsically statistical, and which was written by someone else, is 
asking for trouble.  ("Sorry about the patient, judge.  Our neural 
network diagnostic program gets it wrong sometimes and we aren't quite
sure why.")

     These appendices provide code fragments for a number of neural
network primitive operations as well as several test programs.  The 
fragments were originally written in Pascal.  The specific version of 
Pascal used was VMS Pascal, running on Digital Equipment Corporation 
VAX computers, but porting these programs to other systems is possssible.  
Major problems using these programs on PC's will generally involve array 
size: the size of vectors and matrices may have to be drastically
cut to fit in many PC memories.  Computing speed is often less of
a problem nowadays than fitting large arrays in memory.  The larger
programs described in the last half of the book really require scientific
workstations with large, efficiently managed  memories and fast CPUs to
run properly.

     One comment is in order about the use of Pascal, and its use for
the original version of these programs.  Among serious computer people
Pascal is not very popular.  It is considered a "bondage and discipline"
language, according to the (highly recommended) "New Hacker's Dictionary"
(Cambridge, MA: MIT Press).  This means that Pascal does strict type
checking and does not allow the kind of freedom to bend rules of type
and syntax found in languages like C.  In this book the emphasis is on
scientific computation, specifically numerical simulations.  Scientific
programming usually involves straightforward numerical computation.
Clarity, readability, the ability to locate subtle errors, and the ability
to modify code long after the programs were written are paramount virtues.
Pascal by its unforgiving nature enforces these virtues more than
any other common computer language.  I allow the use of both Pascal
and C in the neural modelling classes that gave rise to this book.
My experience has been that students using C have far more trouble
with time consuming and hard-to-find bugs than students using Pascal.
This is especially noteworthy since the students using C tend to be the
more experienced programmers in the class, often computer science majors.
Programs written in Pascal either don't work at all or they work more or
less correctly.  Bugs are relatively easy to fix, once they are detected.
Even more, it is often possible to understand what a Pascal PROCEDURE
or FUNCTION is supposed to be doing years after it was written, if only
moderate care is taken in the use of names. Properly written C code can
also have these virtues, but for some reason often does not.

     Because visualization of complex systems is so important, we have 
used crude character based graphics in a number of places in our 
programs.  They are crude, but they get the job done.  Unfortunately, 
the greater complexity of modern computer interfaces makes programming 
anything to do with graphics now a job for professionals.  The 
professional graphics look nicer, of course, but it is often difficult 
to get these programs to operate in conjunction with simple test 
programs of the kind we use throughout this book.  In terms of basic 
visual understanding of the behavior of complex systems, the 
attractive interfaces that now come with most computers are regressive 
in that they make it much harder to do simple things than was the case 
ten years ago.  I suspect, however, that the next few years will mark 
the beginning of a golden age in the visualization of complex systems.  
Some remarkable programs are just starting to appear, specifically 
designed to use graphics to facilitate understanding of complexity.  
Until the day that it becomes trivial to make connections between 
graphics and the output of simulation programs there will still be 
a place for quick and dirty graphics, such as those we have used here.

     Earlier versions of Turbo Pascal and a number of other Pascal 
variants have a couple of useful PROCEDURES for screen control that we 
use often in our programs.  One is called GoToXY  and the other is 
ClrScr.  GoToXY directs the cursor to location X,Y on the screen.  
Unfortunately, the (X,Y) locations on the screen are not exactly what
you would think they would be, based on usual mathematical convention. 

     The screen control routines GoToXY() and ClrScr() is the present
series of programs will run on any standard ANSI terminal or terminal 
emulator, for example a VT100 or its successors.  Also, through the
use of conditional compilations, these routines will also work on a
PC when using a Microsoft compiler.
 
 
                               Columns = X

                   X=  1    2    3    4     ...           80
                       -------------------------------------|
                Y = 1  | (1,1)                              |
                    2  |                                    |
       Rows = Y     3  |                                    |
                    4  |                                    |

                       ...

                       |                                    |
                   24  |                                    |
                       |------------------------------------|


*/

/* Includes */
#if defined MSDOS
#include <graph.h>      /* _settextposition _clearscreen */
#endif
#include <math.h>       /* sqrt */
#include <stdio.h>      /* printf */

/* Prototypes for general purpose routines */
void ClrScr(void);
void GoToXY(int x, int y);
void Add_vectors(float Vector_1[], float Vector_2[], float Vector_sum[]);
void Subtract_vectors(float Vector_1[], float Vector_2[], 
    float Vector_difference[]);
float Vector_length(float Vector_1[]);
void Scalar_times_vector(float Scalar, float Vector_in[], float Vector_out[]);
float Inner_product(float Vector_1[], float Vector_2[]);
void Normalize(float Vector_in[], float Vector_out[]);
float Vector_cosine(float Vector_1[], float Vector_2[]);

/* Prototypes for routines specific to this program */
void Initialize(void);

/* Defines */
#define Dimensionality (int)(8)

/* Global arrays */
float a[Dimensionality];
float b[Dimensionality];
float c[Dimensionality];
float d[Dimensionality];
float w2[Dimensionality];
float w3[Dimensionality];
float c1[Dimensionality];
float z[Dimensionality];

/*--------------------------------------------------------------------------
Screen Control Procedure: Clear screen and move cursor to 1,1
--------------------------------------------------------------------------*/
void ClrScr(void) {
#if defined MSDOS
    _clearscreen(_GCLEARSCREEN);
#else
    printf("%c[2J",0x1b);
    printf("%c[%d;%dH",0x1b,1,1);
#endif
}   /* ClrScr */


/*--------------------------------------------------------------------------
Screen Control Procedure: Position cursor

Arguments
    x    column as integer 1..80
    y    row as integer 1..24
--------------------------------------------------------------------------*/
void GoToXY(int x, int y) {
    int x1,y1;          /* coordinates restricted to allowed range */

    x1=(x<1)? 1: (x>80)? 80: x;         /* 1<=x1<=80 */
    y1=(y<1)? 1: (y>24)? 24: y;         /* 1<=y1<=24 */
#if defined MSDOS
    _settextposition(y,x);
#else
    printf("%c[%d;%dH",0x1b,y1,x1);
#endif
}   /* GoToXY */



/*--------------------------------------------------------------------------
Adds two vectors.
Arguments
    float Vector_1[]            One vector, input
    float Vector_2[]            Other vector, input
    float Vector_sum[]          Vector_1+Vector_2, returned
--------------------------------------------------------------------------*/
void Add_vectors(float Vector_1[], float Vector_2[], float Vector_sum[]) {
    int i;      /* loop index */

    for(i=0; i<Dimensionality; i++)
	Vector_sum[i]=Vector_1[i] + Vector_2[i];
}   /* Add_vectors */


/*--------------------------------------------------------------------------
Subtracts two vectors.
Arguments
    float Vector_1[]            One vector, input
    float Vector_2[]            Other vector, input
    float Vector_sum[]          Vector_1-Vector_2, returned
--------------------------------------------------------------------------*/
void Subtract_vectors(float Vector_1[], float Vector_2[], 
    float Vector_difference[]) {
    int i;      /* loop index */

    for(i=0; i<Dimensionality; i++)
	Vector_difference[i]=Vector_1[i] - Vector_2[i];
}   /* Subtract_vectors */


/*--------------------------------------------------------------------------
Returns the length of a vector.
Argument
    float Vector_1[]            Vector, input
Returns
    float length                Magnitude of Vector_1
--------------------------------------------------------------------------*/
float Vector_length(float Vector_1[]) {
    float Sum_of_Squares;       /* Name says it all */
    int i;                      /* loop index */

    Sum_of_Squares=0;
    for(i=0; i<Dimensionality; i++)
	Sum_of_Squares += Vector_1[i]*Vector_1[i];
    return (float)sqrt(Sum_of_Squares);
}   /* Vector_length */



/*--------------------------------------------------------------------------
Scales a vector.
Arguments
    float Scalar                Scalar quantity
    float Vector_in[]           Vector, input
    float Vector_out[]          Scalar*Vector_in, output
--------------------------------------------------------------------------*/
void Scalar_times_vector(float Scalar, float Vector_in[], float Vector_out[]) {
    int i;      /* loop index */

    for(i=0; i<Dimensionality; i++)
	Vector_out[i] = Scalar*Vector_in[i];
}   /* Scalar_times_vector */

     
/*--------------------------------------------------------------------------
Forms the inner product of two vectors.
Arguments
    float Vector_1[]            One vector, input
    float Vector_2[]            Other vector, input
Returns
    float Inner_product         Vector_1 . Vector_2
--------------------------------------------------------------------------*/
float Inner_product(float Vector_1[], float Vector_2[]) {
    float Sum_of_products;      /* Name says it all */
    int i;                      /* loop index */

    Sum_of_products=0;
    for(i=0; i<Dimensionality; i++)
	Sum_of_products += Vector_1[i]*Vector_2[i];
    return Sum_of_products;
}   /* Inner_product */



/*--------------------------------------------------------------------------
Scales the input vector to unit length.  This is not possible if the input
vector has zero length, and in that case an error message is printed, and
the output vector is set to zero.

This routine may be called with both arguments pointing to the same array.

Arguments
    float Vector_in[]           Vector, input
    float Vector_out[]          Normalized vector, output
--------------------------------------------------------------------------*/
void Normalize(float Vector_in[], float Vector_out[]) {
    float Length_of_vector_in;  /* Name says it all */
    int i;                      /* loop index */

    Length_of_vector_in=Vector_length(Vector_in);
    if(Length_of_vector_in==0) {
	printf("Normalize: Input vector has zero length!\n");
	for(i=0; i<Dimensionality; i++)
	    Vector_out[i] = 0;
    } else {
	for(i=0; i<Dimensionality; i++)
	    Vector_out[i] = Vector_in[i]/Length_of_vector_in;
    }
}   /* Normalize */



/*--------------------------------------------------------------------------
Calculates the cosine of the 'angle' between two vectors.
Arguments
    float Vector_1[]            One vector, input
    float Vector_2[]            Other vector, input
Returns
    float Vector_cosine         (Vector_1 . Vector_2)/(|Vector_1| |Vector_2|)
--------------------------------------------------------------------------*/
float Vector_cosine(float Vector_1[], float Vector_2[]) {
    float Length_of_vector_1, Length_of_vector_2;     /* vector magnitudes */

    Length_of_vector_1=Vector_length(Vector_1);
    Length_of_vector_2=Vector_length(Vector_2);

    if(Length_of_vector_1==0 || Length_of_vector_2==0) {
	printf("Vector_cosine: Input vector has zero length!\n");
	return 0;
    } else {
	 return Inner_product(Vector_1,Vector_2)/
                Length_of_vector_1/Length_of_vector_2;
    }
}   /* Vector_cosine */



/*-----------------------------------------------------------------------
     Below is a very simple test program.  I have found it useful to see 
if students can get simple Pascal programs to compile and run with 
their own particular computer and Pascal implementation.  It also 
provides a little practice in use of the editor if the computer is 
unfamiliar.
-----------------------------------------------------------------------*/


/*--------------------------------------------------------------------------
This routine initializes six global vectors
--------------------------------------------------------------------------*/
void Initialize(void) {
    a[0]=-2;  b[0]=  4;  w2[0]= 1;  w3[0]= 1;  c1[0]=1; z[0]=0;
    a[1]= 5;  b[1]=  1;  w2[1]=-1;  w3[1]= 1;  c1[1]=0; z[1]=0;
    a[2]= 1;  b[2]= -3;  w2[2]=-1;  w3[2]=-1;  c1[2]=0; z[2]=0;
    a[3]= 2;  b[3]= -3;  w2[3]= 1;  w3[3]=-1;  c1[3]=0; z[3]=0;
    a[4]=-2;  b[4]=  0;  w2[4]= 1;  w3[4]=-1;  c1[4]=0; z[4]=0;
    a[5]=-1;  b[5]= -2;  w2[5]=-1;  w3[5]=-1;  c1[5]=0; z[5]=0;
    a[6]= 3;  b[6]= -1;  w2[6]=-1;  w3[6]= 1;  c1[6]=0; z[6]=0;
    a[7]= 4;  b[7]=  5;  w2[7]= 1;  w3[7]= 1;  c1[7]=0; z[7]=0;
}   /* Initialize */



/*--------------------------------------------------------------------------
Program TESTVECTOR
--------------------------------------------------------------------------*/
void main(void) {
    int i;	/* loop index */

    Initialize();    /* Initialize a,b,w2,w3,c1 and z */
    printf("Program to test vector routines.\n");
    printf("See p 82 of An Introduction to Neural Networks.\n");

    printf("\nTest vectors.");
    printf("\na :");
    for(i=0; i<Dimensionality; i++)
        printf("%9.3g",a[i]);
    printf("\nb :");
    for(i=0; i<Dimensionality; i++)
        printf("%9.3g",b[i]);
    printf("\nw2:");
    for(i=0; i<Dimensionality; i++)
        printf("%9.3g",w2[i]);
    printf("\nw3:");
    for(i=0; i<Dimensionality; i++)
        printf("%9.3g",w3[i]);
    printf("\nc1:");
    for(i=0; i<Dimensionality; i++)
        printf("%9.3g",c1[i]);
    printf("\nz :");
    for(i=0; i<Dimensionality; i++)
        printf("%9.3g",z[i]);

    printf("\n\nLengths of test vectors.\n");
    printf(" Length of a : %10.3g\n", Vector_length(a));
    printf(" Length of b : %10.3g\n", Vector_length(b));
    printf(" Length of w2: %10.3g\n", Vector_length(w2));
    printf(" Length of w3: %10.3g\n", Vector_length(w3));
    printf(" Length of c1: %10.3g\n", Vector_length(c1));
    printf(" Length of z : %10.3g\n", Vector_length(z));

    printf("\nCosines.\n");
    printf(" Cosine of angle between  a and  b: %10.3g\n",Vector_cosine(a,b));
    printf(" Cosine of angle between w2 and w3: %10.3g\n",Vector_cosine(w2,w3));
    printf(" Cosine of angle between  c1 and z: %10.3g\n",Vector_cosine(c1,z));

    /* Uncomment the following line if it all gets too much */
    /* printf("\n<Ret> to continue: "); getchar(); */

    printf("\nSums and differences.\n");
    Add_vectors (a,b,c);
    Subtract_vectors (a,b,d);
    printf(" Length of c=a+b : %10.3g\n", Vector_length(c));
    printf(" Length of d=a-b : %10.3g\n", Vector_length(d));
    printf(" Cosine (c,d)    : %10.3g\n", Vector_cosine(c,d));

    printf("\nNormalize all vectors: a,b,c,d,w2,w3,z.\n");
    Normalize(a,a);
    Normalize(b,b);
    Normalize(c,c);
    Normalize(d,d);
    Normalize(w2,w2);
    Normalize(w3,w3);
    Normalize(z,z);

    printf("\nSome lengths of normalized vectors.\n");
    printf(" New length of a : %10.3g\n", Vector_length(a));
    printf(" New length of c : %10.3g\n", Vector_length(c));
    printf(" New length of w2: %10.3g\n", Vector_length(w2));
    printf(" New length of z : %10.3g\n", Vector_length(z));

    printf("\nSome cosines of normalized vectors.\n");
    printf(" Cosine of angle between  a and  b: %10.3g\n",Vector_cosine(a,b));
    printf(" Cosine of angle between w2 and w3: %10.3g\n",Vector_cosine(w2,w3));
    printf(" Cosine of angle between  c and  d: %10.3g\n",Vector_cosine(c,d));
    printf(" Cosine of angle between c1 and  z: %10.3g\n",Vector_cosine(c1,z));

}   /* main */



/*--------------------------------------------------------------------------
                      Expected output

Program to test vector routines.
See p 82 of An Introduction to Neural Networks.

Test vectors.
a :       -2        5        1        2       -2       -1        3        4
b :        4        1       -3       -3        0       -2       -1        5
w2:        1       -1       -1        1        1       -1       -1        1
w3:        1        1       -1       -1       -1       -1        1        1
c1:        1        0        0        0        0        0        0        0
z :        0        0        0        0        0        0        0        0

Lengths of test vectors.
 Length of a :          8
 Length of b :       8.06
 Length of w2:       2.83
 Length of w3:       2.83
 Length of c1:          1
 Length of z :          0

Cosines.
 Cosine of angle between  a and  b:      0.109
 Cosine of angle between w2 and w3:          0
Vector_cosine: Input vector has zero length!
 Cosine of angle between  c1 and z:          0

Sums and differences.
 Length of c=a+b :         12
 Length of d=a-b :       10.7
 Cosine (c,d)    :    -0.0078

Normalize all vectors: a,b,c,d,w2,w3,z.
Normalize: Input vector has zero length!

Some lengths of normalized vectors.
 New length of a :          1
 New length of c :          1
 New length of w2:          1
 New length of z :          0

Some cosines of normalized vectors.
 Cosine of angle between  a and  b:      0.109
 Cosine of angle between w2 and w3:  -4.28e-09
 Cosine of angle between  c and  d:    -0.0078
Vector_cosine: Input vector has zero length!
 Cosine of angle between c1 and  z:          0

--------------------------------------------------------------------------*/

