/* File appd.c, Program Inner_Product_Demo */

/*
                            Appendix D
               Inner Product Demonstration (Chapter 7)


     This program demonstrates some properties of the inner product
between random vectors.  The inner product between a pair of random 
vectors is computed.  A histogram of inner product values is given 
as output, along with some summary statistics of the histogram.  

     Normalized vectors (length one) with uniformly distributed
random elements with mean zero are generated using the Unix math library
random number generator, rand().  If other computers are used, 
use their generator.  A word of warning: No computer random number 
generator is really trustworthy.  Use the tests in Chapter Three of 
Donald Knuth's "The Art of Computer Programming: Semi-Numerical
Algorithms, Volume 2," (Reading, MA: Addison-Wesley) if there is any 
doubt.  Because neural networks are so sensitive to correlations, 
they are exceptionally vulnerable to bad random number generators.
Personal computer random number generators are particularly likely to 
be poor.  

     When the program is run, the program will prompt for (1) the
seed for the random number generator, (2) the dimensionality of the 
vectors to be used (maximum 4,000), (3) the number of random vectors 
to be used to form the histogram, and (4) the number of bins between 
0 and 1.  For this last value, 10 bins would mean the bins were of 
size 0.1, 50 bins would be of size 0.02, and so on.  The maximum possible
inner product between two vectors of length one is, of course, one.  
High dimensionality simulations have histograms tightly clustered around 
zero.  Changing the bin size lets the histogram be expanded so its 
structure can be seen.  Some simple summary statistics are computed from
the simulation.
*/

#include <math.h>       /* sqrt */
#include <stdio.h>      /* printf */
#include <stdlib.h>     /* rand srand RAND_MAX */

#define Max_dimensionality (int)(4000)
#define Max_hist (int)(200)
typedef float Vector[Max_dimensionality];

/* Global variables set in Initialize() */
int Dimensionality;
int Nr_vectors;
int Bin_number;
int Histogram[2*Max_hist+1];

float Inner_product(float Vector_1[], float Vector_2[]);
float Vector_length(float Vector_1[]);
void Normalize(float Vector_in[], float Vector_out[]);
float Vector_cosine(float Vector_1[], float Vector_2[]);
void Make_random_vector(float Vect[]);
void Plot_histogram(void);
void Write_statistics(void);
void Initialize(void);
void main(void);

/*--------------------------------------------------------------------------
Forms the inner product of two vectors.
Arguments
    float Vector_1[]            One vector, input
    float Vector_2[]            Other vector, input
Returns
    float Inner_product         Vector_1 . Vector_2
--------------------------------------------------------------------------*/
float Inner_product(float Vector_1[], float Vector_2[]) {
    float Sum_of_products;      /* Name says it all */
    int i;                      /* loop index */

    Sum_of_products=0;
    for(i=0; i<Dimensionality; i++)
	Sum_of_products += Vector_1[i]*Vector_2[i];
    return Sum_of_products;
}   /* Inner_product */




/*--------------------------------------------------------------------------
Returns the length of a vector.
Argument
    float Vector_1[]            Vector, input
Returns
    float length                Magnitude of Vector_1
--------------------------------------------------------------------------*/
float Vector_length(float Vector_1[]) {
    float Sum_of_Squares;       /* Name says it all */
    int i;                      /* loop index */

    Sum_of_Squares=0;
    for(i=0; i<Dimensionality; i++)
	Sum_of_Squares += Vector_1[i]*Vector_1[i];
    return (float)sqrt(Sum_of_Squares);
}   /* Vector_length */



/*--------------------------------------------------------------------------
Scales the input vector to unit length.  This is not possible if the input
vector has zero length, and in that case an error message is printed, and
the output vector is set to zero.

This routine may be called with both arguments pointing to the same array.

Arguments
    float Vector_in[]           Vector, input
    float Vector_out[]          Normalized vector, output
--------------------------------------------------------------------------*/
void Normalize(float Vector_in[], float Vector_out[]) {
    float Length_of_vector_in;  /* Name says it all */
    int i;                      /* loop index */

    Length_of_vector_in=Vector_length(Vector_in);
    if(Length_of_vector_in==0) {
	printf("Normalize: Input vector has zero length!\n");
	for(i=0; i<Dimensionality; i++)
	    Vector_out[i] = 0;
    } else {
	for(i=0; i<Dimensionality; i++)
	    Vector_out[i] = Vector_in[i]/Length_of_vector_in;
    }
}   /* Normalize */



/*--------------------------------------------------------------------------
Calculates the cosine of the 'angle' between two vectors.
Arguments
    float Vector_1[]            One vector, input
    float Vector_2[]            Other vector, input
Returns
    float Vector_cosine         (Vector_1 . Vector_2)/(|Vector_1| |Vector_2|)
--------------------------------------------------------------------------*/
float Vector_cosine(float Vector_1[], float Vector_2[]) {
    float Length_of_vector_1, Length_of_vector_2;     /* vector magnitudes */

    Length_of_vector_1=Vector_length(Vector_1);
    Length_of_vector_2=Vector_length(Vector_2);

    if(Length_of_vector_1==0 || Length_of_vector_2==0) {
	printf("Vector_cosine: Input vector has zero length!\n");
	return 0;
    } else {
	 return Inner_product(Vector_1,Vector_2)/Length_of_vector_1/Length_of_vector_2;
    }
}   /* Vector_cosine */



/*-----------------------------------------------------------------------
This procedure generates a random vector with length one and with
elements of zero mean.  Unless srand() is called, the same sequence
random numbers will be called each time this program is run.
-----------------------------------------------------------------------*/
void Make_random_vector(float Vect[]) {
    int i;                      /* loop index */

    /* Pick a point uniformly distributed in a hypercube centred at origin */
    for(i=0; i<Dimensionality; i++)
	Vect[i]=2*((float)rand())/RAND_MAX - 1.0;

    /* Normalize it */
    Normalize(Vect,Vect);
}   /* Make_random_vector */


/*-----------------------------------------------------------------------
This routine plots the central few bins of the histogram.
-----------------------------------------------------------------------*/
void Plot_histogram(void) {
    int i;              /* bin number: -Max_hist..+Max_hist (or less) */
    int j;              /* count, used in plotting histogram */
    int limit;          /* range of bins to plot */
    int MaxDisp=67;     /* Maximum count to plot, ie 'page width' */
    float Y_axis;       /* inner product corresponding to given bin: -1..1 */

    printf("Inner products between %d pairs of random vectors.  ",Nr_vectors);
    printf("Dimensionality: %d\n\n",Dimensionality);

    limit=(Bin_number>=9)? 9: Bin_number;
    for(i=-limit; i<=limit; i++) {
	Y_axis=(float)i/(Bin_number+0.5f);
	printf("%6.3f%5d:",Y_axis,Histogram[Max_hist+i]);
	if(Histogram[Max_hist+i]>=MaxDisp) {
	    for(j=1; j<MaxDisp; j++) printf("*");
	    printf("+");
	} else {
	    for(j=1; j<=Histogram[Max_hist+i]; j++) printf("*");
	}
	printf("\n");
    }
}   /* Plot_histogram */


/*-----------------------------------------------------------------------
This prints summary statistics describing the distribution of inner
product values.  The statistics are based on the histogram rather than
the actual values.
-----------------------------------------------------------------------*/
void Write_statistics(void) {
    int i;              /* bin number: -Max_hist..+Max_hist (or less) */
    float Y_axis;       /* inner product corresponding to given bin: -1..1 */
    float Sum;          /* used in calculating mean and variance */
    float Mean;         /* average of all inner products */
    float Variance;     /* variance of all inner products */

    /* Calculate mean of distribution */
    for(Sum=0, i=-Bin_number; i<=Bin_number; i++) {
	Y_axis=(float)i/Bin_number;
	Sum += Histogram[Max_hist+i]*Y_axis;
    }
    Mean=Sum/Nr_vectors;
    printf("   Mean: %8.6f",Mean);

    /* Calculate variance of distribution */
    for(Sum=0, i=-Bin_number; i<=Bin_number; i++) {
	Y_axis=(float)i/Bin_number;
	Sum += Histogram[Max_hist+i]*(Y_axis-Mean)*(Y_axis-Mean);
    }
    Variance=Sum/Nr_vectors;
    printf("   Var : %10.8f",Variance);
    printf("   StdDev: %8.6f",sqrt(Variance));
    printf("   StdErr: %8.6f\n\n",sqrt(Variance/Nr_vectors));
}   /* Write_statistics */


/*-----------------------------------------------------------------------
Here the random number generator is seeded, and values assigned to
Dimensionality, Nr_vectors and Bin_number.
-----------------------------------------------------------------------*/
void Initialize(void) {
    int i;              /* bin number: -Max_hist..+Max_hist (or less) */
    unsigned int Seed;  /* seed for srand() */

    printf("This program generates a number of random vectors with      \n");
    printf("average element zero, and length one and computes and plots \n");
    printf("the inner product between them.  See pp 182-189.            \n");
    printf("Bins are centered on the number next to them.             \n\n");

    printf("Seed for random number generator (0..65535): ");
    scanf("%u",&Seed); srand(Seed);

    do {
	printf("Dimensionality (1..%4d)                   : ",
	    Max_dimensionality);
	scanf("%d",&Dimensionality);
    } while(Dimensionality<1 || Dimensionality>Max_dimensionality);

    do {
	printf("Number of random vectors to test (1..32767): ");
	scanf("%d",&Nr_vectors);
    } while(Nr_vectors<1);

    do {
	printf("Number of bins between 0 and 1 (1..%3d)    : ",Max_hist);
	scanf("%d",&Bin_number);
    } while(Bin_number<1 || Bin_number>Max_hist);
    printf("\n");

    for(i=-Max_hist; i<=Max_hist; i++) Histogram[Max_hist+i]=0;
}   /* Initialize */


/*-----------------------------------------------------------------------
Program Inner_Product_Demo
See p 182-189.
-----------------------------------------------------------------------*/
void main(void) {
    int i;              /* loop index */
    int Bin;            /* bin corresponding particular inner product */
    double ScaledCosine;/* float version of Bin */
    static Vector Vector_1;    /* one random vector of pair */
    static Vector Vector_2;    /* other random vector of pair */

    Initialize();
    for(i=0; i<Nr_vectors; i++) {
	Make_random_vector(Vector_1);
	Make_random_vector(Vector_2);
	/* Calculate the cosine, and scale it */
	ScaledCosine=Bin_number*Vector_cosine(Vector_1, Vector_2);
	/* Round ScaledCosine to the nearest integer */
	ScaledCosine += (ScaledCosine>0)? 0.5: -0.5;
	Bin=(int)ScaledCosine;
	if(Bin<-Bin_number || Bin>Bin_number) {
	    /* This should never happen, but lets check anyway */
	    printf("Calculated bin number=%d: out of range!\n",Bin);
	    break;
	}
	Histogram[Max_hist+Bin]++;
    }
    Plot_histogram();
    Write_statistics();
}   /* main */


/*-----------------------------------------------------------------------
               Example of output

This program generates a number of random vectors with      
average element zero, and length one and computes and plots 
the inner product between them.  See pp 182-189.            
Bins are centered on the number next to them.             

Seed for random number generator (0..65535): 123
Dimensionality (1..4000)                   : 1000
Number of random vectors to test (1..32767): 500
Number of bins between 0 and 1 (1..200)    : 50

Inner products between 500 pairs of random vectors.  Dimensionality: 1000

-0.178    0:
-0.158    0:
-0.139    0:
-0.119    0:
-0.099    0:
-0.079    6:******
-0.059   25:*************************
-0.040   55:*******************************************************
-0.020  105:******************************************************************+
 0.000  129:******************************************************************+
 0.020   96:******************************************************************+
 0.040   57:*********************************************************
 0.059   23:***********************
 0.079    3:***
 0.099    1:*
 0.119    0:
 0.139    0:
 0.158    0:
 0.178    0:
   Mean: -0.000720   Var : 0.00099948   StdDev: 0.031615   StdErr: 0.001414


-----------------------------------------------------------------------*/
