/* File appf.c, Programs AMDEMONOGR and AMDEMO */

/*
			       Appendix F
	    Simple Linear Associator Demonstration (Chapter 7)


     This program demonstrates some of the properties of the linear 
associator.  It associates pairs of random vectors with each other.  
When it learns a pair, it reports the cosine of the angle between the 
known correct output vector and what the network gives when the input 
member of the pair is used as input.  After the network has learned as 
many vector pairs as it was asked to, it then checks the performance 
of the entire set of vectors, rather than just the last pair learned.

     The output of the program simply gives some pertinent input and 
output parameters.  The maximum size of the network is limited to 250 
units, but this can be increased if there is sufficient memory 
available on the workstation.

     When it is run, the program prompts for a random number seed, 
the dimensionality of the system (maximum 250), and the number of 
vector pairs to learn (also maximum 250).  The random number generator 
(rand) is the library generator on a Unix.  It will have to 
be changed for other systems. 

     Two versions of this simple test program are presented.  The
first (AMDEMONOGR) uses text output, simply listing numerical
values for the cosine between the actual and the desired output.  The
output is in two groups.  First is the cosine between actual
and desired output during learning.  The second is the cosine between
actual and desired output for each member of the learned set after
learning is complete. 

      The second test program (AMDEMO) uses ASCII graphics to plot
the cosines after learning is complete.  The cosines for the learned 
vectors as well for as fifty random test vectors, for comparision,
are plotted.  

For description of AMDEMONOGR see page 195.  To compile:
        gcc -O -c appf.c
        gcc -o appf1 appf.o -lm
        Executable is named `appf1' instead of `AMDEMONOGR'.

For description of AMDEMO see page 195.  To compile, add the define -DGRAPHICS:
        gcc -O -DGRAPHICS -c appf.c
        gcc -o appf2 appf.o -lm
        Executable is named `appf2' instead of `AMDEMO'.

*/

#if !defined GRAPHICS
/*------------------------------------------------------------------
      This Program Generates ASCII Text as Output
------------------------------------------------------------------*/

#if defined MSDOS
#include <conio.h>      /* getch kbhit */
#include <graph.h>      /* _clearscreen */
#endif
#include <math.h>       /* sqrt */
#include <stdio.h>      /* getchar printf */
#include <stdlib.h>     /* rand srand RAND_MAX */

/* E.g. Max_dimensionality=250, Number_to_learn=250 */
#define Max_dimensionality  (int)(250)
#define Number_to_learn     (int)(250)
#define FALSE 0
#define TRUE  1
typedef float Vector[Max_dimensionality];
typedef float Matrix[Max_dimensionality][Max_dimensionality];

void ClrScr(void);
void Zero_matrix(void);
float Inner_product(float Vector_1[], float Vector_2[]);
float Vector_length(float Vector_1[]);
void Normalize(float Vector_in[], float Vector_out[]);
float Vector_cosine(float Vector_1[], float Vector_2[]);
void Make_normalized_random_vector(float Vect[]);
void Matrix_times_vector(Matrix A, Vector F, Vector G);
void Add_matrices(Matrix Mx_1, Matrix Mx_2, Matrix Matrix_sum);
void Outer_product(Vector G, Vector F, Matrix A);
void Make_learned_vectors(void);
void Initialize(void);
void main(void);

int Dimensionality;
int Number_of_pairs;
Matrix A;
Matrix Delta_A;
Vector G_out;
Vector F_set[Number_to_learn];
Vector G_set[Number_to_learn];


/*--------------------------------------------------------------------------
Screen Control Procedure: Clear screen and move cursor to 1,1
--------------------------------------------------------------------------*/
void ClrScr(void) {
#if defined MSDOS
    _clearscreen(_GCLEARSCREEN);
#else
    printf("%c[2J",0x1b);
    printf("%c[%d;%dH",0x1b,1,1);
#endif
}   /* ClrScr */


/*-------------------------------------------------------------------------
Sets the matrices A and Delta_A to zero.
-------------------------------------------------------------------------*/
void Zero_matrix(void) {
    int i,j;    /* indices */

    for(i=0; i<Dimensionality; i++)
	for(j=0; j<Dimensionality; j++) {
	    A[i][j]=0;
	    Delta_A[i][j]=0;
	}
}   /* Zero_matrix */


/*--------------------------------------------------------------------------
Forms the inner product of two vectors.
Arguments
    float Vector_1[]            One vector, input
    float Vector_2[]            Other vector, input
Returns
    float Inner_product         Vector_1 . Vector_2
--------------------------------------------------------------------------*/
float Inner_product(float Vector_1[], float Vector_2[]) {
    float Sum_of_products;      /* Name says it all */
    int i;                      /* loop index */

    Sum_of_products=0;
    for(i=0; i<Dimensionality; i++)
	Sum_of_products += Vector_1[i]*Vector_2[i];
    return Sum_of_products;
}   /* Inner_product */


/*--------------------------------------------------------------------------
Returns the length of a vector.
Argument
    float Vector_1[]            Vector, input
Returns
    float length                Magnitude of Vector_1
--------------------------------------------------------------------------*/
float Vector_length(float Vector_1[]) {
    float Sum_of_Squares;       /* Name says it all */
    int i;                      /* loop index */

    Sum_of_Squares=0;
    for(i=0; i<Dimensionality; i++)
	Sum_of_Squares += Vector_1[i]*Vector_1[i];
    return (float)sqrt(Sum_of_Squares);
}   /* Vector_length */


/*--------------------------------------------------------------------------
Scales the input vector to unit length.  This is not possible if the input
vector has zero length, and in that case an error message is printed, and
the output vector is set to zero.

This routine may be called with both arguments pointing to the same array.

Arguments
    float Vector_in[]           Vector, input
    float Vector_out[]          Normalized vector, output
--------------------------------------------------------------------------*/
void Normalize(float Vector_in[], float Vector_out[]) {
    float Length_of_vector_in;  /* Name says it all */
    int i;                      /* loop index */

    Length_of_vector_in=Vector_length(Vector_in);
    if(Length_of_vector_in==0) {
	printf("Normalize: Input vector has zero length!\n");
	for(i=0; i<Dimensionality; i++)
	    Vector_out[i] = 0;
    } else {
	for(i=0; i<Dimensionality; i++)
	    Vector_out[i] = Vector_in[i]/Length_of_vector_in;
    }
}   /* Normalize */



/*--------------------------------------------------------------------------
Calculates the cosine of the 'angle' between two vectors.
Arguments
    float Vector_1[]            One vector, input
    float Vector_2[]            Other vector, input
Returns
    float Vector_cosine         (Vector_1 . Vector_2)/(|Vector_1| |Vector_2|)
--------------------------------------------------------------------------*/
float Vector_cosine(float Vector_1[], float Vector_2[]) {
    float Length_of_vector_1, Length_of_vector_2;     /* vector magnitudes */

    Length_of_vector_1=Vector_length(Vector_1);
    Length_of_vector_2=Vector_length(Vector_2);

    if(Length_of_vector_1==0 || Length_of_vector_2==0) {
	printf("Vector_cosine: Input vector has zero length!\n");
	return 0;
    } else {
	 return Inner_product(Vector_1,Vector_2)/
                              Length_of_vector_1/Length_of_vector_2;
    }
}   /* Vector_cosine */



/*-----------------------------------------------------------------------
This procedure generates a random vector with length one and with
elements of zero mean.  Unless srand() is called, the same sequence
random numbers will be called each time this program is run.
-----------------------------------------------------------------------*/
void Make_normalized_random_vector(float Vect[]) {
    int i;                      /* loop index */

    /* Pick a point uniformly distributed in a hypercube centred at origin */
    for(i=0; i<Dimensionality; i++)
	Vect[i]=2*((float)rand())/RAND_MAX - 1.0;

    /* Normalize it */
    Normalize(Vect,Vect);
}   /* Make_normalized_random_vector */



/*-------------------------------------------------------------------------
This PROCEDURE forms the product Af = g where both f and g are column
vectors.
-------------------------------------------------------------------------*/
void Matrix_times_vector(Matrix A, Vector F, Vector G) {
    int i,j;    /* loop indices */

    /* Zero vector G */
    for(i=0; i<Dimensionality; i++)
	G[i]=0;
    for(i=0; i<Dimensionality; i++)
	for(j=0; j<Dimensionality; j++)
	    G[i] += A[i][j]*F[j];
}   /* Matrix_times_vector */


/*-------------------------------------------------------------------------
Calculates the sum of 2 matrices dimensioned [Dimensionality][Dimensionality]
Arguments
    Matrix Mx_1[]            One matrix, input
    Matrix Mx_2[]            Other matrix, input
Returns
    Matrix Matrix_sum        Mx_1 + Mx_2
-------------------------------------------------------------------------*/
void Add_matrices(Matrix Mx_1, Matrix Mx_2, Matrix Matrix_sum) {
    int i,j;    /* loop indices */

    for(i=0; i<Dimensionality; i++)
	for(j=0; j<Dimensionality; j++)
	    Matrix_sum[i][j]=Mx_1[i][j]+Mx_2[i][j];
}   /* Add_matrices */


/*-------------------------------------------------------------------------
We want to generate an outer product matrix A so that
	    T
     A = G F
-------------------------------------------------------------------------*/
void Outer_product(Vector G, Vector F, Matrix A) {
    int i,j;    /* loop indices */

    for(i=0; i<Dimensionality; i++)
	for(j=0; j<Dimensionality; j++)
	    A[i][j] = G[i]*F[j];
}   /* Outer_product */


/*-------------------------------------------------------------------------
Reports progress in building a learned set.
-------------------------------------------------------------------------*/
void Make_learned_vectors(void) {
    int i;      /* loop index */

    for(i=0; i<Number_to_learn; i++) {
	Make_normalized_random_vector(F_set[i]);
	Make_normalized_random_vector(G_set[i]);
    }
}   /* Make_learned_vectors */


/*-------------------------------------------------------------------------
-------------------------------------------------------------------------*/
void Initialize(void) {
    unsigned int Seed;  /* seed for srand() */

    ClrScr();
    printf("\n\nDemonstration of Simple Linear Associative Memory.\n");
    printf("System learns random vectors which are associated using\n");
    printf("an outer product matrix.   Resulting matrix is tested\n");
    printf("with the learned vectors.  See p 195.\n");
    printf("\nMaximum Dimensionality is: %d\n",Max_dimensionality);

    printf("\nSeed for random number generator (0..65535): ");
    scanf("%u",&Seed); srand(Seed);

    do {
	printf("Dimensionality (1..%4d): ",
	    Max_dimensionality);
	scanf("%d",&Dimensionality);
    } while(Dimensionality<1 || Dimensionality>Max_dimensionality);

    do {
	printf("Number of pairs to learn (1..%d): ",Number_to_learn);
	scanf("%d",&Number_of_pairs);
    } while(Number_of_pairs<1 && Number_of_pairs>Number_to_learn);

    /* Initialize vectors */
    Zero_matrix();
    printf("\n\n");
    Make_learned_vectors();
}   /* Initialize */


/*-------------------------------------------------------------------------
Program AMDEMONOGR
See p 195.
-------------------------------------------------------------------------*/
void main(void) {
    int i;              /* loop index */
    int Input_char;     /* keyboard input */

    do {
	Initialize();
	printf("Learning %d pairs of associations.\n",Number_of_pairs);
	for(i=0; i<Number_of_pairs; i++) {
	    Outer_product(G_set[i], F_set[i], Delta_A);
	    Add_matrices(A, Delta_A, A);
	    Matrix_times_vector(A, F_set [i], G_out);
	    printf("Pairs learned: %3d.  ",i);
	    printf("Cosine between G[%d] and actual output: %.5f\n",
		i,Vector_cosine(G_out, G_set[i]));
	}
	printf("\nLearning complete.\n");

	printf("The entire set of learned vectors will be tested "
	       "with the matrix.\n");
	for(i=0; i<Number_of_pairs; i++) {
	    Matrix_times_vector(A, F_set[i], G_out);
	    printf("Pair tested  : %3d.  ",i);
	    printf("Cosine between G[%d] and actual output: %.5f\n",
		i,Vector_cosine(G_out, G_set[i]));
	}

        /* Repeat? */
#if defined MSDOS
	printf("\n>> Do you want to repeat this demonstration? Y/N ");
	Input_char=getch(); while(kbhit()) getch();
#else
        getchar();
        do { 
	    printf("\n>> Do you want to repeat this demonstration? Y/N ");
            Input_char=getchar();
        } while(Input_char!='y' && Input_char!='Y' &&
                Input_char!='n' && Input_char!='N');
#endif
    } while(Input_char=='Y' || Input_char=='y');
}   /* main */


#else
/*------------------------------------------------------------------
      This Program Generates Simple ASCII Graphics as Output
------------------------------------------------------------------*/

#if defined MSDOS
#include <conio.h>      /* getch kbhit */
#include <graph.h>      /* _clearscreen */
#endif
#include <math.h>       /* sqrt */
#include <stdio.h>      /* getchar printf */
#include <stdlib.h>     /* rand srand RAND_MAX */

/* E.g. Max_dimensionality=250, Number_in_test_set=50, Number_to_learn=50 */
#define Max_dimensionality  (int)(250)
#define Number_in_test_set  (int)(50)
#define Number_to_learn     (int)(50)
#define Max_histogram       (int)(20)
#define FALSE 0
#define TRUE  1
typedef float Vector[Max_dimensionality];
typedef float Matrix[Max_dimensionality][Max_dimensionality];

void ClrScr(void);
void GoToXY(int x, int y);
float Inner_product(float Vector_1[], float Vector_2[]);
float Vector_length(float Vector_1[]);
void Normalize(float Vector_in[], float Vector_out[]);
float Vector_cosine(float Vector_1[], float Vector_2[]);
void Make_normalized_random_vector(float Vect[]);
void Matrix_times_vector(Matrix A, Vector F, Vector G);
void Add_matrices(Matrix Mx_1, Matrix Mx_2, Matrix Matrix_sum);
void Outer_product(Vector G, Vector F, Matrix A);
void Make_test_vectors(void);
void Make_learned_vectors(void);
void Display_outputs(void);
void Initialize(void);
void main(void);

int Dimensionality;
int Number_of_pairs;
int Index;
Matrix A;
Matrix Delta_A;
Vector First,Second;
Vector Test_vectors[Number_in_test_set];
Vector F_vectors[Number_to_learn];
Vector G_vectors[Number_to_learn];
int TS_histogram[2*Max_histogram+1];
int LS_histogram[2*Max_histogram+1];


/*--------------------------------------------------------------------------
Screen Control Procedure: Clear screen and move cursor to 1,1
--------------------------------------------------------------------------*/
void ClrScr(void) {
#if defined MSDOS
    _clearscreen(_GCLEARSCREEN);
#else
    printf("%c[2J",0x1b);
    printf("%c[%d;%dH",0x1b,1,1);
#endif
}   /* ClrScr */


/*--------------------------------------------------------------------------
Screen Control Procedure: Position cursor

Arguments
    x    column as integer 1..80
    y    row as integer 1..24
--------------------------------------------------------------------------*/
void GoToXY(int x, int y) {
    int x1,y1;          /* coordinates restricted to allowed range */

    x1=(x<1)? 1: (x>80)? 80: x;         /* 1<=x1<=80 */
    y1=(y<1)? 1: (y>24)? 24: y;         /* 1<=y1<=24 */
#if defined MSDOS
    _settextposition(y,x);
#else
    printf("%c[%d;%dH",0x1b,y1,x1);
#endif
}   /* GoToXY */



/*--------------------------------------------------------------------------
Forms the inner product of two vectors.
Arguments
    float Vector_1[]            One vector, input
    float Vector_2[]            Other vector, input
Returns
    float Inner_product         Vector_1 . Vector_2
--------------------------------------------------------------------------*/
float Inner_product(float Vector_1[], float Vector_2[]) {
    float Sum_of_products;      /* Name says it all */
    int i;                      /* loop index */

    Sum_of_products=0;
    for(i=0; i<Dimensionality; i++)
	Sum_of_products += Vector_1[i]*Vector_2[i];
    return Sum_of_products;
}   /* Inner_product */


/*--------------------------------------------------------------------------
Returns the length of a vector.
Argument
    float Vector_1[]            Vector, input
Returns
    float length                Magnitude of Vector_1
--------------------------------------------------------------------------*/
float Vector_length(float Vector_1[]) {
    float Sum_of_Squares;       /* Name says it all */
    int i;                      /* loop index */

    Sum_of_Squares=0;
    for(i=0; i<Dimensionality; i++)
	Sum_of_Squares += Vector_1[i]*Vector_1[i];
    return (float)sqrt(Sum_of_Squares);
}   /* Vector_length */


/*--------------------------------------------------------------------------
Scales the input vector to unit length.  This is not possible if the input
vector has zero length, and in that case an error message is printed, and
the output vector is set to zero.

This routine may be called with both arguments pointing to the same array.

Arguments
    float Vector_in[]           Vector, input
    float Vector_out[]          Normalized vector, output
--------------------------------------------------------------------------*/
void Normalize(float Vector_in[], float Vector_out[]) {
    float Length_of_vector_in;  /* Name says it all */
    int i;                      /* loop index */

    Length_of_vector_in=Vector_length(Vector_in);
    if(Length_of_vector_in==0) {
	printf("Normalize: Input vector has zero length!\n");
	for(i=0; i<Dimensionality; i++)
	    Vector_out[i] = 0;
    } else {
	for(i=0; i<Dimensionality; i++)
	    Vector_out[i] = Vector_in[i]/Length_of_vector_in;
    }
}   /* Normalize */



/*--------------------------------------------------------------------------
Calculates the cosine of the 'angle' between two vectors.
Arguments
    float Vector_1[]            One vector, input
    float Vector_2[]            Other vector, input
Returns
    float Vector_cosine         (Vector_1 . Vector_2)/(|Vector_1| |Vector_2|)
--------------------------------------------------------------------------*/
float Vector_cosine(float Vector_1[], float Vector_2[]) {
    float Length_of_vector_1, Length_of_vector_2;     /* vector magnitudes */

    Length_of_vector_1=Vector_length(Vector_1);
    Length_of_vector_2=Vector_length(Vector_2);

    if(Length_of_vector_1==0 || Length_of_vector_2==0) {
	printf("Vector_cosine: Input vector has zero length!\n");
	return 0;
    } else {
	 return Inner_product(Vector_1,Vector_2)/
                                Length_of_vector_1/Length_of_vector_2;
    }
}   /* Vector_cosine */



/*-----------------------------------------------------------------------
This procedure generates a random vector with length one and with
elements of zero mean.  Unless srand() is called, the same sequence
random numbers will be called each time this program is run.
-----------------------------------------------------------------------*/
void Make_normalized_random_vector(float Vect[]) {
    int i;                      /* loop index */

    /* Pick a point uniformly distributed in a hypercube centred at origin */
    for(i=0; i<Dimensionality; i++)
	Vect[i]=2*((float)rand())/RAND_MAX - 1.0;

    /* Normalize it */
    Normalize(Vect,Vect);
}   /* Make_normalized_random_vector */



/*-------------------------------------------------------------------------
This PROCEDURE forms the product Af = g where both f and g are column
vectors.
-------------------------------------------------------------------------*/
void Matrix_times_vector(Matrix A, Vector F, Vector G) {
    int i,j;    /* loop indices */

    /* Zero vector G */
    for(i=0; i<Dimensionality; i++)
	G[i]=0;
    for(i=0; i<Dimensionality; i++)
	for(j=0; j<Dimensionality; j++)
	    G[i] += A[i][j]*F[j];
}   /* Matrix_times_vector */


/*-------------------------------------------------------------------------
Adds two square matrices of floats.
-------------------------------------------------------------------------*/
void Add_matrices(Matrix Mx_1, Matrix Mx_2, Matrix Matrix_sum) {
    int i,j;    /* loop indices */

    for(i=0; i<Dimensionality; i++)
	for(j=0; j<Dimensionality; j++)
	    Matrix_sum[i][j]=Mx_1[i][j]+Mx_2[i][j];
}   /* Add_matrices */


/*-------------------------------------------------------------------------
We want to generate an outer product matrix A so that
	    T
     A = G F
-------------------------------------------------------------------------*/
void Outer_product(Vector G, Vector F, Matrix A) {
    int i,j;    /* loop indices */

    for(i=0; i<Dimensionality; i++)
	for(j=0; j<Dimensionality; j++)
	    A[i][j] = G[i]*F[j];
}   /* Outer_product */


/*-------------------------------------------------------------------------
Reports progress in building a test set.
-------------------------------------------------------------------------*/
void Make_test_vectors(void) {
    int i;      /* loop index */

    GoToXY(1,22);
    printf("Making test vectors   :         ");
    for(i=0; i<Number_in_test_set; i++) {
	Make_normalized_random_vector(Test_vectors[i]);
	GoToXY(24,22);
	printf("%d",i);
    }
}   /* Make_test_vectors */


/*-------------------------------------------------------------------------
Reports progress in building a learned set.
-------------------------------------------------------------------------*/
void Make_learned_vectors(void) {
    int i;      /* loop index */

    GoToXY(1,23);
    printf("Making learned vectors:         ");
    for(i=0; i<Number_to_learn; i++) {
	Make_normalized_random_vector(F_vectors[i]);
	Make_normalized_random_vector(G_vectors[i]);
	GoToXY(24,23);
	printf("%d",i);
    }
}   /* Make_learned_vectors */


/*-------------------------------------------------------------------------
A histogram of the lengths of the of the test and learned vectors.
The histogram is binned in units of 0.1, but is labelled with integers.

The distribution of learned vectors is plotted first, and should be 
centred on 1.0, or the bin labelled `10'.

The distribution of test vectors is superimposed, and should be 
centred on 0.
-------------------------------------------------------------------------*/
void Display_outputs(void) {
    int i;      /* row of screen */
    int j;      /* column of screen */
    int i0;     /* mid-point of histograms */

    i0=Max_histogram;   /* offset to mid_point of histograms */
    ClrScr();
    GoToXY(1,1);
    printf("Learned Vectors. (*), Test Vectors (+).\n");

    /* Histogram corresponding to learned vectors */
    GoToXY(1,2);
    for(i=-7; i<=11; i++) {
	if(i<-Max_histogram || i>Max_histogram) continue;
	printf("%3d ",i);
	for(j=1; j<=LS_histogram[i0+i]; j++) printf("*");
	printf("\n");
    }

    /* Histogram corresponding to test vectors */
    GoToXY(1,2);
    for(i=-7; i<=11; i++) {
	if(i<-Max_histogram || i>Max_histogram) continue;
	printf("%3d ",i);
	for(j=1; j<=TS_histogram[i0+i]; j++) printf("+");
	printf("\n");
    }
}   /* Display_outputs */


/*-------------------------------------------------------------------------
Here the random number generator is seeded, and values assigned to
Dimensionality and Number_of_vectors.
-------------------------------------------------------------------------*/
void Initialize(void) {
    int i;              /* loop index */
    unsigned int Seed;  /* seed for srand() */

    ClrScr();
    printf("\n\nDemonstration of Simple Linear Associative Memory.\n");
    printf("System learns random vectors which are associated using\n");
    printf("an outer product matrix.   Resulting matrix is tested\n");
    printf("with learned and unlearned vectors.  See p 195.\n");
    printf("\nMaximum Dimensionality is: %d\n",Max_dimensionality);

    printf("\nSeed for random number generator (0..65535): ");
    scanf("%u",&Seed); srand(Seed);

    do {
	printf("Dimensionality (1..%4d): ",
	    Max_dimensionality);
	scanf("%d",&Dimensionality);
    } while(Dimensionality<1 || Dimensionality>Max_dimensionality);

    do {
	printf("Number of pairs to learn (1..%d): ",Number_to_learn);
	scanf("%d",&Number_of_pairs);
    } while(Number_of_pairs<1 && Number_of_pairs>Number_to_learn);

    /* Initialize histograms */
    for(i=-Max_histogram; i<=Max_histogram; i++) {
	LS_histogram[Max_histogram+i]=0;
	TS_histogram[Max_histogram+i]=0;
    }

    /* Initialize vectors */
    Make_test_vectors();
    Make_learned_vectors();
}   /* Initialize */


/*-------------------------------------------------------------------------
Program AMDEMO
See p 195.
-------------------------------------------------------------------------*/
void main(void) {
    int i;              /* loop index */
    int Input_char;     /* keyboard input */
    float Goodness_of_recall;
    Vector G_out;

    do {
	Initialize();

	GoToXY(1,1);
	printf(">> Learning:               ");
	GoToXY(1,13); for(i=0; i<80; i++) printf(" ");      /* clear line */
	for(i=0; i<Number_of_pairs; i++) {
	    GoToXY(1,13);
	    Outer_product(G_vectors[i], F_vectors[i], Delta_A);
	    Add_matrices(A, Delta_A, A);
	    Matrix_times_vector(A, F_vectors[i], G_out);
	    printf("Pairs learned: %4d  ",i);
	    printf("Cosine of last pair: %8.4f\n",
		    Vector_cosine(G_out, G_vectors[i]));
	}

	GoToXY(1,1);
	printf(">> Testing after learning.");
	GoToXY(1,13); for(i=0; i<80; i++) printf(" ");      /* clear line */
	for(i=0; i<Number_of_pairs; i++) {
	    GoToXY(1,13);
	    Matrix_times_vector(A, F_vectors[i], G_out);
	    Goodness_of_recall=Vector_cosine(G_out, G_vectors[i]);
	    printf("Pairs tested: %4d  ",i);
	    printf("Cosine of output    : %8.4f\n",Goodness_of_recall);
	    Goodness_of_recall*=10;
	    /* Round Goodness_of_recall to the nearest integer */
	    Goodness_of_recall += (Goodness_of_recall>0)? 0.5: -0.5;
	    Index=(int)Goodness_of_recall;
	    if(Index>=-Max_histogram && Index<=Max_histogram)
		LS_histogram[Max_histogram+Index]++;
	}

	GoToXY(1,1);
	printf(">> Using test vectors.     ");
	GoToXY(1,13); for(i=0; i<80; i++) printf(" ");      /* clear line */
	for(i=0; i<Number_in_test_set; i++) {
	    GoToXY(1,13);
	    Matrix_times_vector(A, Test_vectors[i], G_out);
	    Goodness_of_recall=Vector_cosine(G_out, G_vectors[i]);
	    printf("Pairs tested: %4d  ",i);
	    printf("Cosine of test      : %8.4f\n",Goodness_of_recall);
	    Goodness_of_recall*=10;
	    /* Round Goodness_of_recall to the nearest integer */
	    Goodness_of_recall += (Goodness_of_recall>0)? 0.5: -0.5;
	    Index=(int)Goodness_of_recall;
	    if(Index>=-Max_histogram && Index<=Max_histogram)
		TS_histogram[Max_histogram+Index]++;
	}
	Display_outputs();

        /* Repeat? */
	GoToXY(1,21);
#if defined MSDOS
	printf("\n>> Do you want to repeat this demonstration? Y/N ");
	Input_char=getch(); while(kbhit()) getch();
#else
        getchar();
        do { 
	    printf("\n>> Do you want to repeat this demonstration? Y/N ");
            Input_char=getchar();
        } while(Input_char!='y' && Input_char!='Y' &&
                Input_char!='n' && Input_char!='N');
#endif
    } while(Input_char=='Y' || Input_char=='y');
    GoToXY(1,23);
}   /* main */

#endif
