/* File appg.c, Program AMDEMOL */

/*
                             Appendix G
                   Associative Memory Demonstration
                 Discrimination by Length (Chapter 7)

     This program shows that the simple linear associator can
discriminate new from old patterns, based on their length.  It is 
essentially the same as the test program presented in Appendix F.

     The program associates pairs of normalized zero mean random 
vectors.  The program prompts for the seed of the random number generator,
the dimensionality of the system (maximum 250) and the number of pairs
to learn (maximum 50).  These numbers can be easily changed.

     When the pairs have been learned, the program computes the length 
of the output vector for both the test and the learned vectors and 
plots the histogram of values.  If the network performed perfectly, the 
length of the learned output vector would be about 1 and the length of 
the output from a test vector would be around zero.

     The histogram is displayed using crude character based graphics.  
The vertical axis gives 10 times the value of the inner product and the 
horizontal the number of output lengths falling in that bin.  
*/

#if defined MSDOS
#include <conio.h>      /* getch kbhit */
#include <graph.h>      /* _clearscreen */
#endif
#include <math.h>       /* sqrt */
#include <stdio.h>      /* printf */
#include <stdlib.h>     /* rand srand RAND_MAX */

/* E.g. Max_dimensionality=250 */
#define Max_dimensionality  (int)(250)
#define Number_in_test_set  (int)(50)
#define Number_to_learn     (int)(50)
#define Max_histogram       (int)(50)
#define FALSE 0
#define TRUE  1
typedef float Vector[Max_dimensionality];
typedef float Matrix[Max_dimensionality][Max_dimensionality];

void ClrScr(void);
void GoToXY(int x, int y);
void Zero_matrix(void);
float Inner_product(float Vector_1[], float Vector_2[]);
float Vector_length(float Vector_1[]);
void Normalize(float Vector_in[], float Vector_out[]);
void Make_normalized_random_vector(float Vect[]);
void Matrix_times_vector(Matrix A, Vector F, Vector G);
void Add_matrices(Matrix Mx_1, Matrix Mx_2, Matrix Matrix_sum);
void Outer_product(Vector G, Vector F, Matrix A);
void Make_learned_vectors(void);
void Initialize(void);
void main(void);

int Dimensionality;
int Number_of_pairs;
int Index;
Matrix A;
Matrix Delta_A;
Vector First,Second;
Vector Test_vectors[Number_in_test_set];
Vector F_vectors[Number_to_learn];
Vector G_vectors[Number_to_learn];
int TS_histogram[2*Max_histogram+1];
int LS_histogram[2*Max_histogram+1];


/*--------------------------------------------------------------------------
Screen Control Procedure: Clear screen and move cursor to 1,1
--------------------------------------------------------------------------*/
void ClrScr(void) {
#if defined MSDOS
    _clearscreen(_GCLEARSCREEN);
#else
    printf("%c[2J",0x1b);
    printf("%c[%d;%dH",0x1b,1,1);
#endif
}   /* ClrScr */


/*--------------------------------------------------------------------------
Screen Control Procedure: Position cursor

Arguments
    x    column as integer 1..80
    y    row as integer 1..24
--------------------------------------------------------------------------*/
void GoToXY(int x, int y) {
    int x1,y1;          /* coordinates restricted to allowed range */

    x1=(x<1)? 1: (x>80)? 80: x;         /* 1<=x1<=80 */
    y1=(y<1)? 1: (y>24)? 24: y;         /* 1<=y1<=24 */
#if defined MSDOS
    _settextposition(y,x);
#else
    printf("%c[%d;%dH",0x1b,y1,x1);
#endif
}   /* GoToXY */



/*--------------------------------------------------------------------------
Forms the inner product of two vectors.
Arguments
    float Vector_1[]            One vector, input
    float Vector_2[]            Other vector, input
Returns
    float Inner_product         Vector_1 . Vector_2
--------------------------------------------------------------------------*/
float Inner_product(float Vector_1[], float Vector_2[]) {
    float Sum_of_products;      /* Name says it all */
    int i;                      /* loop index */

    Sum_of_products=0;
    for(i=0; i<Dimensionality; i++)
	Sum_of_products += Vector_1[i]*Vector_2[i];
    return Sum_of_products;
}   /* Inner_product */


/*--------------------------------------------------------------------------
Returns the length of a vector.
Argument
    float Vector_1[]            Vector, input
Returns
    float length                Magnitude of Vector_1
--------------------------------------------------------------------------*/
float Vector_length(float Vector_1[]) {
    float Sum_of_Squares;       /* Name says it all */
    int i;                      /* loop index */

    Sum_of_Squares=0;
    for(i=0; i<Dimensionality; i++)
	Sum_of_Squares += Vector_1[i]*Vector_1[i];
    return (float)sqrt(Sum_of_Squares);
}   /* Vector_length */


/*--------------------------------------------------------------------------
Scales the input vector to unit length.  This is not possible if the input
vector has zero length, and in that case an error message is printed, and
the output vector is set to zero.

This routine may be called with both arguments pointing to the same array.

Arguments
    float Vector_in[]           Vector, input
    float Vector_out[]          Normalized vector, output
--------------------------------------------------------------------------*/
void Normalize(float Vector_in[], float Vector_out[]) {
    float Length_of_vector_in;  /* Name says it all */
    int i;                      /* loop index */

    Length_of_vector_in=Vector_length(Vector_in);
    if(Length_of_vector_in==0) {
	printf("Normalize: Input vector has zero length!\n");
	for(i=0; i<Dimensionality; i++)
	    Vector_out[i] = 0;
    } else {
	for(i=0; i<Dimensionality; i++)
	    Vector_out[i] = Vector_in[i]/Length_of_vector_in;
    }
}   /* Normalize */


/*-----------------------------------------------------------------------
This procedure generates a random vector with length one and with
elements of zero mean.  Unless srand() is called, the same sequence
random numbers will be called each time this program is run.
-----------------------------------------------------------------------*/
void Make_normalized_random_vector(float Vect[]) {
    int i;                      /* loop index */

    /* Pick a point uniformly distributed in a hypercube centred at origin */
    for(i=0; i<Dimensionality; i++)
	Vect[i]=2*((float)rand())/RAND_MAX - 1.0;

    /* Normalize it */
    Normalize(Vect,Vect);
}   /* Make_normalized_random_vector */



/*-------------------------------------------------------------------------
This PROCEDURE forms the product Af = g where both f and g are column
vectors.
-------------------------------------------------------------------------*/
void Matrix_times_vector(Matrix A, Vector F, Vector G) {
    int i,j;    /* loop indices */

    /* Zero vector G */
    for(i=0; i<Dimensionality; i++)
	G[i]=0;
    for(i=0; i<Dimensionality; i++)
	for(j=0; j<Dimensionality; j++)
	    G[i] += A[i][j]*F[j];
}   /* Matrix_times_vector */


/*-------------------------------------------------------------------------
Calculates the sum of 2 matrices dimensioned [Dimensionality][Dimensionality]
Arguments
    Matrix Mx_1[]            One matrix, input
    Matrix Mx_2[]            Other matrix, input
Returns
    Matrix Matrix_sum        Mx_1 + Mx_2
-------------------------------------------------------------------------*/
void Add_matrices(Matrix Mx_1, Matrix Mx_2, Matrix Matrix_sum) {
    int i,j;    /* loop indices */

    for(i=0; i<Dimensionality; i++)
	for(j=0; j<Dimensionality; j++)
	    Matrix_sum[i][j]=Mx_1[i][j]+Mx_2[i][j];
}   /* Add_matrices */


/*-------------------------------------------------------------------------
We want to generate an outer product matrix A so that
	    T
     A = G F
-------------------------------------------------------------------------*/
void Outer_product(Vector G, Vector F, Matrix A) {
    int i,j;    /* loop indices */

    for(i=0; i<Dimensionality; i++)
	for(j=0; j<Dimensionality; j++)
	    A[i][j] = G[i]*F[j];
}   /* Outer_product */


/*-------------------------------------------------------------------------
-------------------------------------------------------------------------*/
void Make_test_vectors(void) {
    int i;      /* loop index */

    GoToXY(1,22);
    printf("Making test vectors   :         ");
    for(i=0; i<Number_in_test_set; i++) {
	Make_normalized_random_vector(Test_vectors[i]);
	GoToXY(24,22);
	printf("%d",i);
    }
}   /* Make_test_vectors */


/*-------------------------------------------------------------------------
-------------------------------------------------------------------------*/
void Make_learned_vectors(void) {
    int i;      /* loop index */

    GoToXY(1,23);
    printf("Making learned vectors:         ");
    for(i=0; i<Number_to_learn; i++) {
	Make_normalized_random_vector(F_vectors[i]);
	Make_normalized_random_vector(G_vectors[i]);
	GoToXY(24,23);
	printf("%d",i);
    }
}   /* Make_learned_vectors */


/*-------------------------------------------------------------------------
A histogram of the lengths of the of the test and learned vectors.
The histogram is binned in units of 0.1, but is labelled with integers.

The distribution of learned vectors is plotted first, and should be 
centred on 1.0, or the bin labelled `10'.

The distribution of test vectors is superimposed, and should be 
centred on 0.
-------------------------------------------------------------------------*/
void Display_outputs(void) {
    int i;      /* row of screen */
    int j;      /* column of screen */
    int i0;     /* mid-point of histograms */

    i0=Max_histogram;   /* offset to mid_point of histograms */
    ClrScr();
    GoToXY(1,2);
    printf("Learned Vectors. (*), Test Vectors (+).\n");

    /* Histogram corresponding to learned vectors */
    GoToXY(1,3);
    for(i=0; i<=14; i++) {
	if(i<-Max_histogram || i>Max_histogram) continue;
	printf("%3.1f ",(float)i/10);
	for(j=1; j<=LS_histogram[i0+i]; j++) printf("*");
	printf("\n");
    }

    /* Histogram corresponding to test vectors */
    GoToXY(1,3);
    for(i=0; i<=14; i++) {
	if(i<-Max_histogram || i>Max_histogram) continue;
	printf("%3.1f ",(float)i/10);
	for(j=1; j<=TS_histogram[i0+i]; j++) printf("+");
	printf("\n");
    }
}   /* Display_outputs */


/*-------------------------------------------------------------------------
-------------------------------------------------------------------------*/
void Initialize(void) {
    int i,j;            /* loop indices */
    unsigned int Seed;  /* seed for srand() */

    ClrScr();
    printf("\n\nDemonstration of Associative Memory.\n");
    printf("Learns random vectors which are associated using\n");
    printf("an outer product matrix.   Tested with learned and\n");
    printf("unlearned vectors.  Lengths of output vectors plotted.\n");
    printf("See p 197.\n");

    printf("\nSeed for random number generator (0..65535): ");
    scanf("%u",&Seed); srand(Seed);

    do {
	printf("Dimensionality (1..%4d): ",
	    Max_dimensionality);
	scanf("%d",&Dimensionality);
    } while(Dimensionality<1 || Dimensionality>Max_dimensionality);

    do {
	printf("Number of pairs to learn (1..%d): ",Number_to_learn);
	scanf("%d",&Number_of_pairs);
    } while(Number_of_pairs<1 && Number_of_pairs>Number_to_learn);

    /* Initialize histograms */
    for(i=-Max_histogram; i<=Max_histogram; i++) {
	LS_histogram[Max_histogram+i]=0;
	TS_histogram[Max_histogram+i]=0;
    }

    /* Initialize vectors */
    Make_test_vectors();
    Make_learned_vectors();

    /* Initialize matrices */
    for(j=0; j<Dimensionality; j++)
    for(i=0; i<Dimensionality; i++) {
	A[j][i]=0;
	Delta_A[j][i]=0;
    }
}   /* Initialize */



/*-------------------------------------------------------------------------
Program AMDEMOL
See p 197.
-------------------------------------------------------------------------*/
void main(void) {
    int i;              /* loop index */
    int Input_char;     /* keyboard input */
    float Goodness_of_recall;
    Vector G_out;

    do {
	Initialize();

	GoToXY(1,1);
	printf(">> Learning:               ");
	GoToXY(1,12); for(i=0; i<80; i++) printf(" ");      /* clear line */
	for(i=0; i<Number_of_pairs; i++) {
	    GoToXY(1,12);
	    Outer_product(G_vectors[i], F_vectors[i], Delta_A);
	    Add_matrices(A, Delta_A, A);
	    Matrix_times_vector(A, F_vectors[i], G_out);
	    printf("Pairs learned: %4d  ",i);
	    printf("Length of last output: %8.4f\n",Vector_length(G_out));
	}

	GoToXY(1,1);
	printf(">> Testing after learning.");
	GoToXY(1,12); for(i=0; i<80; i++) printf(" ");      /* clear line */
	for(i=0; i<Number_of_pairs; i++) {
	    GoToXY(1,12);
	    Matrix_times_vector(A, F_vectors[i], G_out);
	    Goodness_of_recall=Vector_length(G_out);
	    printf("Pairs tested: %4d  ",i);
	    printf("Length of last output    : %8.4f\n",Goodness_of_recall);
	    Goodness_of_recall*=10;
	    /* Round Goodness_of_recall to the nearest integer */
	    Goodness_of_recall += (Goodness_of_recall>0)? 0.5: -0.5;
	    Index=(int)Goodness_of_recall;
	    if(Index>=-Max_histogram && Index<=Max_histogram)
		LS_histogram[Max_histogram+Index]++;
	}

	GoToXY(1,1);
	printf(">> Using test vectors.     ");
	GoToXY(1,12); for(i=0; i<80; i++) printf(" ");      /* clear line */
	for(i=0; i<Number_in_test_set; i++) {
	    GoToXY(1,12);
	    Matrix_times_vector(A, Test_vectors[i], G_out);
	    Goodness_of_recall=Vector_length(G_out);
	    printf("Pairs tested: %4d  ",i);
	    printf("Length of test output    : %8.4f\n",Goodness_of_recall);
	    Goodness_of_recall*=10;
	    /* Round Goodness_of_recall to the nearest integer */
	    Goodness_of_recall += (Goodness_of_recall>0)? 0.5: -0.5;
	    Index=(int)Goodness_of_recall;
	    if(Index>=-Max_histogram && Index<=Max_histogram)
		TS_histogram[Max_histogram+Index]++;
	}
	Display_outputs();

        /* Repeat? */
        GoToXY(1,21);
#if defined MSDOS
        printf("\n>> Do you want to repeat this demonstration? Y/N ");
        Input_char=getch(); while(kbhit()) getch();
#else
        getchar();
        do {
            printf("\n>> Do you want to repeat this demonstration? Y/N ");
            Input_char=getchar();
        } while(Input_char!='y' && Input_char!='Y' &&
                Input_char!='n' && Input_char!='N');
#endif
    } while(Input_char=='Y' || Input_char=='y');
    GoToXY(1,23);
}   /* main */

