/* File appj.c, Program WhChar */

/*
                           Appendix J
              Character Based Demonstration of the 
                     Widrow-Hoff Algorithm

     This program is almost identical to the program in Appendix I 
except it implements the Widrow-Hoff gradient descent algorithm.  The 
data base of states used to test it in Chapter 9 is appended.  The
state data base is autoassociative, with the input and output
patterns identical.  A number of test stimuli are presented in
this data base to check the accuracy of retrieval, that is,
to see whether, given a state name, it can retrieve the name of
the capital, and vice versa.

     When the program is run, it prompts for the name of the file
containing the strings to learn.  Then it requires a random number seed.  
(Strings are presented randomly during learning.)  This program uses 
the Unix library function rand(), which will have to be replaced with 
the random number generator for other computers if necessary.  Next, an 
interpretation threshold is required.  Zero is the value to use for most 
applications.  The next prompt is for the number of pairs to learn.  This
give the number of strings to be associated.  In the state database at 
the end of the Appendix, this would be 15.  The other patterns, above 15,  
are test patterns and will be tested after learning is finished.  The 
next prompt, total learning trials, determines how long iteration will 
last. Widrow-Hoff is an iterative procedure.  Associations from the training
set will be chosen at random until the number of learning trials is reached.  
For 15 patterns relating to states, 200 or 300 pattern presentations is 
more than satisfactory.  As learning proceeds, the program will print out 
information about how well learning is going.  To save time, this version 
of the program only prints to the screen statistics when an odd-numbered 
pair is learned.  After learning, the entire set of input patterns in the 
input file is tested.

    See pp246-254, in particular the example on p 254.
*/

#if defined MSDOS
#include <io.h>         /* access */
#else
#include <unistd.h>     /* access */
#endif
#include <math.h>       /* fabs sqrt */
#include <stdio.h>      /* feof fgets fopen fclose printf scanf */
#include <stdlib.h>     /* rand srand RAND_MAX */
#include <string.h>     /* strchr strncpy strcat*/

#define Number_of_characters (int)(25)
#define Dimensionality       (int)(8*Number_of_characters)
#define Max_no_associations  (int)(100)

typedef float Matrix[Dimensionality][Dimensionality];
typedef float Vector[Dimensionality];
typedef char String[Number_of_characters+1];
typedef int Byte[8];
typedef int IVector[Dimensionality];

int Number_of_associations;
int Pairs_to_learn;
int Total_learning_trials;
Matrix A,Delta_A;
char File_name[40];
float Threshold=0;
float Eta_prime;
Vector F_Set[Max_no_associations];
Vector G_Set[Max_no_associations];

void Subtract_vectors(float Vector_1[], float Vector_2[], float Vector_difference[]);
void Scalar_times_vector(float Scalar, Vector Vector_in, Vector Vector_out);
float Inner_product(float Vector_1[], float Vector_2[]);
float Vector_length(float Vector_1[]);
void Normalize(float Vector_in[], float Vector_out[]);
float Vector_cosine(float Vector_1[], float Vector_2[]);
void Matrix_times_vector(Matrix A, Vector F, Vector G);
void Add_matrices(Matrix Mx_1, Matrix Mx_2, Matrix Matrix_sum);
void Outer_product(Vector G, Vector F, Matrix A);

void Interpret(float Threshold,   /* Determines interpretation threshold.*/
	       Vector V,                /* Input state vector. */
	       String Interpretation);  /* Output interpreted string. */
static void Make_Binary_Vector(float Threshold, Vector V, IVector IV);
static int Count(Byte B);
static char Byte_Interpret(Byte B);
static void Read_IVector_as_ASCII(IVector IV, String Interpretation);
void Make_stimulus(String Input_string,   /* Character string to be */
					  /* turned into a state vector. */
		   Vector Output_vector); /* Output state vector. */
static void Zero_byte(Byte B);
static void Nr_to_byte(int Nr, Byte B);
static void Add_Byte_to_Vector(Vector V, int i, Byte B);
void Write_string(String S);
void Read_input_line(FILE* IPFile, String Input_string);
void Read_stimuli(void);
void Associate(Vector F, Vector G);
void Test_whole_set(void);
void Initialize(void);
float Random(void);
void main(void);

/*--------------------------------------------------------------------------
Subtracts two vectors.
Arguments
    float Vector_1[]            One vector, input
    float Vector_2[]            Other vector, input
    float Vector_difference[]   Vector_1-Vector_2, returned
--------------------------------------------------------------------------*/
void Subtract_vectors(float Vector_1[], float Vector_2[], 
                      float Vector_difference[]) {
    int i;      /* loop index */

    for(i=0; i<Dimensionality; i++)
	Vector_difference[i]=Vector_1[i] - Vector_2[i];
}   /* Subtract_vectors */



/*--------------------------------------------------------------------------
Scales a vector.
Arguments
    float Scalar                Factor, input
    float Vector_in[]           Vector, input
    float Vector_out[]          Factor*Vector_in, returned
--------------------------------------------------------------------------*/
void Scalar_times_vector(float Scalar, Vector Vector_in, Vector Vector_out) {
    int i;      /* loop index */

    for(i=0; i<Dimensionality; i++)
	Vector_out[i]=Scalar*Vector_in[i];
}   /* Subtract_vectors */



/*--------------------------------------------------------------------------
Forms the inner product of two vectors.
Arguments
    float Vector_1[]            One vector, input
    float Vector_2[]            Other vector, input
Returns
    float Inner_product         Vector_1 . Vector_2
--------------------------------------------------------------------------*/
float Inner_product(float Vector_1[], float Vector_2[]) {
    float Sum_of_products;      /* Name says it all */
    int i;                      /* loop index */

    Sum_of_products=0;
    for(i=0; i<Dimensionality; i++)
	Sum_of_products += Vector_1[i]*Vector_2[i];
    return Sum_of_products;
}   /* Inner_product */


/*--------------------------------------------------------------------------
Returns the length of a vector.
Argument
    float Vector_1[]            Vector, input
Returns
    float length                Magnitude of Vector_1
--------------------------------------------------------------------------*/
float Vector_length(float Vector_1[]) {
    float Sum_of_Squares;       /* Name says it all */
    int i;                      /* loop index */

    Sum_of_Squares=0;
    for(i=0; i<Dimensionality; i++)
	Sum_of_Squares += Vector_1[i]*Vector_1[i];
    return (float)sqrt(Sum_of_Squares);
}   /* Vector_length */



/*--------------------------------------------------------------------------
Scales the input vector to unit length.  This is not possible if the input
vector has zero length, and in that case an error message is printed, and
the output vector is set to zero.

This routine may be called with both arguments pointing to the same array.

Arguments
    float Vector_in[]           Vector, input
    float Vector_out[]          Normalized vector, output
--------------------------------------------------------------------------*/
void Normalize(float Vector_in[], float Vector_out[]) {
    float Length_of_vector_in;  /* Name says it all */
    int i;                      /* loop index */

    Length_of_vector_in=Vector_length(Vector_in);
    if(Length_of_vector_in==0) {
	printf("Normalize: Input vector has zero length!\n");
	for(i=0; i<Dimensionality; i++)
	    Vector_out[i] = 0;
    } else {
	for(i=0; i<Dimensionality; i++)
	    Vector_out[i] = Vector_in[i]/Length_of_vector_in;
    }
}   /* Normalize */



/*--------------------------------------------------------------------------
Calculates the cosine of the 'angle' between two vectors.
Arguments
    float Vector_1[]            One vector, input
    float Vector_2[]            Other vector, input
Returns
    float Vector_cosine         (Vector_1 . Vector_2)/(|Vector_1| |Vector_2|)
--------------------------------------------------------------------------*/
float Vector_cosine(float Vector_1[], float Vector_2[]) {
    float Length_of_vector_1, Length_of_vector_2;     /* vector magnitudes */

    Length_of_vector_1=Vector_length(Vector_1);
    Length_of_vector_2=Vector_length(Vector_2);

    if(Length_of_vector_1==0 || Length_of_vector_2==0) {
	printf("Vector_cosine: Input vector has zero length!\n");
	return 0;
    } else {
	 return Inner_product(Vector_1,Vector_2)/
                                   Length_of_vector_1/Length_of_vector_2;
    }
}   /* Vector_cosine */



/*-------------------------------------------------------------------------
This function forms the product Af = g where both f and g are column
vectors.
-------------------------------------------------------------------------*/
void Matrix_times_vector(Matrix A, Vector F, Vector G) {
    int i,j;    /* loop indices */

    /* Zero vector G */
    for(i=0; i<Dimensionality; i++)
	G[i]=0;
    for(i=0; i<Dimensionality; i++)
	for(j=0; j<Dimensionality; j++)
	    G[i] += A[i][j]*F[j];
}   /* Matrix_times_vector */


/*-------------------------------------------------------------------------
Calculates the sum of 2 matrices dimensioned [Dimensionality][Dimensionality]
Arguments
    Matrix Mx_1[]            One matrix, input
    Matrix Mx_2[]            Other matrix, input
Returns
    Matrix Matrix_sum        Mx_1 + Mx_2
-------------------------------------------------------------------------*/
void Add_matrices(Matrix Mx_1, Matrix Mx_2, Matrix Matrix_sum) {
    int i,j;    /* loop indices */

    for(i=0; i<Dimensionality; i++)
	for(j=0; j<Dimensionality; j++)
	    Matrix_sum[i][j]=Mx_1[i][j]+Mx_2[i][j];
}   /* Add_matrices */


/*-------------------------------------------------------------------------
We want to generate an outer product matrix A so that
	    T
     A = G F
-------------------------------------------------------------------------*/
void Outer_product(Vector G, Vector F, Matrix A) {
    int i,j;    /* loop indices */

    for(i=0; i<Dimensionality; i++)
	for(j=0; j<Dimensionality; j++)
	    A[i][j] = G[i]*F[j];
}   /* Outer_product */


/*------------------------------------------------------------------------
		Interpret
This routine takes a vector of floats, applies a threshold to so that
each element is classifed as +1) probably set; -1) probably clear; and 0)
ambiguous.  These intermediate values are stored in int IV[].  Then this
array is turned into a character string and returned as <Interpretation>.
The string will in general contain both ascii characters that have been
identified, and the character '_' if it is still ambiguous.

This routine is used in Test_whole_set() to turn the raw association
output from Matrix_times_vector() into a string.  It is also used there
to recover the input and required output strings from F_Set[] and G_Set[].
The arrays F_Set[] and G_Set[] consist of float values that are either
zero or greater than zero, and so for this second application it is just
a matter of setting Threshold to zero and of supplying a vector from
F_Set[] or G_Set[].

Uses void Make_Binary_Vector()
     int Count()
     char Byte_Interpret()
     void Read_IVector_as_ASCII()
Arguments
    float Threshold          Value between 0 and 1
    Vector V                 array to be interpreted
Returns
    String Interpretation    string of Number_of_characters, plus null
------------------------------------------------------------------------*/
void Interpret(float Threshold,   /* Determines interpretation threshold. */
	       Vector V,                /* Input state vector. */
	       String Interpretation) { /* Output interpreted string. */
    IVector IV;

    Make_Binary_Vector(Threshold, V, IV);
    Read_IVector_as_ASCII(IV, Interpretation);
}   /* Interpret */


/*------------------------------------------------------------------------
Applies a threshold to the values in <V>, and writes result to <IV>.
Each value in these arrays is considered to correspond to a bit of
a character string.  If the value is >Threshold, that bit is considered
set; if it is < -Threshold, that bit is considered cleared; otherwise it
is considered ambiguous.  'Set' is indicated by +1 in IV[], 'Cleared' is
indicated by -1 in IV[], and 'Ambiguous' is indicated by 0.

This routine is also used to recover the bit pattern of the input strings
from the float arrays F_Set[] and G_Set[].  These arrays consist of values
that are either less than zero or greater than zero.  Hence Threshold
is set to zero, and the if statements below should then classify each
float value either as +1 or -1 and never 0.

To make the behaviour when Threshold=0 clearer, I have changed the
inequality below from >= to >.

Used by Interpret()
Arguments
    float Threshold          Value between 0 and 1
    Vector V                 array to be interpreted
Returns
    IVector IV               result of applying threshold: array of -1,0 +1
------------------------------------------------------------------------*/
static void Make_Binary_Vector(float Threshold, Vector V, IVector IV) {
    int j;      /* loop index */

    for(j=0; j<Dimensionality; j++) {
	IV[j]=0;
	if(V[j] >   (float)fabs(Threshold)) IV[j]= 1;
	if(V[j] <= -(float)fabs(Threshold)) IV[j]=-1;
    }
}   /* Make_Binary_Vector */


/*------------------------------------------------------------------------
Counts the number of ambiguous bits (B[i]==0) indicated by the array <B>.
Used by Interpret()
------------------------------------------------------------------------*/
static int Count(Byte B) {
    int i;      /* loop index */
    int sum;    /* number of zeros in byte */

    sum=0;
    for(i=0; i<7; i++)          /* Ignore parity bit. */
	if(B[i]==0) sum++;
    return sum;
}   /* Count */


/*------------------------------------------------------------------------
B is an array of eight integers each equal to +1, 0 or -1.  This routine
converts it to an ascii character.  If a group of 8 contains any 0 values,
the corresponding character is considered ambiguous and a '_' results.
Otherwise +1 causes the corresponding bit to be set, and -1 causes the
corresponding bit to be cleared.

Used by Interpret()
------------------------------------------------------------------------*/
static char Byte_Interpret(Byte B) {
    int i;      /* loop index */
    int Charnr; /* ASCII code of character, as an integer */
    Byte C;	/* copy of B */

    /* First check to see if all byte elements + 1 or - 1. */
    if(Count(B)>0)
	return '_';     /* character is still ambiguous */
    else {
	/* Convert B[] to an byte of zero and +1. */
	for(i=0; i<8; i++)
            C[i]=(B[i]==-1)? 0: 1;
	Charnr=64*C[6] + 32*C[5] + 16*C[4] + 8*C[3] + 4*C[2] + 2*C[1] + C[0];
	if(Charnr==0 || Charnr==127)    /* Zero/delete convention is '_' */
	    return '_';
	else if(Charnr<32)              /* Non-printing character. */
	    return '#';
	else                            /* Print character. */
	    return (char)Charnr;
    }
}   /* Byte_Interpret */


/*------------------------------------------------------------------------
Takes the vector <IV>, consisting of 8*Number_of_characters integers, and
converts it to an ascii string.

Used by Interpret()
Uses Byte_Interpret()
Arguments
    IVector IV               result of applying threshold: array of -1,0 +1
Returns
    String Interpretation    ascii string corresponding to IV
------------------------------------------------------------------------*/
static void Read_IVector_as_ASCII(IVector IV, String Interpretation) {
    int i,j;    /* loop indices */
    Byte B;     /* eight consecutive values copied from IV[] */

    for(j=0; j<Number_of_characters; j++) {
	/* Next 'for loop' flips order of vector elements. */
	for(i=0; i<8; i++)
	    B[7-i]=IV[j*8+i];
	Interpretation[j]=Byte_Interpret(B);
    }
    Interpretation[Number_of_characters]='\0';  /* Add null termination */
}   /* Read_IVector_as_ASCII */



/*------------------------------------------------------------------------
		Make_stimulus
This routine is called by Read_stimuli.  It takes a string containing
<Number_of_characters> ascii characters, and for each of the
8*Number_of_characters bits sets the corresponding element of
<Output_vector> to +1 (if bit is set), or to 0 (if bit is clear).
Subsequently in Initialize(), the output vectors will be 'normalized'
by dividing each element by sqrt(Dimensionality).
Uses void Zero_byte()
     void Nr_to_byte()
     void Add_Byte_to_Vector()
Arguments
    String Input_string      string of Number_of_characters, plus null
Returns
    Vector Output_vector     8*Number_of_characters integers, each 1 or 0
------------------------------------------------------------------------*/
void Make_stimulus(String Input_string, Vector Output_vector) {
    int i;      /* loop index */
    Byte B;     /* One character, expanded to an array of 8 integers */

    for(i=0; i<Number_of_characters; i++) {
	Nr_to_byte((int)Input_string[i],B);
	if(Input_string[i]=='_') Zero_byte(B);
	Add_Byte_to_Vector(Output_vector, i ,B);
    }
}   /* Make_stimulus */


/*------------------------------------------------------------------------
Sets the array of 8 integers to zero.
Called by Make_stimulus()
------------------------------------------------------------------------*/
static void Zero_byte(Byte B) {
    int i;      /* loop index */

    for(i=0; i<8; i++)
	B[i]=0;
}   /* Zero_byte */


/*------------------------------------------------------------------------
Copies each of the seven least significant bits of <Nr> to the elements
of B[].
Called by Make_stimulus()
------------------------------------------------------------------------*/
static void Nr_to_byte(int Nr, Byte B) {
    int i;              /* loop index */
    int Parity=0;       /* sum of all bits */
    int mask;           /* bit mask */

    if(Nr>=128) {
	printf("Too large. ASCII conversion fails.\n"); return;
    } else if(Nr<0) {
	printf("Negative number. ASCII conversion fails.\n"); return;
    }

    /* Copy bits 6..0 of Nr to the corresponding elements of B[] */
    for(mask=0x40,i=6; i>=0; mask>>=1,i--) {
	if(Nr&mask) {
	    B[i]=1; Parity++;
	} else
	    B[i]=-1;
    }
    /* Set bit 7 such that the sum of all bits mod 2 is equal to 0 */
    B[7]=(Parity&1)? 1: -1;
}   /* Nr_to_byte */


/*------------------------------------------------------------------------
Copies 8 integers from B to sequential elements of V, starting from
element 8*i.
Called by Make_stimulus()
------------------------------------------------------------------------*/
static void Add_Byte_to_Vector(Vector V, int i, Byte B) {
    int j;      /* loop index */

    if(i<0 || i>=Number_of_characters)
        printf("Add_Byte_to_Vector(): i=%d is out of range\n",i);
    else
        for(j=0; j<8; j++)
            V[i*8+j]=B[7-j];
}   /* Add_Byte_to_Vector */



/*-------------------------------------------------------------------------
Prints the specified string to screen.
-------------------------------------------------------------------------*/
void Write_string(String S) {
    printf("%s",S);
}   /* Write_string */


/*-------------------------------------------------------------------------
Reads one line of input from stream IPFile, and copies to the string
Input_string.
Up to 80 characters can be read, but only the first 25 will be copied.
If there are less than 25 (even 0) characters read, then the contents
of Input_string will be padded with blank characters to a length of 25.
-------------------------------------------------------------------------*/
void Read_input_line(FILE* IPFile, String Input_string) {
    char Input_line[80+2];      /* maximum of 80 characters+newline+null */

    /* read one line from input, and remove newline character, if present */
    fgets(Input_line,sizeof(Input_line),IPFile);
    if(strchr(Input_line,'\n'))
	*strchr(Input_line,'\n')='\0';

    /* Copy at most <Number_of_characters> characters from input line */
    strncpy(Input_string,Input_line,sizeof(String));
    Input_string[sizeof(String)-1]='\0';        /* Ensure null termination */

    /* Pad input with blanks, if necessary */
    while(strlen(Input_string)<Number_of_characters)
	strcat(Input_string," ");
}   /* Read_input_line */


/*-------------------------------------------------------------------------
Read the contents of the file specified by <File_name>, which should
consist of groups of three lines.  The first line can contain anything,
the second will be copied to F_Set[n], and the third to G_Set[n].  The
total number of groups read successfully is stored as Number_of_associations.
-------------------------------------------------------------------------*/
void Read_stimuli(void) {
    String Input_string;        /* 25 characters read from input */
    FILE *Stimuli;              /* pointer to data file */

    Stimuli=fopen(File_name,"rt");
    Number_of_associations=0;
    do {
	/* Read 1st line, but this is ignored */
	Read_input_line(Stimuli,Input_string);
	if(feof(Stimuli)) break;

	/* Read 2nd line.  Second line is the F stimulus. */
	Read_input_line(Stimuli,Input_string);
	if(feof(Stimuli)) break;
	Make_stimulus(Input_string, F_Set[Number_of_associations]);

	/* Read 3rd line.  Third line is the G stimulus. */
	Read_input_line(Stimuli,Input_string);
	if(feof(Stimuli)) {printf("Incomplete input\n"); break;}
	Make_stimulus(Input_string, G_Set[Number_of_associations]);

	Number_of_associations++;
    } while(!feof(Stimuli) && Number_of_associations<=Max_no_associations);
    fclose(Stimuli);
}   /* Read_stimuli */


/*-------------------------------------------------------------------------
Form the association of a pair of F and G column vectors (calculate the
outer product G F^T), and add this into the array A.
-------------------------------------------------------------------------*/
void Associate(Vector F, Vector G) {
   Outer_product(G, F, Delta_A);
   Add_matrices(A, Delta_A, A);
}   /* Associate */


/*-------------------------------------------------------------------------
For each of the associations, input, ideal output and actual output are
compared.
-------------------------------------------------------------------------*/
void Test_whole_set(void) {
    int i;                      /* loop index */
    Vector Output_vector;       /* A*F, which is an approximation to G */
    String F_string;            /* Input string */
    String G_string;            /* Required output string */
    String Interpretation;      /* Actual output string */

    for(i=0; i<Number_of_associations; i++) {
	Matrix_times_vector(A, F_Set[i], Output_vector);
	Interpret(0,         F_Set[i], F_string);
	Interpret(0,         G_Set[i], G_string);
	Interpret(Threshold, Output_vector, Interpretation);

	printf("\n%d",i+1);
	printf(".   Cosine: %7.5f", Vector_cosine(Output_vector, G_Set[i]));
	printf(".   Length: %7.5f", Vector_length(Output_vector));
	printf("\n");
	printf(" Input          ==> %s\n",F_string);
	printf(" Desired Output ==> %s\n",G_string);
	printf(" Actual Output  ==> %s\n",Interpretation);
    }
}   /* Test_whole_set */


/*-------------------------------------------------------------------------
This routine gets the name of the file containing association strings,
and a threshold value for use by Interpret().  It calls the routine
Read_stimuli() to fill the arrays F_Set[] and G_Set[], and finally
zeros the association arrays A[][] and Delta_A[][].
-------------------------------------------------------------------------*/
void Initialize(void) {
    int i,j;            /* loop indices */
    unsigned int Seed;  /* seed for srand() */
    float Eta;          /* learning constant */

    printf("\n\nDemonstration of Associative Memory. Uses Widrow-Hoff\n");
    printf("Reads character vectors from a file.  Learns and recalls\n");
    printf("the pairs of associations.  See pp 249-254.\n");
    printf("\n");

    /* Get file containing association strings, and threshold */
    do {
	printf("File name of stimuli (eg states10.txt) : "); 
        scanf("%39s",File_name);
    } while(access(File_name,4)!=0);
    Read_stimuli();     /* Read data file to F_Set[][] and G_Set[][] */

    printf("Seed for random number generator (0..65535): ");
    scanf("%u",&Seed); srand(Seed);

    do {
	printf("Threshold (0..1; approx 0.0): "); scanf("%f",&Threshold);
    } while(Threshold<0 || Threshold>1);

    do {
	printf("Different pairs to learn   (1..%d): ",Number_of_associations);
	scanf("%d",&Pairs_to_learn);
    } while(Pairs_to_learn<1 && Pairs_to_learn>Number_of_associations);

    do {
	printf("Total learning trials  (1...; approx 300): ");
	scanf("%d",&Total_learning_trials);
    } while(Total_learning_trials<1);

    /* This is length of a vector completely composed of +1 and -1 */
    Eta=(float)(sqrt((float)Dimensionality));
    Eta_prime=10*Eta;   /* Learning constant. This is length of a vector */
			/* completely composed of +1 and -1 */

    /* Normalize input vectors, and zero A[][] and Delta_A[][] */
    /*
    for(i=0; i<Number_of_associations; i++)
	for(j=0; j<Dimensionality; j++)
	    F_Set[i][j]/=Eta;
    */
    for(i=0; i<Dimensionality; i++)
	for(j=0; j<Dimensionality; j++) {
	    A[i][j]*=0;
	    Delta_A[i][j]*=0;
	}
}   /* Initialize */


/*--------------------------------------------------------------------------
Returns a float number >=0 and < 1.0
--------------------------------------------------------------------------*/
float Random(void) {
    return ((float)rand())/((float)RAND_MAX+1);
}   /* Random */


/*--------------------------------------------------------------------------
Program WhChar
--------------------------------------------------------------------------*/
void main(void) {
    int i,j;            /* loop indices */
    int Random_pair;    /* randomly selected pair */
    static Vector Error_signal;
    static Vector Output_vector;  /* A*F, which is an approximation to G */

    Initialize();
    printf("Threshold: %6.2f\n",Threshold);
    printf("Correct length of output vectors: %8.5f\n",sqrt((float)Dimensionality));

    /* Report progess of training */
    for(i=0; i<Total_learning_trials; i++) {
	Random_pair=(int)(Random()*Pairs_to_learn); /* Chooses random pair. */
	Matrix_times_vector(A, F_Set[Random_pair], Output_vector);
	for(j=0; j<Dimensionality; j++)
	    Error_signal[j]=(G_Set[Random_pair][j]-Output_vector[j])/Eta_prime;
	Associate(F_Set[Random_pair], Error_signal);
	if(Random_pair&1) {
	    Matrix_times_vector(A, F_Set[Random_pair], Output_vector);
	    printf("Trial: %3d  Pair learned: %3d",i+1,Random_pair);
	    printf("  Cosine: %7.5f",
                Vector_cosine(Output_vector, G_Set[Random_pair]));
	    printf("  Length: %7.3f\n",Vector_length(Output_vector));
	}
    }

    /* Show result of training */
    Test_whole_set();
}   /* main */


/*--------------------------------------------------------------
File states.txt for testing purposes.  Includes 15 states to
be learned, plus a number of test vectors.
This is an autoassociative stimulus set.
--------------------------------------------------------------
1234567890123456789012345
Boston      Massachusetts
Boston      Massachusetts
2
Providence   Rhode Island
Providence   Rhode Island
3
Sacramento     California
Sacramento     California
4
Bismarck     North Dakota
Bismarck     North Dakota
5
Oklahoma City    Oklahoma
Oklahoma City    Oklahoma
6
Harrisburg   Pennsylvania
Harrisburg   Pennsylvania
7
Charleston  West Virginia
Charleston  West Virginia
8
Baton Rouge     Louisiana
Baton Rouge     Louisiana
9
Concord     New Hampshire
Concord     New Hampshire
10
Columbia   South Carolina 
Columbia   South Carolina 
11
Albany           New York
Albany           New York
12
Trenton        New Jersey
Trenton        New Jersey
13
Juneau             Alaska
Juneau             Alaska
14
Talahassee        Florida
Talahassee        Florida
15
Austin              Texas
Austin              Texas
1234567890123456789012345
Boston___________________
Boston      Massachusetts
17
Providence_______________
Providence   Rhode Island
18
Sacramento_______________
Sacramento     California
19
Bismarck_________________
Bismarck     North Dakota
20
Oklahoma City____________
Oklahoma City    Oklahoma
21
Harrisburg_______________
Harrisburg   Pennsylvania
22
Charleston_______________
Charleston  West Virginia
23
Baton Rouge______________
Baton Rouge     Louisiana
24
Concord__________________
Concord     New Hampshire
25
Columbia_________________ 
Columbia   South Carolina
26
Albany___________________
Albany           New York
27
Trenton__________________
Trenton        New Jersey
28
Juneau___________________
Juneau             Alaska
29
Talahassee_______________
Talahassee        Florida
30
Austin___________________
Austin              Texas
1234567890123456789012345
____________Massachusetts
Boston      Massachusetts
32
_____________Rhode Island
Providence   Rhode Island
33
_______________California
Sacramento     California
34
_____________North Dakota
Bismarck     North Dakota
35
_________________Oklahoma
Oklahoma City    Oklahoma
36
_____________Pennsylvania
Harrisburg   Pennsylvania
37
____________West Virginia
Charleston  West Virginia
38
________________Louisiana
Baton Rouge     Louisiana
39
____________New Hampshire
Concord     New Hampshire
40
___________South Carolina 
Columbia   South Carolina 
41
_________________New York
Albany           New York
42
_______________New Jersey
Trenton        New Jersey
43
___________________Alaska
Juneau             Alaska
44
__________________Florida
Talahassee        Florida
45
____________________Texas
Austin              Texas
*/
