/* File associat.c, Program associat */

/*----------------------------------------------------------------------
Copyright James Anderson, Department of Psychology, Brown University.
This research was primarily supported by National Science
Foundation Grants BNS-82-14728, BNS-85-18675 and BNS-90-23283
to: James Anderson,  Box 1978
    Department of Cognitive and Linguistic Sciences, Brown University
    Providence, RI 02912.
    James_Anderson@brown.edu

Please acknowledge these grants if you make use of these program
in published material.
 
Versions of these programs are used for the demonstrations in 
Chapters 15, 16, and 17 of "An Introduction to Neural Networks"
by James A. Anderson, MIT Press
----------------------------------------------------------------------*/

#include <math.h>       /* sqrt */
#include <stdio.h>      /* fgets getchar printf scanf */
#include <stdlib.h>	/* rand srand */
#include <string.h>     /* strncpy */

/* Number_of_neurons is related to Dimensionality of vectors */
#define Number_of_neurons   (int)(200)
/* Maximum_set_size is the size of each training set */
#define Maximum_set_size    (int)(100)
/* String_length refers to strings in the struct Stimulus.Name */
#define String_length  (int)(60)
#define FALSE 0
#define TRUE  1

typedef float Vector[Number_of_neurons];
typedef char String[String_length+1];
typedef struct {
    String Name;
    Vector Val;
} Stimulus;
typedef struct {
    float Strength;
    int From;   /* 0<=From<=Number_of_neurons-1 */
} Synapse;
typedef struct {
    float Activity;
    Synapse Synapses[Number_of_neurons];
    float Neuron_output;
    float Learning;
    float Neuron_upper_limit;
    float Neuron_lower_limit;
} Neuron;
 
/* Defined in associat.c */
/* Element [0] does not appear to be used anywhere */
Stimulus F_set[Maximum_set_size+1];
Stimulus G_set[Maximum_set_size+1];
Neuron Neurons[Number_of_neurons];
int Number_of_synapses;
char FFileName[FILENAME_MAX];
char GFileName[FILENAME_MAX];
char NFileName[FILENAME_MAX];
int Nr_in_f_set;
int Nr_of_stimuli;
int Nr_to_learn;
int Correction;


void Read_f_file(void);
void Read_g_file(void);
void Write_f_and_g(void);
float Random(void);
float Vector_length(int Dimensionality, float Vector_1[]);
float Inner_product(int Dimensionality, float Vector_1[], float Vector_2[]);
void Read_Nfile(void);
void Write_Nfile(void);
int Check_froms(int NR, Neuron *Neu);
void Setup(int Fully_connected);
void Initialize(void);
void Calculate(Vector F, Vector O);
void Memorize(Vector F, Vector G, Vector O);
void Write_cosine(float Vector_1[], float Vector_2[]);
void main(int argc, char *argv[]);


/*----------------------------------------------------------------------
Read stimuli from the input file <FFileName> to F_set[1], F_set[2], ...
F_set[Nr_in_f_set].  (F_set[0] is left blank.)
----------------------------------------------------------------------*/
void Read_f_file(void) {
    FILE *IPFile;	/* for accessing F file */

    printf("\nThe program is reading F file \'%s\'.\n",FFileName);
    printf("The dimensionality of the system is %d\n", Number_of_neurons);

    Nr_in_f_set=0;
    IPFile=fopen(FFileName,"r");
    if(IPFile==NULL) {
        printf("Read_f_file(): Error opening input file\n");
    } else {
        while(!feof(IPFile) && Nr_in_f_set<Maximum_set_size) {
            if(fread(&F_set[Nr_in_f_set+1],sizeof(Stimulus),1,IPFile)==1)
                Nr_in_f_set++;
        }
    }
    fclose(IPFile);
    printf("\n");
}   /* Read_f_file */


/*----------------------------------------------------------------------
Read stimuli from the input file <GFileName> to G_set[1], G_set[2], ...
G_set[Nr_in_g_set].  (G_set[0] is left blank.)

It is not necessary to have separate f and g sets in a strictly
autoassociative system, but in other applications, the f and g sets
can be different.
----------------------------------------------------------------------*/
void Read_g_file(void) {
    FILE *IPFile;	/* for accessing G file */
    int Nr_in_g_set;

    printf("\nThe program is reading G file \'%s\'.\n",GFileName);

    Nr_in_g_set=0;
    IPFile=fopen(GFileName,"r");
    if(IPFile==NULL) {
        printf("Read_g_file(): Error opening input file\n");
    } else {
        while(!feof(IPFile) && Nr_in_g_set<Maximum_set_size) {
            if(fread(&G_set[Nr_in_g_set+1],sizeof(Stimulus),1,IPFile)==1)
                Nr_in_g_set++;
        }
    }
    fclose(IPFile);
    printf("\n");

    /* Set Nr_of_stimuli */
    if(Nr_in_f_set!=Nr_in_g_set)
        printf("WARNING! Number of Fs does not equal number of Gs.\n");
    Nr_of_stimuli=(Nr_in_f_set<Nr_in_g_set)? Nr_in_f_set: Nr_in_g_set;
}   /* Read_g_file */


/*----------------------------------------------------------------------
Print all F and G stimuli
----------------------------------------------------------------------*/
void Write_f_and_g(void) {
    int i;	/* loop index */

    printf("\nF and G stimuli used.\n");
    for(i=1; i<=Nr_of_stimuli; i++) {
        printf("\n F[%2d] : %s\n",i,F_set[i].Name);
        printf(" G[%2d] : %s\n",i,G_set[i].Name);
    }
    printf("\n");
}   /* Write_f_and_g */

/*--------------------------------------------------------------------------
Returns a float number >=0 and < 1.0
--------------------------------------------------------------------------*/
float Random(void) {
    return ((float)rand())/((float)RAND_MAX+1);
}   /* Random */

/*--------------------------------------------------------------------------
Returns the length of a vector.
Arguments
    int Dimensionality          Number of elements in vector
    float Vector_1[]            Vector, input
Returns
    float length                Magnitude of Vector_1
--------------------------------------------------------------------------*/
float Vector_length(int Dimensionality, float Vector_1[]) {
    float Sum_of_Squares;       /* Name says it all */
    int i;                      /* loop index */

    Sum_of_Squares=0;
    for(i=0; i<Dimensionality; i++)
        Sum_of_Squares += Vector_1[i]*Vector_1[i];
    return (float)sqrt(Sum_of_Squares);
}   /* Vector_length */

/*--------------------------------------------------------------------------
Forms the inner product of two vectors.
Arguments
    int Dimensionality          Number of elements in vector
    float Vector_1[]            One vector, input
    float Vector_2[]            Other vector, input
Returns
    float Inner_product         Vector_1 . Vector_2
--------------------------------------------------------------------------*/
float Inner_product(int Dimensionality, float Vector_1[], float Vector_2[]) {
    float Sum_of_products;      /* Name says it all */
    int i;                      /* loop index */

    Sum_of_products=0;
    for(i=0; i<Dimensionality; i++)
        Sum_of_products += Vector_1[i]*Vector_2[i];
    return Sum_of_products;
}   /* Inner_product */


/*----------------------------------------------------------------------
----------------------------------------------------------------------*/
void Read_Nfile(void) {
    int Nr_of_neurons;  /* size of Nfile */
    FILE *Nfile;

    printf("\nThe program is reading N file \'%s\'.\n",NFileName);
    printf("This is a file of type Neuron\n");

    Nfile=fopen(NFileName,"r");
    if(Nfile==NULL) {
        printf("Neuron file could not be opened.\n");
        Nr_of_neurons=0;
        return;
    }

    /* Read file */
    for(Nr_of_neurons=0; Nr_of_neurons<Number_of_neurons; Nr_of_neurons++)
        if(fread(&Neurons[Nr_of_neurons],sizeof(Neuron),1,Nfile)==0) break;
    fclose(Nfile);

    /* Check size */
    printf("The neuron file contains %d neurons.\n",Nr_of_neurons);
    if(Nr_of_neurons!=Number_of_neurons)
        printf("ERROR! Dimensionality does not agree with file size.\n");
    printf("\n");
}   /* Read_Nfile */

 
/*----------------------------------------------------------------------
----------------------------------------------------------------------*/
void Write_Nfile(void) {
    int i;      /* loop index */
    FILE *OPFile;

    OPFile=fopen(NFileName,"w");
    if(OPFile==NULL) {
        printf("Failed to open %s\n",NFileName);
    } else {
        printf("\nWriting to neuron output file.\n");
        for(i=0; i<Number_of_neurons; i++)
            fwrite(&Neurons[i],sizeof(Neuron),1,OPFile);
        fclose(OPFile);
    }
}   /* Write_Nfile */

 
/*----------------------------------------------------------------------
Checks for the existence of an afferent connection between the given 
neuron and another neuron, number <NR>.
int NR is a random number in the range 0..Number_of_neurons-1
----------------------------------------------------------------------*/
int Check_froms(int NR, Neuron *Neu) {
    int i;	/* loop index */

    for(i=0; i<Number_of_synapses; i++)
        if(Neu->Synapses[i].From == NR) return TRUE;
    return FALSE;
}   /* Check_froms */


/*----------------------------------------------------------------------
Initializes the array of neurons
----------------------------------------------------------------------*/
void Setup(int Fully_connected) {
    int i,j;	/* loop indices */
    int Try;	/* Trial presynaptic neuron: 0..Number_of_neurons-1 */
    
    for(i=0; i<Number_of_neurons; i++) {
        /* Initialize synapses */
        for(j=0; j<Number_of_neurons; j++) {
            Neurons[i].Synapses[j].Strength=0;
            Neurons[i].Synapses[j].From=-1;
        }
        if(Fully_connected) {
            for(j=0; j<Number_of_synapses; j++)
                Neurons[i].Synapses[j].From=(i+j)%Number_of_synapses;
        } else {
            /* Number_of_synapses is typically 50% of Number_of_neurons */
            for(j=0; j<Number_of_synapses; j++) {
                do {
                    Try=(int)(Random()*Number_of_neurons);
                } while(Check_froms(Try,&Neurons[i]));
                Neurons[i].Synapses[j].From=Try;
            }
        } 

        /* Initialize other elements of neuron */
        Neurons[i].Activity   = 0;
	Neurons[i].Learning   = 1;
        Neurons[i].Neuron_output      = 0;
	Neurons[i].Neuron_upper_limit = 0;
        Neurons[i].Neuron_lower_limit = 0;
    }
    printf("Setup completed.\n");
}   /* Setup */

 
/*----------------------------------------------------------------------
----------------------------------------------------------------------*/
void Initialize(void) {
    int Use_Nfile;	/* TRUE if existing Nfile is to be used */
    char InputLine[80]; /* Keyboard input */
    unsigned int Seed;  /* seed for srand() */
    char ch;            /* response */
    int Fully_connected;

    printf("ASSOCIAT program.  October, 1994.\n\n");
    Read_f_file();
    Read_g_file();
    Write_f_and_g();

    printf("Seed for random number generator (0..65535): ");
    scanf("%u",&Seed); srand(Seed);
    getchar();
    printf("\n");

    do {
        printf("Number of associations to learn   (1..%d..): ",Nr_of_stimuli);
        scanf("%d",&Nr_to_learn);
        getchar();
    } while(Nr_to_learn<1);
    printf("\n");

    do {
        printf("Use Windrow-Hoff CORRECTION procedure? (Y or N) : ");
        fgets(InputLine,sizeof(InputLine),stdin);
        ch=InputLine[0];
    } while(ch!='n' && ch!='N' && ch!='y' && ch!='Y');
    Correction=(ch=='y' || ch=='Y');
    printf("\n");

    do {
        printf("Use old Nfile as start? (Y or N) : ");
        fgets(InputLine,sizeof(InputLine),stdin);
        ch=InputLine[0];
    } while(ch!='n' && ch!='N' && ch!='y' && ch!='Y');
    Use_Nfile=(ch=='y' || ch=='Y');
    printf("\n");

    /* When everything else is working, try moving this to below next block */
    /* and deleting the line initializing Number_of_synapses */
    if(Use_Nfile) {
        /* Disk files of type .neu always have <Number_of_neurons> synapses. */
        Number_of_synapses=Number_of_neurons;
        Read_Nfile();
    }

    do {
        printf("Number of synapses  (1..%d): ",Number_of_neurons);
        scanf("%d",&Number_of_synapses);
        getchar();
    } while(Number_of_synapses<1 && Number_of_synapses>Number_of_neurons);
    printf("\n");

    /* Full connectivity simply fills in the connections in order. */
    Fully_connected=(Number_of_synapses==Number_of_neurons);

    /* Connectivities over 80% may take a long time to connect the */
    /* matrix because connections may not be repeated. */
    if(Number_of_synapses > 4*Number_of_neurons/5)
        printf("Warning!  It may take a long time to set up matrix!\n");

    /* Create random set of neuron strengths and connections */
    if(!Use_Nfile)
        Setup(Fully_connected);
}   /* Initialize */


/*--------------------------------------------------------------------------
Calculate does most of the computational work.  This procedure
should be in very fast vector hardware or assembly language.
--------------------------------------------------------------------------*/
void Calculate(Vector F, Vector O) {
    int i,j;            /* loop indices */
    float Sum;          /* total neuronal input */

    for(i=0; i<Number_of_neurons; i++) {
        Sum=0;
        for(j=0; j<Number_of_synapses; j++)
            Sum += Neurons[i].Synapses[j].Strength
                  *F[Neurons[i].Synapses[j].From];
        O[i]=Sum;
    }
}   /* Calculate */


/*----------------------------------------------------------------------
This procedure changes the synaptic weights.  It is the equivalent
of the Associate() and Learn() functions in the Appendix M program.
----------------------------------------------------------------------*/
void Memorize(Vector F, Vector G, Vector O) {
    int i,j;	/* loop indices */
    float Gterm;
    float LP;	/* learning parameter */

    /* Guess at learning parameter, LP.   Increase for speed.  */
    /* If LP is too large, the system can become unstable. */
    LP=Vector_length(Number_of_neurons,F);
    LP=0.5/LP/LP;

    for(i=0; i<Number_of_neurons; i++) {
        /* If Widrow-Hoff correction technique, then the error is */
        /* learned, otherwise, the associated output vector is learned. */
        if(Correction)
            Gterm=G[i]-O[i];	/* Error term */
        else
            Gterm=G[i];		/* Association */

        for(j=0; j<Number_of_synapses; j++) 
            Neurons[i].Synapses[j].Strength += 
                LP * Gterm * F[Neurons[i].Synapses[j].From];
    }
}   /* Memorize */


/*--------------------------------------------------------------------------
Calculates the cosine of the 'angle' between two vectors.
Arguments
    float Vector_1[]            One vector, input
    float Vector_2[]            Other vector, input
--------------------------------------------------------------------------*/
void Write_cosine(float Vector_1[], float Vector_2[]) {
    float Length_of_vector_1, Length_of_vector_2;     /* vector magnitudes */
    float cos_12;

    Length_of_vector_1=Vector_length(Number_of_neurons,Vector_1);
    Length_of_vector_2=Vector_length(Number_of_neurons,Vector_2);

    if(Length_of_vector_1==0 || Length_of_vector_2==0) {
        printf("Write_cosine(): Input vector has zero length!\n");
    } else {
        cos_12=Inner_product(Number_of_neurons,Vector_1,Vector_2)/
                Length_of_vector_1/Length_of_vector_2;
    }
    printf("%9.4f",cos_12);
}   /* Write_cosine */



/*----------------------------------------------------------------------
Typical invocation:
    ./associat fdis.stm gdis.stm ndis.neu
of which fdis.stm and gdis.stm must be pre-existing files, and ndis.neu
is optionally used as input and is the output of this program.
----------------------------------------------------------------------*/
void main(int argc, char *argv[]) {
    Vector Outp;
    Vector Gout;
    int i,k;		/* loop indices */
    int How_often;

    if(argc>1) strncpy(FFileName,argv[1],sizeof(FFileName));
    else       strncpy(FFileName,"Ffile",sizeof(FFileName));
    if(argc>2) strncpy(GFileName,argv[2],sizeof(GFileName));
    else       strncpy(GFileName,"Gfile",sizeof(GFileName));
    if(argc>3) strncpy(NFileName,argv[3],sizeof(NFileName));
    else       strncpy(NFileName,"Nfile",sizeof(NFileName));

    Initialize();
    printf("\n");

    /* If there are a small number of stimuli, then each is presented */
    /* exactly once, in sequence. Effectively, this can be used to generate */
    /* an outer product matrix if the correction (Widrow-Hoff) procedure */
    /* is turned off. */
    if(Nr_to_learn <= Nr_of_stimuli)
        printf("Each stimulus is presented exactly once.\n");
    if(Nr_to_learn < Nr_of_stimuli) {
        printf("Some stimuli not learned.  ");
        printf("Number to learn = %d   ",Nr_to_learn);
        printf("Number of stimuli = %d\n",Nr_of_stimuli);
    }

    /* Does weight updating. */
    for(k=1; k<=Nr_to_learn; k++) {
        if(Nr_to_learn<=Nr_of_stimuli)
            i=k;				/* learn in sequence */
        else
            i=1+(int)(Random()*Nr_of_stimuli);	/* learn random pair */
        Calculate(F_set[i].Val, Gout);
        Memorize(F_set[i].Val, G_set[i].Val, Gout);

        /* Display every tenth pair, so progress can be monitored */
        How_often=10;
        if(k%How_often == 0) {
            printf("%5d  Nr: %4d   Cosine: ",k,i);
            Write_cosine(G_set[i].Val,Gout);
            printf("\n");
        }
    }

    /* Gives recall accuracy for entire stimulus set for checking purposes. */
    /* Computes cosine between stored and retrieved association and also */
    /* length of retrieved vector. */
    printf("\nAccuracy of recall of input set.\n\n");
    for(i=1; i<=Nr_of_stimuli; i++) {
        Calculate(F_set[i].Val, Outp);
        printf("\n%3d  Name: %s\n",i,G_set[i].Name);
        printf("   Cosine: ");
        Write_cosine(G_set[i].Val,Outp);
        printf("  Length: %.4f\n",Vector_length(Number_of_neurons,Outp));
    }

    /* Write connections to output file. */
    Write_Nfile();
}   /* main */

