/* File bsb_proc.c */

/*--------------------------------------------------------------------------
This is the procedure that does all the work in the BSB model.
--------------------------------------------------------------------------*/

#include <ctype.h>	/* toupper */
#include <math.h>	/* fabs */
#include <stdio.h>	/* fgets getchar printf scanf */

/* Number_of_characters is related to Dimensionality of vectors */
#define Number_of_characters (int)(25)
#define Dimensionality       (int)(8*Number_of_characters)
/* Maximum_set_size is the size of each training set */
#define Maximum_set_size     (int)(100)
/* String_length refers to strings in the struct Stimulus.Name */
#define String_length  (int)(60)       /* Must be > Number_of_characters */
#define FALSE 0
#define TRUE 1

typedef float Vector[Dimensionality];
typedef char String[String_length+1];
typedef struct {
    String Name;
    Vector Val;
} Stimulus;
typedef struct {
    float Strength;
    int From;
} Synapse;
typedef struct {
    float Activity;
    Synapse Synapses[Dimensionality];
    float Neuron_output;
    float Learning;
    float Neuron_upper_limit;
    float Neuron_lower_limit;
} Neuron;

/* Defined in bsb.c */
/* Element [0] does not appear to be used anywhere */
Stimulus F_set[Maximum_set_size+1];
Stimulus G_set[Maximum_set_size+1];
Stimulus T_set[Maximum_set_size+1];
Neuron Neurons[Dimensionality];
int Neuron_file_present;
float Threshold;
int FG_mode;

/* Defined in bsb_int.c */
void Interpret(float Threshold,   /* Determines interpretation threshold.*/
               Vector V,                /* Input state vector. */
               String Interpretation);  /* Output interpreted string. */

/* Defined in bsb_list.c */
void List_test_set(void);

/* Defined in bsb_make.c */
void Make_stimulus(void);

/* Defined in bsb_proc.c */
static char Test_stimulus_set;
static int Test_stimulus_nr;
static int Delta_passes;
static int Start_nr_of_passes; 
static int Finish_nr_of_passes; 
static int Current_check;
static float Upper_limit;
static float Lower_limit;
static float Feedback_constant;
static float Decay_constant;
static Vector Starting_vector; 
static Vector Inp;
static Vector Outp;
static int Unchanged_original;
static int Quit;
static int Restart;
static Stimulus Stim;
static void BSBvals_initialization(void);
static void Display_stimulus(void);
static void BSBvals(void);
static void Initial_BSB_parameters(void);
static void Write_BSB_parameters(void);
static void BSB_command(void);
static void Limit_vector(Vector V);
static void Calculate(Vector F, Vector O);
static int Check_vector(Vector V);
static void Constant_times_vector(float C, Vector V_in, Vector V_out);
static void Vector_plus_vector(Vector V1, Vector V2, Vector V_Sum);
static void W(String S);
static void One_pass(int i, Vector V_in, Vector V_out);
static void Restart_initialization(void);
static void Do_iterations(int Start, int Finish);
void BSBProcedure(void);

/* Defined in bsb_readnf.c */
int Number_of_synapses[Dimensionality];
int Avg_number_of_synapses;

/* Defined in bsb_vt100.c */
void Clear_box(void);
void Clear_block(void);
void Erase_line(void);
void Scroll_block(void);
void Save_cursor(void);
void Restore_cursor(void);
void Bottom_line(void);
void Restore_scrolling(void);

/*--------------------------------------------------------------------------
--------------------------------------------------------------------------*/
static void BSBvals_initialization(void) {
    int i;	/* loop index */

    if(Test_stimulus_set=='F' || Test_stimulus_set=='f')
        Stim=F_set[Test_stimulus_nr];
    else if(Test_stimulus_set=='G' || Test_stimulus_set=='g')
        Stim=G_set[Test_stimulus_nr];
    else if(Test_stimulus_set=='T' || Test_stimulus_set=='t')
        Stim=T_set[Test_stimulus_nr];
    else
        printf("\nBSBval_initialization(): Illegal value of Stimulus_set\n");
    if(!Restart)
        for(i=0; i<Dimensionality; i++) Stim.Val[i]=Outp[i];
}   /* BSBvals_initialization */

/*--------------------------------------------------------------------------
--------------------------------------------------------------------------*/
static void Display_stimulus(void) {
    int i;			/* loop index */
    int Elements_per_unit=8;	/* Number of 3-byte elements per report unit */
    int Elements_per_line=24;	/* Number of 3-byte elements per line */
    int Unit_nr;		/* Unit corresponding to element */
    int Truncated_value;	/* One char version of element */
    String Stim_interpretation;	/* Vector interpretated as a string */
    char Sign_char;		/* Used to represent element */
    char Value_char;		/* Used to respresent element */

    Interpret(Threshold,Stim.Val,Stim_interpretation);
    Clear_block();
    Scroll_block();
    printf("            Name: %s\n", Stim.Name);
    printf("  Interpretation: %s\n\n", Stim_interpretation);

    for(i=0; i<Dimensionality; i++) {
        if((i%Elements_per_unit)==0) {
            Unit_nr=i/Elements_per_unit;
            printf("%6d",Unit_nr+1);
            printf(" %c ",Stim_interpretation[Unit_nr]);
        }
        if(Stim.Val[i] > 0) {
            Sign_char='+';
            Truncated_value=(int)(10*Stim.Val[i]/Upper_limit);
            Value_char=(Truncated_value>=10)? 'L': (char)('0'+Truncated_value);
        } else if(Stim.Val[i] < 0) {
            Sign_char='-';
            Truncated_value=(int)(fabs(10*Stim.Val[i]/Lower_limit));
            Value_char=(Truncated_value>=10)? 'L': (char)('0'+Truncated_value);
        } else {
	    Sign_char='-';
            Value_char='0';
        }

        printf("%c%c", Sign_char, Value_char);
        if(i%Elements_per_line==Elements_per_line-1) printf("%4d\n",i+1);
    }
}   /* Display_stimulus */



/*--------------------------------------------------------------------------
--------------------------------------------------------------------------*/
static void BSBvals(void) {
    BSBvals_initialization();
    Display_stimulus();
}   /* BSBvals */


/*--------------------------------------------------------------------------
The variable `delta' is discussed in Limit_vector(), but in effect setting
    Unchanged_original= TRUE;
implies delta=1, while the opposite implies delta=0.
--------------------------------------------------------------------------*/
static void Initial_BSB_parameters(void) {
    Unchanged_original= TRUE;
    FG_mode           = FALSE;
    Test_stimulus_set = 'T';
    Test_stimulus_nr  = 1;
    Delta_passes      = 16;
    Threshold         = 0.5;
    Restart           = TRUE;
    Quit              = FALSE;

    /* BSB Computational Parameters. */
    Decay_constant    = 0.90;    /* gamma */
    Feedback_constant = 0.20;    /* alpha */
    Lower_limit       = -1.3;
    Upper_limit       =  1.3;
}   /* Initial_BSB_parameters */

/*--------------------------------------------------------------------------
This routine clears the screen and prints three lines of information 
listing the program settings.
--------------------------------------------------------------------------*/
static void Write_BSB_parameters(void) {
    Clear_box();

    /* First line */
    printf(
    "BSB>P)asses :%4d     U)Limit:%8.3f  T)hreshold:%8.3f   Stim. #) : %c%d\n",
        Delta_passes,Upper_limit,Threshold,Test_stimulus_set,Test_stimulus_nr);

    /* Second line */
    if(Neuron_file_present)
        printf("Mx: Synapses:%4d   F)eedback:%8.3f      D)ecay:%8.3f\n",
            Avg_number_of_synapses,Feedback_constant,Decay_constant);

    /* Third line */
    if(Restart)
        printf("RESTART");
    else
        printf("It:%d", Finish_nr_of_passes);
    if(Unchanged_original) printf("  Add original stimulus.");
    printf("\n"); 
}   /* Write_BSB_parameters */


/*--------------------------------------------------------------------------
--------------------------------------------------------------------------*/
static void BSB_command(void) {
    String BSB_string;
    int i;
    int Set_parameters;
    char ch;
    int Test_stimulus_set_OK;        /* Set only if == 'f', 'g' or 't' */
    int Test_stimulus_nr_OK;         /* Set only if == 1..Maximum_set_size */

    Save_cursor();
    Set_parameters=FALSE;

    /* Get and process commands */
    do {
        Write_BSB_parameters();
        Bottom_line();
        printf("%s", FG_mode? "FG ": "TF ");
        printf("BSB X)ecute C)hange L)ist R)estart V)als Q)uit > ");
        fgets(BSB_string,sizeof(BSB_string),stdin);
        Bottom_line();
        Erase_line();

        /* Process commands */
        for(i=0; BSB_string[i]!='\0'; i++) {
            ch=(char)toupper(BSB_string[i]);
            if(ch=='C') {
                Make_stimulus();
            } else if(ch=='D') {
                printf("Trace Decay: ");
                scanf("%f",&Decay_constant);
                getchar();
            } else if(ch=='F') {
                printf("Matrix Feedback: ");
                scanf("%f",&Feedback_constant);
                getchar();
            } else if(ch=='L') {
                List_test_set();
            } else if(ch=='M') {
                FG_mode=!FG_mode;
            } else if(ch=='O') {
                Unchanged_original=!Unchanged_original;
                if(Unchanged_original)
                    printf("Will add original stimulus to output.");
                else
                    printf("Input stimulus will be treated as data.");
            } else if(ch=='P') {
                printf("Number of passes: ");
                scanf("%d",&Delta_passes);
                getchar();
            } else if(ch=='Q') {
                Quit=TRUE;
            } else if(ch=='R') {
                Restart=TRUE;
            } else if(ch=='T') {
                printf("New threshold: ");
                scanf("%f",&Threshold);
                getchar();
            } else if(ch=='U') {
                printf("Upper Limit: ");
                scanf("%f",&Upper_limit);
                getchar();
                printf("Lower Limit: ");
                scanf("%f",&Lower_limit);
                getchar();
            } else if(ch=='V') {
                BSBvals();
            } else if(ch=='X') {
                printf("Xecute Iterations.\n");
                Set_parameters=TRUE;
            } else if(ch=='#') {
                do {
                    printf("Choose stimulus (Set, number) : ");
                    scanf("%c %d",&Test_stimulus_set,&Test_stimulus_nr);
                    getchar();
                    Test_stimulus_set=(char)toupper(Test_stimulus_set);
                    Test_stimulus_set_OK=(
                         Test_stimulus_set=='F' || Test_stimulus_set=='f' ||
                         Test_stimulus_set=='G' || Test_stimulus_set=='g' ||
                         Test_stimulus_set=='T' || Test_stimulus_set=='t');
                    Test_stimulus_nr_OK=(Test_stimulus_nr>=1 &&
                            Test_stimulus_nr<=Maximum_set_size);
                } while(!Test_stimulus_set_OK && !Test_stimulus_nr_OK);

            }
        }
    } while(!Set_parameters && !Quit);
    Restore_cursor();
}   /* BSB_command */

    
/*--------------------------------------------------------------------------
This performs two operations.  First (but only if Unchanged_original 
true) it adds the input vector to alpha A x + gamma x, where x is the 
current state vector.  The input vector is unscaled before being added.
Sometimes (as in BSB_Iteration() in appm.c) it is multiplied by a factor
delta.  In the present case delta is implicitly 1.0.

Next it truncates each element of the sum to the range Lower_limit ..
Upper_limit.

This routine impliments the equation at the bottom of page 504.
--------------------------------------------------------------------------*/
static void Limit_vector(Vector V) {
    int i;		/* loop index */

    /* The data input is added to the output of the iteration. */
    if(Unchanged_original) 
        for(i=0; i<Dimensionality; i++) 
            V[i] += Starting_vector[i];

    /* Clipping */
    for(i=0; i<Dimensionality; i++) {
        if(V[i]>Upper_limit) V[i]=Upper_limit;
        if(V[i]<Lower_limit) V[i]=Lower_limit;
    }
}   /* Limit_vector */


/*--------------------------------------------------------------------------
Calculate does most of the computational work.  This procedure
should be in very fast vector hardware or assembly language.
--------------------------------------------------------------------------*/
static void Calculate(Vector F, Vector O) {
    int i,j;		/* loop indices */
    float Sum;		/* total neuronal input */

    for(i=0; i<Dimensionality; i++) {
        Sum=0;
        for(j=0; j<Number_of_synapses[i]; j++)
            Sum += Neurons[i].Synapses[j].Strength
                  *F[Neurons[i].Synapses[j].From];
        O[i]=Sum;
    }
}   /* Calculate */



/*--------------------------------------------------------------------------
Useful for seeing how far a vector is from a corner.
--------------------------------------------------------------------------*/
static int Check_vector(Vector V) {
    int i;		/* loop index */
    int Count_so_far;	/* number of limited elements */

    Count_so_far=0;
    for(i=0; i<Dimensionality; i++) {
        if(V[i]>=Upper_limit) Count_so_far++;
        if(V[i]<=Lower_limit) Count_so_far++;
    }
    return Count_so_far;
}   /* Check_vector */


/*--------------------------------------------------------------------------
--------------------------------------------------------------------------*/
static void Constant_times_vector(float C, Vector V_in, Vector V_out) {
    int i;		/* loop index */

    for(i=0; i<Dimensionality; i++)
        V_out[i]=C*V_in[i];
}   /* Constant_times_vector */


/*--------------------------------------------------------------------------
--------------------------------------------------------------------------*/
static void Vector_plus_vector(Vector V1, Vector V2, Vector V_Sum) {
    int i;		/* loop index */

    for(i=0; i<Dimensionality; i++)
        V_Sum[i]=V1[i]+V2[i];
}   /* Vector_plus_vector */


/*--------------------------------------------------------------------------
--------------------------------------------------------------------------*/
static void W(String S) {
    int i;	/* loop index */

    for(i=0; i<Number_of_characters; i++)
        printf("%c",S[i]);
}    /* W */


/*--------------------------------------------------------------------------
One iterative pass through the matrix.
--------------------------------------------------------------------------*/
static void One_pass(int i, Vector V_in, Vector V_out) {
    Vector Ax;
    Vector DecayIx;
    String Interpreted_vector;

    Calculate(V_in,Ax);

    /* Result is Ax=alpha*A*x */
    Constant_times_vector(Feedback_constant,Ax,Ax);

    /* Previous states decays: DecayIx=gamma*x */
    Constant_times_vector(Decay_constant,V_in,DecayIx);

    /* Output is sum of these two terms */
    Vector_plus_vector(Ax, DecayIx, V_out);

    /* Limit element values.  Also will add original state if called for */
    Limit_vector(V_out);
    Current_check=Check_vector(V_out);

    /* Write output to file or screen */
    Interpret(Threshold,V_out,Interpreted_vector);
    printf("%6d.  ",i);
    W(Interpreted_vector);
    printf("  Check: %3d\n",Current_check);
}   /* One_pass */


/*--------------------------------------------------------------------------
--------------------------------------------------------------------------*/
static void Restart_initialization(void) {
    int i;	/* loop index */

    Clear_block();
    Scroll_block();
    Start_nr_of_passes=0;
    if(Test_stimulus_set=='F' || Test_stimulus_set=='f') 
        for(i=0; i<Dimensionality; i++)
            Starting_vector[i]=F_set[Test_stimulus_nr].Val[i];
    else if(Test_stimulus_set=='G' || Test_stimulus_set=='g')
        for(i=0; i<Dimensionality; i++)
            Starting_vector[i]=G_set[Test_stimulus_nr].Val[i];
    else if(Test_stimulus_set=='T' || Test_stimulus_set=='t')
        for(i=0; i<Dimensionality; i++)
            Starting_vector[i]=T_set[Test_stimulus_nr].Val[i];
    else
        printf("\nRestart_initialization(): Illegal value of Stimulus_set\n");
    for(i=0; i<Dimensionality; i++)
        Inp[i]=Starting_vector[i];
    Limit_vector(Inp);
    Restart=FALSE;
    Clear_block();
    Scroll_block();
}   /* Restart_initialization */


/*--------------------------------------------------------------------------
--------------------------------------------------------------------------*/
static void Do_iterations(int Start, int Finish) {
    int n;	/* iteration number */
    int i;	/* loop index */

    n=Start;
    do {
        One_pass(n,Inp,Outp);
        n++;
        for(i=0; i<Dimensionality; i++) Inp[i]=Outp[i];
    } while(Current_check<Dimensionality && n<=Finish);
}   /* Do_iterations */


/*--------------------------------------------------------------------------
Brain state in a box iterations
--------------------------------------------------------------------------*/
void BSBProcedure(void) {
    Initial_BSB_parameters();
    do {
        BSB_command();
        if(Restart) Restart_initialization();
        Finish_nr_of_passes=Start_nr_of_passes+Delta_passes;
        if(!Quit) {
            Do_iterations(Start_nr_of_passes+1,Finish_nr_of_passes);
            Start_nr_of_passes=Finish_nr_of_passes;
            if(Current_check==Dimensionality)
                printf(">> Fully limited.  Finished.\n");
        }
    } while(!Quit); 
    Restore_scrolling();
    printf("\n");
}   /* BSBProcedure */

