/* File appc.c */

/*
                             Appendix C
          Procedures, Functions, and Types from Chapter 5

    Pascal versions are preserved as comments, and additional routines
    added to make a runable demonstration.  Note that for the purposes
    of the demonstration, the constant Dimensionality has been drastically
    from 100 to 2.
*/

#include <stdio.h>      /* getchar printf */

/*-------------------------------------------------------------------------
     TYPE Matrix

        CONST Dimensionality = 100
        TYPE Matrix = ARRAY [1..Dimensionality, 1..Dimensionality] OF REAL;
-------------------------------------------------------------------------*/
#define Dimensionality (int)(2)
typedef float Vector[Dimensionality];
typedef float Matrix[Dimensionality][Dimensionality];

void main(void);
void Print_vector(char *name, Vector Vec);
void Print_matrix(char *name, Matrix Mx);

void Add_matrices(Matrix Mx_1, Matrix Mx_2, Matrix Matrix_sum);
void Subtract_matrices(Matrix Mx_1, Matrix Mx_2, Matrix Matrix_difference);
void Constant_times_matrix(float c, Matrix Mx_in, Matrix Mx_out);
void Matrix_times_vector(Matrix A, Vector F, Vector G);
void Outer_product(Vector G, Vector F, Matrix A);

/*-------------------------------------------------------------------------
Test program
-------------------------------------------------------------------------*/
void main(void) {
    float scalar;
    Vector Vec_1, Vec_2, Vec_out;
    Matrix Mx_1,Mx_2,Matrix_out;

    /* Initialize scalar and arrays */
    scalar=(float)0.1;
    Vec_1[0]=2;   Vec_1[1]=3;
    Vec_2[0]=-1;  Vec_2[1]=1;
    Mx_1[0][0]=0; Mx_1[0][1]=1;
    Mx_1[1][0]=2; Mx_1[1][1]=3;
    Mx_2[0][0]=5; Mx_2[0][1]=4;
    Mx_2[1][0]=2; Mx_2[1][1]=0;

    /* Display input matrices */
    Print_matrix("Mx_1",Mx_1);
    Print_matrix("Mx_2",Mx_2);

    Add_matrices(Mx_1, Mx_2, Matrix_out);
    Print_matrix("Sum:   Mx_1 + Mx_2",Matrix_out);

    Subtract_matrices(Mx_1, Mx_2, Matrix_out);
    Print_matrix("Difference:  Mx_1 - Mx_2",Matrix_out);

    Constant_times_matrix(scalar, Mx_1, Matrix_out);
    Print_matrix("Scalar product:  0.1 * Mx_1",Matrix_out);

    printf("Press <Ret> to continue: "); getchar();

    /* Display input matrices and vectors */
    Print_matrix("Mx_1",Mx_1);
    Print_vector("Vec_1", Vec_1);
    Print_vector("Vec_2", Vec_2);

    Matrix_times_vector(Mx_1, Vec_1, Vec_out);
    Print_vector("Inner product: Mx_1 . Vec_1",Vec_out);

    Outer_product(Vec_1, Vec_2, Matrix_out);
    Print_matrix("Outer product: Vec_1 * Vec_2^T",Matrix_out);
}   /* main */


/*-------------------------------------------------------------------------
Prints a vector as a column of values, as is the convention for vectors.
-------------------------------------------------------------------------*/
void Print_vector(char *name, Vector Vec) {
    int i;      /* loop index */

    printf("\n%s\n",name);
    for(i=0; i<Dimensionality; i++) {
	printf(" %8.2f\n",Vec[i]);
    }
}   /* Print_vector */


/*-------------------------------------------------------------------------
Prints a matrix as a square array of values.
-------------------------------------------------------------------------*/
void Print_matrix(char *name, Matrix Mx) {
    int i,j;    /* loop indices */

    printf("\n%s\n",name);
    for(i=0; i<Dimensionality; i++) {
	for(j=0; j<Dimensionality; j++)
	    printf(" %8.2f",Mx[i][j]);
	printf("\n");
    }
}   /* Print_matrix */


/*-------------------------------------------------------------------------
     PROCEDURE Add_matrices (Mx_1, Mx_2: Matrix; VAR Matrix_sum: Matrix);
        VAR I,J: INTEGER;
        BEGIN
        FOR I:= 1 TO Dimensionality DO
            FOR J:= 1 TO Dimensionality DO
		Matrix_sum [I,J] := Mx_1 [I,J] + Mx_2 [I,J];
        END; 
-------------------------------------------------------------------------*/
void Add_matrices(Matrix Mx_1, Matrix Mx_2, Matrix Matrix_sum) {
    int i,j;    /* loop indices */

    for(i=0; i<Dimensionality; i++)
	for(j=0; j<Dimensionality; j++)
	    Matrix_sum[i][j]=Mx_1[i][j]+Mx_2[i][j];
}   /* Add_matrices */

/*-------------------------------------------------------------------------
     PROCEDURE Subtract_matrices (Mx_1, Mx_2: Matrix; 
                       VAR Matrix_difference: Matrix);
        VAR I,J: INTEGER;
        BEGIN
        FOR I:= 1 TO Dimensionality DO
            FOR J:= 1 TO Dimensionality DO
                Matrix_difference [I,J] := Mx_1 [I,J] - Mx_2 [I,J];
        END; 
-------------------------------------------------------------------------*/
void Subtract_matrices(Matrix Mx_1, Matrix Mx_2, Matrix Matrix_difference) {
    int i,j;    /* loop indices */

    for(i=0; i<Dimensionality; i++)
	for(j=0; j<Dimensionality; j++)
	    Matrix_difference[i][j]=Mx_1[i][j]-Mx_2[i][j];
}   /* Subtract_matrices */


/*-------------------------------------------------------------------------
     PROCEDURE Constant_times_matrix   (C: REAL; Mx_in: Matrix; 
                                            VAR Mx_out: Matrix);
        VAR I,J: INTEGER;
        BEGIN
        FOR I:= 1 TO Dimensionality DO
            FOR J:= 1 TO Dimensionality DO
                Mx_out [I,J] := C * Mx_in [I,J];
        END; 
-------------------------------------------------------------------------*/
void Constant_times_matrix(float c, Matrix Mx_in, Matrix Mx_out) {
    int i,j;    /* loop indices */

    for(i=0; i<Dimensionality; i++)
	for(j=0; j<Dimensionality; j++)
	    Mx_out[i][j]=c*Mx_in[i][j];
}   /* Constant_times_matrix */


/*-------------------------------------------------------------------------
     PROCEDURE Matrix_times_vector (A: Matrix; F: Vector;
                                           VAR G: Vector);

          {This PROCEDURE forms the product Af = g where }
          {both f and g are column vectors.              }

          VAR I,J: INTEGER;
          BEGIN
          FOR I:= 1 TO Dimensionality DO G[I]:= 0;   {Zero vector G.}
          FOR I:= 1 TO Dimensionality DO
              FOR J:= 1 TO Dimensionality DO G[I]:= G[I] + A[I,J] * F[J];
          END;
-------------------------------------------------------------------------*/
void Matrix_times_vector(Matrix A, Vector F, Vector G) {
    int i,j;    /* loop indices */

    /* Zero vector G */
    for(i=0; i<Dimensionality; i++)
	G[i]=0;
    for(i=0; i<Dimensionality; i++)
	for(j=0; j<Dimensionality; j++)
	    G[i] += A[i][j]*F[j];
}   /* Matrix_times_vector */


/*-------------------------------------------------------------------------
     PROCEDURE Outer_product (G,F: Vector; VAR A: Matrix);
         {                                                       }
         {We want to generate an outer product matrix A so that  }
         {            T                                          }
         {     A = g f                                           }
         {                                                       }
         VAR I,J: INTEGER;
         BEGIN
         FOR I:= 1 TO Dimensionality DO
             FOR J:= 1 TO Dimensionality DO 
                 A[I,J] := G[I] * F[J];
         END;
-------------------------------------------------------------------------*/
void Outer_product(Vector G, Vector F, Matrix A) {
    int i,j;    /* loop indices */

    for(i=0; i<Dimensionality; i++)
	for(j=0; j<Dimensionality; j++)
	    A[i][j] = G[i]*F[j];
}   /* Outer_product */



/*-------------------------------------------------------------------------
Example of output:

Mx_1
     0.00     1.00
     2.00     3.00

Mx_2
     5.00     4.00
     2.00     0.00

Sum:   Mx_1 + Mx_2
     5.00     5.00
     4.00     3.00

Difference:  Mx_1 - Mx_2
    -5.00    -3.00
     0.00     3.00

Scalar product:  0.1 * Mx_1
     0.00     0.10
     0.20     0.30

Mx_1
     0.00     1.00
     2.00     3.00

Vec_1
     2.00
     3.00

Vec_2
    -1.00
     1.00

Inner product: Mx_1 . Vec_1
     3.00
    13.00

Outer product: Vec_1 * Vec_2^T
    -2.00     2.00
    -3.00     3.00

-------------------------------------------------------------------------*/
