/* File appe.c, Program SVDEMO */

/*
                               Appendix E
                    Summed Vector Memory (Chapter 7)

     This program gives a demonstration of a simple summed vector 
recognition memory.  The memory (variable Memory) is a vector which
is the sum of the learned vectors.  In the program, the vectors 
have maximum dimensionality of 2,000 and the distribution of elements 
has mean zero and an average length of one.  The learned vectors are 
generated by the Unix library random number generator rand.

     When the program is run, the program will prompt for the seed 
for the random number generator, the dimensionality of the system 
(2,000 maximum) and the number of random vectors to learn (200 maximum).  
 
     The output of the system uses character based graphics to present 
the histogram of system output values when the vectors the system learned 
are tested, and when 100 new zero mean random vectors are tested.  New 
random vectors should give output values about zero; the learned vectors 
should give outputs around one.  The y-axis (vertical) gives the inner 
product value times 10; the x-axis (horizontal) gives the number of 
output values in the bin.  The program can be run again if wanted, by 
typing `y' to the prompt.

See p 194 of An Introduction to Neural Networks.
*/

#if defined MSDOS
#include <graph.h>      /* _settextposition _clearscreen */
#endif
#include <math.h>       /* sqrt */
#include <stdio.h>      /* fflush getchar printf */
#include <stdlib.h>     /* rand srand RAND_MAX */

/* Max_dimensionality=2000, Number_in_test_set=100, Number_to_learn=200 */
#define Max_dimensionality (int)(2000)
#define Number_in_test_set (int)(100)
#define Number_to_learn    (int)(200)
#define Max_histogram      (int)(20)
#define FALSE 0
#define TRUE  1
typedef float Vector[Max_dimensionality];

void ClrScr(void);
void GoToXY(int x, int y);
void Add_vectors(float Vector_1[], float Vector_2[], float Vector_sum[]);
float Inner_product(float Vector_1[], float Vector_2[]);
float Vector_length(float Vector_1[]);
void Make_random_vector(float Vect[]);
void Make_test_vectors(void);
void Make_learned_vectors(void);
void Display_outputs(void);
void Initialize(void);
void Demonstration(void);
void main(void);

int Exit;
int Dimensionality;
int Number_of_vectors;
int TS_histogram[2*Max_histogram+1];
int LS_histogram[2*Max_histogram+1];
Vector Memory;                  /* Sum of all F_vectors */
Vector Test_vectors[Number_in_test_set];
Vector F_vectors[Number_to_learn];


/*--------------------------------------------------------------------------
Screen Control Procedure: Clear screen and move cursor to 1,1
--------------------------------------------------------------------------*/
void ClrScr(void) {
#if defined MSDOS
    _clearscreen(_GCLEARSCREEN);
#else
    printf("%c[2J",0x1b);
    printf("%c[%d;%dH",0x1b,1,1);
#endif
}   /* ClrScr */


/*--------------------------------------------------------------------------
Screen Control Procedure: Position cursor

Arguments
    x    column as integer 1..80
    y    row as integer 1..24
--------------------------------------------------------------------------*/
void GoToXY(int x, int y) {
    int x1,y1;          /* coordinates restricted to allowed range */

    x1=(x<1)? 1: (x>80)? 80: x;         /* 1<=x1<=80 */
    y1=(y<1)? 1: (y>24)? 24: y;         /* 1<=y1<=24 */
#if defined MSDOS
    _settextposition(y,x);
#else
    printf("%c[%d;%dH",0x1b,y1,x1);
#endif
}   /* GoToXY */



/*--------------------------------------------------------------------------
Adds two vectors.
Arguments
    float Vector_1[]            One vector, input
    float Vector_2[]            Other vector, input
    float Vector_sum[]          Vector_1+Vector_2, returned
--------------------------------------------------------------------------*/
void Add_vectors(float Vector_1[], float Vector_2[], float Vector_sum[]) {
    int i;      /* loop index */

    for(i=0; i<Dimensionality; i++)
	Vector_sum[i]=Vector_1[i] + Vector_2[i];
}   /* Add_vectors */


/*--------------------------------------------------------------------------
Forms the inner product of two vectors.
Arguments
    float Vector_1[]            One vector, input
    float Vector_2[]            Other vector, input
Returns
    float Inner_product         Vector_1 . Vector_2
--------------------------------------------------------------------------*/
float Inner_product(float Vector_1[], float Vector_2[]) {
    float Sum_of_products;      /* Name says it all */
    int i;                      /* loop index */

    Sum_of_products=0;
    for(i=0; i<Dimensionality; i++)
	Sum_of_products += Vector_1[i]*Vector_2[i];
    return Sum_of_products;
}   /* Inner_product */



/*--------------------------------------------------------------------------
Returns the length of a vector.
Argument
    float Vector_1[]            Vector, input
Returns
    float length                Magnitude of Vector_1
--------------------------------------------------------------------------*/
float Vector_length(float Vector_1[]) {
    float Sum_of_Squares;       /* Name says it all */
    int i;                      /* loop index */

    Sum_of_Squares=0;
    for(i=0; i<Dimensionality; i++)
	Sum_of_Squares += Vector_1[i]*Vector_1[i];
    return (float)sqrt(Sum_of_Squares);
}   /* Vector_length */


/*-----------------------------------------------------------------------
This procedure generates a random vector.  The vector is randomly directed
and has random length.  The average length is one.  Unless srand() is
called, the same sequence random numbers will be called each time this
program is run.
-----------------------------------------------------------------------*/
void Make_random_vector(float Vect[]) {
    float Sqrt_dim;             /* */
    int i;                      /* loop index */

    Sqrt_dim=(float)sqrt(Dimensionality/12.0);
    for(i=0; i<Dimensionality; i++)
	Vect[i]=((float)rand()/RAND_MAX - 0.5f)/Sqrt_dim;

}   /* Make_random_vector */


/*--------------------------------------------------------------------------
Reports progress in building a test set.
--------------------------------------------------------------------------*/
void Make_test_vectors(void) {
    int i;                      /* loop index */

    /* Write test vector title on line 22 */
    GoToXY(1,22);
    printf("Making test vectors   :         ");

    for(i=0; i<Number_in_test_set; i++) {
	Make_random_vector(Test_vectors[i]);
	/* Report progress */
	GoToXY(24,22);
	printf("%4d Length: %10.5f",i,Vector_length(Test_vectors[i]));
    }
}   /* Make_test_vectors */


/*--------------------------------------------------------------------------
Reports progress in building a learned set.
--------------------------------------------------------------------------*/
void Make_learned_vectors(void) {
    int i;                      /* loop index */

    /* Write learned vector title on line 23 */
    GoToXY(1,23);
    printf("Making learned vectors:         ");

    for(i=0; i<Number_of_vectors; i++) {
	Make_random_vector(F_vectors[i]);
	/* Report progress */
	GoToXY(24,23);
	printf("%4d Length: %10.5f",i,Vector_length(F_vectors[i]));
    }
}   /* Make_learned_vectors */


/*--------------------------------------------------------------------------
A histogram of the lengths of the of the test and learned vectors.
The histogram is binned in units of 0.1, but is labelled with integers.

The distribution of learned vectors is plotted first, and should be 
centred on 1.0, or the bin labelled `10'.

The distribution of test vectors is superimposed, and should be 
centred on 0.
--------------------------------------------------------------------------*/
void Display_outputs(void) {
    int i;      /* loop index: range of histogram values to plot */
    int j;      /* loop index: used to plot histogram bar */
    int i0;     /* mid-point of histograms */

    i0=Max_histogram;   /* offset to mid_point of histograms */
    ClrScr();
    GoToXY(1,1);
    printf("Inner product of learned Vectors (*) and  test vectors (+).");

    /* Histogram corresponding to learned vectors */
    GoToXY(1,2);
    for(i=-5; i<=14; i++) {
	if(i<-Max_histogram || i>Max_histogram) continue;
	printf("%3d ",i);
	for(j=1; j<=LS_histogram[i0+i]; j++) printf("*");
	printf("\n");
    }

    /* Histogram corresponding to test vectors */
    GoToXY(1,2);
    for(i=-5; i<=14; i++) {
	if(i<-Max_histogram || i>Max_histogram) continue;
	printf("%3d ",i);
	for(j=1; j<=TS_histogram[i0+i]; j++) printf("+");
	printf("\n");
    }
}   /* Display_outputs */


/*--------------------------------------------------------------------------
Here the random number generator is seeded, and values assigned to
Dimensionality and Number_of_vectors.
--------------------------------------------------------------------------*/
void Initialize(void) {
    int i;              /* loop index */
    unsigned int Seed;  /* seed for srand() */

    ClrScr();
    printf("\n\nDemonstration of Simple Summed Vector Memory.\n");
    printf("Uses random vectors which are simply summed.\n");
    printf("Recall is tested using inner products of stored and\n");
    printf("random vectors.  See p194.\n");
    printf("Maximum Dimensionality is: %d\n",Max_dimensionality);

    printf("Seed for random number generator (0..65535): ");
    scanf("%u",&Seed); srand(Seed);

    do {
	printf("Dimensionality (1..%4d)                   : ",
	    Max_dimensionality);
	scanf("%d",&Dimensionality);
    } while(Dimensionality<1 || Dimensionality>Max_dimensionality);

    do {
	printf("Number of vectors to learn (1..%d): ",Number_to_learn);
	scanf("%d",&Number_of_vectors);
    } while(Number_of_vectors<1 && Number_of_vectors>Number_to_learn);

    /* Initialize histograms */
    for(i=-Max_histogram; i<=Max_histogram; i++) {
	LS_histogram[Max_histogram+i]=0;
	TS_histogram[Max_histogram+i]=0;
    }

    /* Initialize vectors */
    for(i=0; i<Dimensionality; i++) Memory[i]=0;
    Make_test_vectors();
    Make_learned_vectors();
}   /* Initialize */


/*--------------------------------------------------------------------------
--------------------------------------------------------------------------*/
void Demonstration(void) {
    int i;              /* loop index */
    int Index;          /* index to histograms */
    float Goodness_of_recall;   /* measure of success */
    int Input_char;     /* keyboard input */

    Initialize();

    GoToXY(1,1);
    printf(">> Learning.                ");
    GoToXY(1,12); for(i=0; i<80; i++) printf(" ");      /* clear line */
    for(i=0; i<Number_of_vectors; i++) {
	GoToXY(1,12);
	Add_vectors(Memory, F_vectors[i], Memory);
	Goodness_of_recall=Inner_product(Memory, F_vectors[i]);
	printf("Vectors learned : %4d  Recall of last vector: %12.6f",
		 i,Goodness_of_recall);
    }

    GoToXY(1,1);
    printf(">> Testing learned vectors after complete set is learned.");
    GoToXY(1,12); for(i=0; i<80; i++) printf(" ");      /* clear line */
    for(i=0; i<Number_of_vectors; i++) {
	GoToXY(1,12);
	Goodness_of_recall=Inner_product(Memory, F_vectors[i]);
	printf("Vectors tested  : %4d  Inner product        : %12.6f",
		 i,Goodness_of_recall);
	Goodness_of_recall*=10;
	/* Round Goodness_of_recall to the nearest integer */
	Goodness_of_recall += (Goodness_of_recall>0)? 0.5: -0.5;
	Index=(int)Goodness_of_recall;
	if(Index>=-Max_histogram && Index<=Max_histogram)
	    LS_histogram[Max_histogram+Index]++;
    }

    GoToXY(1,1);
    printf(">> Using test vectors after complete set is learned.       ");
    GoToXY(1,12); for(i=0; i<80; i++) printf(" ");      /* clear line */
    for(i=0; i<Number_in_test_set; i++) {
	GoToXY(1,12);
	Goodness_of_recall=Inner_product(Memory, Test_vectors[i]);
	printf("Vectors tested  : %4d  Inner product        : %12.6f",
		 i,Goodness_of_recall);
	Goodness_of_recall*=10;
	/* Round Goodness_of_recall to the nearest integer */
	Goodness_of_recall += (Goodness_of_recall>0)? 0.5: -0.5;
	Index=(int)Goodness_of_recall;
	if(Index>=-Max_histogram && Index<=Max_histogram)
	    TS_histogram[Max_histogram+Index]++;
    }

    Display_outputs();

    GoToXY(1,24);
    fflush(stdin);
    printf(">> Do you want to repeat this demonstration? Y/N : ");
    do {
        Input_char=getchar();
    } while(Input_char!='y' && Input_char!='Y' &&
            Input_char!='n' && Input_char!='N');
    if(Input_char=='N' || Input_char=='n')
	Exit=TRUE;
    else
	Exit=FALSE;
}   /* Demonstration */


/*--------------------------------------------------------------------------
Program SVDEMO
See p 194.
--------------------------------------------------------------------------*/
void main(void) {
    do {
	Demonstration();
    } while(!Exit);
}   /* main */


/*--------------------------------------------------------------------------
                    Example of output

Demonstration of Simple Summed Vector Memory.
Uses random vectors which are simply summed.
Recall is tested using inner products of stored and
random vectors.  See p194.
Maximum Dimensionality is: 2000
Seed for random number generator (0..65535): 123
Dimensionality (1..2000)                   : 1000
Number of vectors to learn (1..200): 200

Inner product of learned Vectors (*) and  test vectors (+).
 -5 +++++++
 -4 +++++++
 -3 ++++
 -2 ++++
 -1 +++++++++
  0 ++++++++++++
  1 +++++++
  2 ++++++
  3 +++++++++++++
  4 +++++++
  5 +++++++***
  6 ++++******
  7 ++*************
  8 +++**************
  9 +++*****************
 10 +**********
 11 **********************
 12 ********************
 13 ******************
 14 *****


>> Do you want to repeat this demonstration? Y/N : 


--------------------------------------------------------------------------*/
