/* File apph.c, Program AMCHARDEMO */ 

/*
                           Appendix H
      Utility Programs for Character Based Simulations (Chapter 7)


     This Appendix gives a number of lines of boring code that turn 
character strings into state vectors and vice versa.  These routines 
are used in program BSB, which is also included with these Appendices.

     Two routines, Make_stimulus(), and Interpret(), 
are used to make state vectors from character strings and to interpret 
state vectors back into character strings.  The routines require the 
the maximum string length to be specified.  Here we use
    #define Number_of_characters 25
The size of numeric arrays having type Vector or IVector is then 8x25=200.
*/

#include <math.h>       /* fabs */
#include <stdio.h>      /* printf */

#define Number_of_characters (int)(25)
#define Dimensionality       (int)(8*Number_of_characters)

typedef float Vector[Dimensionality];
typedef char String[Number_of_characters+1];
typedef int Byte[8];
typedef int IVector[Dimensionality];

/*------------------------------------------------------------------------
        The two PROCEDUREs have straightforward calls
------------------------------------------------------------------------*/
void Interpret(float Threshold,   /* Determines interpretation threshold.*/
	       Vector V,                /* Input state vector. */
	       String Interpretation);  /* Output interpreted string. */
void Make_stimulus(String Input_string,   /* Character string to be */
					  /* turned into a state vector. */
		   Vector Output_vector); /* Output state vector. */

static void Make_Binary_Vector(float Threshold, Vector V, IVector IV);
static int Count(Byte B);
static char Byte_Interpret(Byte B);
static void Read_IVector_as_ASCII(IVector IV, String Interpretation);
static void Zero_byte(Byte B);
static void Nr_to_byte(int Nr, Byte B);
static void Add_Byte_to_Vector(Vector V, int i, Byte B);

/*------------------------------------------------------------------------
     Both PROCEDUREs use an internal TYPE Byte, that is, an eight element 
array of integers and TYPE IVector, which is a vector composed of integer 
components. The programs are straightforward realizations of their 
functions.  It would be easy to make more compact and faster code, but, 
in practice, the amount of time spent doing the interpretation and 
translations are negligible.
 
     A minor problem arises because of numbering conventions for vector 
elements and bytes.  It is customary for vector elements to be numbered, 
as objects, in ascending order, left to right.  When bytes are listed 
in computer manuals, they are almost always listed as binary numbers, 
which means the elements are numbered from right to left.  This convention 
has been maintained, though at the cost of awkwardness in the code:


Vector     [ 1  2  3  4  5  6  7  8  9  10 11 12 13 14 15 16 17 ...
elements: 

                First Byte                 Second Byte          ...

Byte order ( 8  7  6  5  4  3  2  1)(8   7  6  5  4  3  2  1)(8 ...
in vector:

     The underline character, `_', gives rise to a byte of all zeros,
(0 0 0 0 0 0 0 0), allowing a convenient way of inserting zeros into the 
state vectors.  When a vector is interpreted, a byte at least one of
whose elements has an absolute value below the interpretation threshold 
is printed as the underline, `_'.

     A byte which would give rise to a nonprinting character like 
`Line Feed', `Form Feed', `Bell', `Escape', etc. is printed as `#.' 
------------------------------------------------------------------------*/


/*------------------------------------------------------------------------
		     Interpret
This routine takes a vector of floats, applies a threshold to so that
each element is classifed as +1) probably set; -1) probably clear; and 0)
ambiguous.  These intermediate values are stored in int IV[].  Then this
array is turned into a character string and returned as <Interpretation>.
The string will in general contain both ascii characters that have been
identified, and the character '_' if it is still ambiguous.

Uses void Make_Binary_Vector()
     int Count()
     char Byte_Interpret()
     void Read_IVector_as_ASCII()
Arguments
    float Threshold          Value between 0 and 1
    Vector V                 array to be interpreted
Returns
    String Interpretation    string of Number_of_characters, plus null
------------------------------------------------------------------------*/
void Interpret(float Threshold,   /* Determines interpretation threshold. */
	       Vector V,                /* Input state vector. */
	       String Interpretation) { /* Output interpreted string. */
    IVector IV;

    Make_Binary_Vector(Threshold, V, IV);
    Read_IVector_as_ASCII(IV, Interpretation);
}   /* Interpret */


/*------------------------------------------------------------------------
Applies a threshold to the values in <V>, and writes result to <IV>.
Each value in these arrays is considered to correspond to a bit of
a character string.  If the value is >Threshold, that bit is considered
set; if it is < -Threshold, that bit is considered cleared; otherwise it
is considered ambiguous.  'Set' is indicated by +1 in IV[], 'Cleared' is
indicated by -1 in IV[], and 'Ambiguous' is indicated by 0.

This routine is also used to recover the bit pattern of the input strings
from the float arrays F_set[] and G_set[].  These arrays consist of values
that are either less than zero or greater than zero.  Hence Threshold
is set to zero, and the if statements below should then classify each
float value either as +1 or -1 and never 0.

To make the behaviour when Threshold=0 clearer, I have changed the
inequality below from >= to >.

Used by Interpret()
Arguments
    float Threshold          Value between 0 and 1
    Vector V                 array to be interpreted
Returns
    IVector IV               result of applying threshold: array of -1,0 +1
------------------------------------------------------------------------*/
static void Make_Binary_Vector(float Threshold, Vector V, IVector IV) {
    int j;      /* loop index */

    for(j=0; j<Dimensionality; j++) {
	IV[j]=0;
	if(V[j] >   (float)fabs(Threshold)) IV[j]= 1;
	if(V[j] <= -(float)fabs(Threshold)) IV[j]=-1;
    }
}   /* Make_Binary_Vector */


/*------------------------------------------------------------------------
Counts the number of ambiguous bits (B[i]==0) indicated by the array <B>.
Used by Interpret()
------------------------------------------------------------------------*/
static int Count(Byte B) {
    int i;      /* loop index */
    int sum;    /* number of zeros in byte */

    sum=0;
    for(i=0; i<7; i++)          /* Ignore parity bit. */
	if(B[i]==0) sum++;
    return sum;
}   /* Count */


/*------------------------------------------------------------------------
B is an array of eight integers each equal to +1, 0 or -1.  This routine
converts it to an ascii character.  If a group of 8 contains any 0 values,
the corresponding character is considered ambiguous and a '_' results.
Otherwise +1 causes the corresponding bit to be set, and -1 causes the
corresponding bit to be cleared.

Used by Interpret()
------------------------------------------------------------------------*/
static char Byte_Interpret(Byte B) {
    int i;      /* loop index */
    int Charnr; /* ASCII code of character, as an integer */

    /* First check to see if all byte elements + 1 or - 1. */
    if(Count(B)>0)
	return '_';     /* character is still ambiguous */
    else {
	/* Now convert to byte of zero and +1. */
	for(i=0; i<8; i++)
	    if(B[i]==-1) B[i]=0;
	Charnr=64*B[6] + 32*B[5] + 16*B[4] + 8*B[3] + 4*B[2] + 2*B[1] + B[0];
	if(Charnr==0 || Charnr==127)    /* Zero/delete convention is '_' */
	    return '_';
	else if(Charnr<32)              /* Non-printing character. */
	    return '#';
	else                            /* Print character. */
	    return (char)Charnr;
    }
}   /* Byte_Interpret */

 
/*------------------------------------------------------------------------
Takes the vector <IV>, consisting of 8*Number_of_characters integers, and
converts it to an ascii string.

Used by Interpret()
Uses Byte_Interpret()
Arguments
    IVector IV               result of applying threshold: array of -1,0 +1
Returns
    String Interpretation    ascii string corresponding to IV
------------------------------------------------------------------------*/
static void Read_IVector_as_ASCII(IVector IV, String Interpretation) {
    int i,j;    /* loop indices */
    Byte B;     /* eight consecutive values copied from IV[] */

    for(j=0; j<Number_of_characters; j++) {
	/* Next 'for loop' flips order of vector elements. */
	for(i=0; i<8; i++)
	    B[7-i]=IV[j*8+i];
	Interpretation[j]=Byte_Interpret(B);
    }
    Interpretation[Number_of_characters]='\0';  /* Add null termination */
}   /* Read_IVector_as_ASCII */



/*------------------------------------------------------------------------
			 Make_stimulus
This routine is called by Read_stimuli.  It takes a string containing
<Number_of_characters> ascii characters, and for each of the
8*Number_of_characters bits sets the corresponding element of
<Output_vector> to +1 (if bit is set), or to 0 (if bit is clear).
Subsequently in Initialize(), the output vectors will be 'normalized'
by dividing each element by sqrt(Dimensionality).
Uses void Zero_byte()
     void Nr_to_byte()
     void Add_Byte_to_Vector()
Arguments
    String Input_string      string of Number_of_characters, plus null
Returns
    Vector Output_vector     8*Number_of_characters integers, each 1 or 0
------------------------------------------------------------------------*/
void Make_stimulus(String Input_string, Vector Output_vector) {
    int i;      /* loop index */
    Byte B;     /* One character, expanded to an array of 8 integers */

    for(i=0; i<Number_of_characters; i++) {
	Nr_to_byte((int)Input_string[i],B);
	if(Input_string[i]=='_') Zero_byte(B);
	Add_Byte_to_Vector(Output_vector, i ,B);
    }
}   /* Make_stimulus */


/*------------------------------------------------------------------------
Sets the array of 8 integers to zero.
Called by Make_stimulus()
------------------------------------------------------------------------*/
static void Zero_byte(Byte B) {
    int i;      /* loop index */

    for(i=0; i<8; i++)
	B[i]=0;
}   /* Zero_byte */


/*------------------------------------------------------------------------
Copies each of the seven least significant bits of <Nr> to the elements
of B[].
Called by Make_stimulus()
------------------------------------------------------------------------*/
static void Nr_to_byte(int Nr, Byte B) {
    int i;              /* loop index */
    int Parity=0;       /* sum of all bits */
    int mask;           /* bit mask */

    if(Nr>=128) {
	printf("Too large. ASCII conversion fails.\n"); return;
    } else if(Nr<0) {
	printf("Negative number. ASCII conversion fails.\n"); return;
    }

    /* Copy bits 6..0 of Nr to the corresponding elements of B[] */
    for(mask=0x40,i=6; i>=0; mask>>=1,i--) {
	if(Nr&mask) {
	    B[i]=1; Parity++;
	} else
	    B[i]=-1;
    }
    /* Set bit 7 such that the sum of all bits mod 2 is equal to 0 */
    B[7]=(Parity&1)? 1: -1;
}   /* Nr_to_byte */


/*------------------------------------------------------------------------
Copies 8 integers from B to sequential elements of V, starting from
element 8*i.
Called by Make_stimulus()
------------------------------------------------------------------------*/
static void Add_Byte_to_Vector(Vector V, int i, Byte B) {
    int j;      /* loop index */

    if(i<0 || i>=Number_of_characters)
        printf("Add_Byte_to_Vector(): i=%d is out of range\n",i);
    else
        for(j=0; j<8; j++)
            V[i*8+j]=B[7-j];
}   /* Add_Byte_to_Vector */


/*------------------------------------------------------------------------
See p200-201.
------------------------------------------------------------------------*/
void main() {
    String Input_string="00abcdefghijklmnopqrstuvw";
    String Interpretation;
    Vector Output_vector;
    int i;              /* loop index */

    Input_string[0]=0;
    Input_string[1]=5;
    Make_stimulus(Input_string, Output_vector);
    for(i=0; i<Dimensionality; i++)
	Output_vector[i]=(float)(20*Output_vector[i]-10);
    Interpret(5, Output_vector, Interpretation);
    printf("Results=%s\n",Interpretation);
}   /* main */
