/* File appk.c, Program Boltz */

/*
                            Appendix K
                    A Simple Boltzmann Machine


     This program implements a simple Boltzmann machine.  Connectivity 
is local and symmetric, with all units connected to their four neighbors, 
above, below, to the left and to the right, with weights of +1.  There 
are two global minima, with system energy -450 units, one composed of 
all +1, and composed of all -1.  The units are connected in a toroid, 
with top and bottom and left and right rows and columns neighbors.  The 
Gibbs distribution is used to decide whether or not a unit changes state.

     The program prompts for a random number seed, total number of 
updatings to perform, a system temperature, and the number of 
asynchronous updatings to perform before updating the screen display.  
(Screen updates take much more time than the computation itself.)  The 
lowest energy solution is displayed first.  A <CR> shows the initial 
configuration and a second <CR> starts the updating process.  

     The annealing schedule defined in function Flip is too fast, so 
a very deep local energy minimum -- a band of one sign in the midst of 
the other sign -- occurs occasionally.  This local minimum has system 
energy of 390 units, well above the lowest energy state.   Experimentation 
with the annealing schedule is interesting but time consuming.  The 
program will stop if the lowest energy state is reached or when it 
finishes the required number of iterations. 

     The program itself is a straightforward realization of the dynamics 
of a Boltzmann network as discussed in Chapter 12.  Any complexity in the 
program is primarily due to the human factors and the display, as is 
usually the case.  

     For reasons of readability, a state vector unit or element, called
f[i,j] in the text, is incorporated into a structure "Element", 
called Surface[I][J] in the program,  which contains the weights as well
as the activities of units in the simulation.  The activities are called 
Surface[I][J].F in the program.

     An input system temperature of 0 units will lead to a system 
behaving like a Hopfield network, where system energy decreases 
monotonically.  The only difference between this network and a Hopfield 
network as we defined it in Chapter 12 is that a state change that 
produces a zero change in energy is never accepted.  This was done 
because the extreme simplicity of the network used and its intrinsic 
symmetries will always lead to a solution of all '+' if the original 
updating rule is used.  This is because the original rule says a unit 
changes to the '+' state if the synaptic inner product is equal to 
or greater than zero.  The rule used by the program for zero temperature
says that the state stays unchanged if the energy is unchanged.

     Some of the quantities involving EXP (Delta_Energy/T), computed in 
the program can get large as temperature drops and floating point under- 
and overflows may occur. A few lines of code are incorporated in 
function Flip to prevent this.

     See pp 426-431.
*/

#if defined MSDOS
#include <graph.h>      /* _clearscreen _settextposition */
#endif
#include <math.h>       /* exp */
#include <stdio.h>      /* fflush fgetc printf scanf */
#include <stdlib.h>     /* rand srand RAND_MAX */

#define X_size 15
#define Y_size 15

/* Each element consists of five integers .F .C.Up .C.Down .C.Right .C.Left */
typedef struct {
    int F;
    struct {
	int Up;
	int Down;
	int Right;
	int Left;
    } C;
} Element;


Element Surface[Y_size+2][X_size+2];
long Iteration;                 /* Iteration counter */
long Number_of_iterations;      /* Required number of iterations */
unsigned int Old_seed;          /* Seed entered by user */
float T_zero;                   /* initial temperature as float */
int W_T_zero;                   /* initial temperature as int */
float T;                        /* current temperature as float */
int Display;                    /* number of iterations per display update */
int Ground_state;
int Current_energy;

float Random(void);
void ClrScr(void);
void GoToXY(int x, int y);
void Draw_Surface(void);
void Initialize(void);
int B_Random(double F);
void Randomize_surface(void);
void Expand_array(void);
int Element_energy(int x, int y);
int System_energy(void);
void Flip(void);
void Boltz(void);

/*--------------------------------------------------------------------------
Returns a float number >=0 and < 1.0
--------------------------------------------------------------------------*/
float Random(void) {
    return ((float)rand())/((float)RAND_MAX+1);
}   /* Random */


/*--------------------------------------------------------------------------
Screen Control Procedure: Clear screen and move cursor to 1,1
--------------------------------------------------------------------------*/
void ClrScr(void) {
#if defined MSDOS
    _clearscreen(_GCLEARSCREEN);
#else
    printf("%c[2J",0x1b);
    printf("%c[%d;%dH",0x1b,1,1);
#endif
}   /* ClrScr */


/*--------------------------------------------------------------------------
Screen Control Procedure: Position cursor

Arguments
    x    column as integer 1..80
    y    row as integer 1..24
--------------------------------------------------------------------------*/
void GoToXY(int x, int y) {
    int x1,y1;          /* coordinates restricted to allowed range */

    x1=(x<1)? 1: (x>80)? 80: x;         /* 1<=x1<=80 */
    y1=(y<1)? 1: (y>24)? 24: y;         /* 1<=y1<=24 */
#if defined MSDOS
    _settextposition(y,x);
#else
    printf("%c[%d;%dH",0x1b,y1,x1);
#endif
}   /* GoToXY */



/*------------------------------------------------------------------------
This routine draws a grid representing the state of the <X_size> x <Y_size>
nodes of the system.  For each node, the symbols '+' or '-' are drawn
depending on the sign of the value at that node.
------------------------------------------------------------------------*/
void Draw_Surface(void) {
    int i,j;    /* column and row */

    GoToXY(1,1);
    printf("Iteration  : %ld\n\n",Iteration);

    for(j=1; j<=Y_size; j++) {
	for(i=1; i<=X_size; i++)
	    printf("%s", (Surface[j][i].F>0)? " +": " -");
	printf("\n");
    }
    GoToXY(1,20); printf("Temperature   : %6d",(int)(T+0.5));
    GoToXY(1,21); printf("Current energy: %6d",Current_energy);
    GoToXY(1,22); printf("Ground State  : %6d",Ground_state);
}   /* Draw_Surface */


/*------------------------------------------------------------------------
------------------------------------------------------------------------*/
void Initialize(void) {
    int i,j;            /* loop indices */
    unsigned int Seed;  /* seed for srand() */

    ClrScr();
    printf("Demonstration of simple spin flip Boltzmann machine.\n");
    printf("Only nearest neighbor interactions are used.\n");
    printf("Lowest energy final solution will be homogeneous.\n");
    printf("Choice of zero temperature disables noise addition.\n");
    printf("Initial states +1, -1 with equal probability.\n");
    printf("See pp 426-431.\n");

    printf("\nSeed for random number generator (0..65535): ");
    scanf("%u",&Seed); srand(Seed);
    Old_seed=Seed;

    do {
	printf("Number of iterations   (>=1; 20000 approx): ");
	scanf("%ld",&Number_of_iterations);
    } while(Number_of_iterations<1);

    do {
	printf("Initial Temperature    (>0.0; 10 approx): ");
	scanf("%f",&T_zero);
    } while(T_zero<=0);
    T=T_zero;
    W_T_zero=(int)(T_zero+0.5);

    do {
	printf("Number of iterations per display update (>=1; 10 approx): ");
	scanf("%d",&Display);
    } while(Display<1);

    for(j=0; j<=Y_size+1; j++)
	for(i=0; i<=X_size+1; i++) {
	    Surface[j][i].F       =  1;  /* Initialize all values to +1 */
	    Surface[j][i].C.Up    =  1;  /* All weights are set at + 1 */
	    Surface[j][i].C.Down  =  1;
	    Surface[j][i].C.Right =  1;
	    Surface[j][i].C.Left  =  1;
	}
}   /* Initialize */


/*------------------------------------------------------------------------
Returns a random binary value.  FALSE has a probablility of <F>, and TRUE
has a probability of 1-<F>.
Argument
    float F            probability of FALSE, 0<=F<=1, input
------------------------------------------------------------------------*/
int B_Random(double F) {
    return (Random()>=(float)F)? 1: 0;
}   /* B_Random */


/*------------------------------------------------------------------------
This routine is called once near the start of the program in order to
generate a random starting pattern.
Weights all remain at +1 from Initialize().
------------------------------------------------------------------------*/
void Randomize_surface(void) {
    int i,j;            /* loop indices */

    Iteration=1;

    /* Randomize state; 0.5 probability of a +1 */
    for(j=0; j<=Y_size+1; j++)
	for(i=0; i<=X_size+1; i++) {
	    Surface[j][i].F=B_Random(0.5)? 1: -1;
	}
}   /* Randomize_surface */


/*------------------------------------------------------------------------
Copies rows and columns to produce toroidal wrap around.  There are just
X_size x Y_size independent nodes, so the extra two rows and two columns
just mirror the values on the opposite side.
------------------------------------------------------------------------*/
void Expand_array(void) {
    int i,j;            /* loop indices */

    for(i=1; i<=X_size; i++) {
	Surface[0][i]       =Surface[Y_size][i];  /* Copies row 15 to row 0 */
	Surface[Y_size+1][i]=Surface[1][i];       /* Copies row 1 to row 16 */
    }
    for(j=1; j<=Y_size; j++) {
	Surface[j][0]       =Surface[j][X_size];  /* Copies col 15 to col 0 */
	Surface[j][X_size+1]=Surface[j][1];       /* Copies col 1 to col 16 */
    }
}   /* Expand_array */


/*------------------------------------------------------------------------
------------------------------------------------------------------------*/
int Element_energy(int x, int y) {
    int C_value, R_value, L_value, U_value, D_value;
    int Energy;

    /* Neighbors of element we are looking at. */
    U_value=Surface[y-1][x].F;
    D_value=Surface[y+1][x].F;
    R_value=Surface[y][x+1].F;
    L_value=Surface[y][x-1].F;
    C_value=Surface[y][x].F;

    /* Computation of energy associated with connnections to single element */
    Energy = -(C_value * U_value * Surface[y][x].C.Up    +
	       C_value * D_value * Surface[y][x].C.Down  +
	       C_value * R_value * Surface[y][x].C.Right +
	       C_value * L_value * Surface[y][x].C.Left )/2;

    return Energy;
}   /* Element_energy */


/*------------------------------------------------------------------------
------------------------------------------------------------------------*/
int System_energy(void) {
    int i,j;            /* loop indices */
    int Sum;            /* total system energy */

    Expand_array();  /* Copy rows and columns to give toroidal geometry. */
    Sum=0;
    for(j=1; j<=Y_size; j++)
	for(i=1; i<=X_size; i++)
	    Sum+=Element_energy(i,j);
    return Sum;
}   /* System_energy */


/*------------------------------------------------------------------------
------------------------------------------------------------------------*/
void Flip(void) {
    int i,j;        /* coords of randomly chosen node */
    int New,Old;    /* system energy corresponding to flipped and original */
    int Delta_energy;   /* 2(New-Old) */
    double Sched;       /* annealing schedule */
    float Ratio;        /* normalized Delta_energy */
    double Prob_unchanged_state;        /* prob of no state change */

    i=(int)(Random()*X_size)+1;
    j=(int)(Random()*Y_size)+1;
    Expand_array();     /* Copy rows and columns to give toroidal geometry. */

    /* Flip to the new state. */
    Surface[j][i].F *= -1;
    New=Element_energy(i,j);
    /* Flip back to original state. */
    Surface[j][i].F *= -1;
    Old=Element_energy(i,j);
    Delta_energy=2*(New - Old);  /* Factor of 2 because we are  */
				 /* computing for a single unit. */

    /* Hopfield network condition. */
    /* Slight variation on Hopfield: Change state only if energy is lowered */
    /* no matter what weights indicate. */
    if(T_zero==0)
	if(Delta_energy<0) Surface[j][i].F *= -1;

    /* Boltzmann machine condition. */
    if(T_zero>0) {
	Sched= -(double)Iteration/3000; /* Change this for different schedule */
	T=T_zero*(float)exp(Sched);     /* Annealing schedule */
	Ratio=Delta_energy/T;
	/* Limit Ratio to -50..+50 */
	if(Ratio>50) Ratio=50;
	if(Ratio<-50) Ratio=-50;
	/* Otherwise we have floating point overflows at very low temps. */

	/* Gibbs distribution: prob (1)/prob(2) = exp (Ratio). */
	/* prob(1) = unchanged state; prob(2) = changed state. */
	Prob_unchanged_state=exp(Ratio)/(1+exp(Ratio));

	/* Call B_Random.  If Delta_E large, positive, very unlikely to flip */
	/* If Delta_E zero, then equally likely to flip or not flip. */
	if(B_Random(Prob_unchanged_state))
	    Surface[j][i].F *= -1;
    }
}   /* Flip */


/*------------------------------------------------------------------------
PROGRAM Boltz
------------------------------------------------------------------------*/
void main(void) {

    Initialize();
    fflush(stdin);

    /* Display endpoint, and calculate Ground_state */
    ClrScr();
    GoToXY(1,19);
    printf("One of Two Lowest Energy States of this System.");
    Ground_state=System_energy();
    Draw_Surface();
    GoToXY(1,18); printf(">> Hit return to continue.");
    fgetc(stdin); fflush(stdin);
    ClrScr();

    /* Display random starting pattern */
    Randomize_surface();
    Current_energy=System_energy();
    Iteration=0;
    Draw_Surface();
    GoToXY(1,18);  printf(">> Hit return to continue.");
    fgetc(stdin); fflush(stdin);
    if(Number_of_iterations > 0) ClrScr();

    while((Iteration<Number_of_iterations) && (Current_energy>Ground_state)) {
	Flip();
	Iteration++;
	if(Iteration%Display==0) {
	    Current_energy=System_energy();
	    Draw_Surface();
	}
    }

    GoToXY(1,23);
    if(Iteration >= Number_of_iterations)
	printf(">> Finished Iterations. Initial Temp: %4d Seed: %10u\n",
	    W_T_zero,Old_seed);
    if(Current_energy==Ground_state)
	printf(">> Reached ground state. Initial Temp: %4d Seed: %10u\n",
	    W_T_zero,Old_seed);

    /* fgetc(stdin); fflush(stdin); */
}   /* Boltz */
