/* File appm.c, Program BSB25 */

/*
                             Appendix M
                           BSB Procedure 

     This program implements a simple version of the BSB procedure.  It 
is essentially the program and data set used in Appendix J, which 
implemented the Widrow-Hoff learning rule.  This program was modified 
by the addition of code to implement the BSB procedure during retrieval.  
The Widrow-Hoff learning has been placed in a single routine Learn().  
The BSB Retrieval process has been placed in BSB_Iteration().  
Note: A much more flexible version of BSB is present with these files 
in Program BSB.  See the Manual for details of its use.

     The retrieval constants used are embedded in the program in 
routine BSB_Parameters().  They can be easily changed and are not 
critical.  A few additional global parameters were defined as well.  

     When the program is run, it will prompt for (1) the name of the 
file containing the character strings to be learned; (2) a random number 
seed; (3) the interpretation threshold (0.5 is a good starting point with 
BSB so that small noisy element values will not be displayed); (4) the 
number of vectors to learn (15 for the state data base) and (5) the 
number of Widrow-Hoff learning trials. (In the simulation in Figure 15.11 
this was 150.).

     The BSB iterations will stop when an attractor is reached, that is, 
a fully limited state vector, or when 25 iterations (parameter  
Max_number_of_iterations) have been performed.  The display tells how 
many units have reached their limits so the progress of growth of the 
state vector can be followed.  The program will run through the entire 
set of vectors in the file.  Test vectors should be placed in the data 
base after the vectors to be learned, as was done in previous versions of 
this program. 

    See p515.

*/

#if defined MSDOS
#include <io.h>         /* access */
#else
#include <unistd.h>     /* access */
#endif
#include <math.h>       /* fabs sqrt */
#include <stdio.h>      /* feof fgets fopen fclose printf scanf */
#include <stdlib.h>     /* rand srand RAND_MAX */
#include <string.h>     /* strchr strncpy strlen strcat */

/* Number_of_characters was 25, and Max_no_associations was 100 */
#define Number_of_characters (int)(25)
#define Dimensionality       (int)(8*Number_of_characters)
#define Max_no_associations  (int)(100)

typedef float Matrix[Dimensionality][Dimensionality];
typedef float Vector[Dimensionality];
typedef char String[Number_of_characters+1];
typedef int Byte[8];
typedef int IVector[Dimensionality];

int Number_of_associations;
int Pairs_to_learn;
int Total_learning_trials;
Matrix A,Delta_A;
char File_name[40];
float Threshold;
float Eta_prime;
Vector F_Set[Max_no_associations];
Vector G_Set[Max_no_associations];

void Subtract_vectors(float Vector_1[], float Vector_2[], 
                      float Vector_difference[]);
void Scalar_times_vector(float Scalar, Vector Vector_in, Vector Vector_out);
float Inner_product(float Vector_1[], float Vector_2[]);
float Vector_length(float Vector_1[]);
void Normalize(float Vector_in[], float Vector_out[]);
void Matrix_times_vector(Matrix A, Vector F, Vector G);
void Add_matrices(Matrix Mx_1, Matrix Mx_2, Matrix Matrix_sum);
void Outer_product(Vector G, Vector F, Matrix A);

void Interpret(float Threshold,   /* Determines interpretation threshold.*/
	       Vector V,                /* Input state vector. */
	       String Interpretation);  /* Output interpreted string. */
static void Make_Binary_Vector(float Threshold, Vector V, IVector IV);
static int Count(Byte B);
static char Byte_Interpret(Byte B);
static void Read_IVector_as_ASCII(IVector IV, String Interpretation);
void Make_stimulus(String Input_string,   /* Character string to be */
					  /* turned into a state vector. */
		   Vector Output_vector); /* Output state vector. */
static void Zero_byte(Byte B);
static void Nr_to_byte(int Nr, Byte B);
static void Add_Byte_to_Vector(Vector V, int i, Byte B);
void Write_string(String S);
void Read_input_line(FILE* IPFile, String Input_string);
void Read_stimuli(void);
void Associate(Vector F, Vector G);
void Learn(void);
void BSB_Iteration(int *pNumber_limited,
		   Vector Original_vector,      /* Starting vector. */
		   Vector V_in,                 /* Input vector. */
		   Vector V_out);               /* Output vector. */
void Write_output(int Iteration_number,
		  int Number_limited,
		  Vector V_in);
void Initialize(void);
float Random(void);
void main(void);


/*--------------------------------------------------------------------------
Subtracts two vectors.
Arguments
    float Vector_1[]            One vector, input
    float Vector_2[]            Other vector, input
    float Vector_difference[]   Vector_1-Vector_2, returned
--------------------------------------------------------------------------*/
void Subtract_vectors(float Vector_1[], float Vector_2[], 
                      float Vector_difference[]) {
    int i;      /* loop index */

    for(i=0; i<Dimensionality; i++)
	Vector_difference[i]=Vector_1[i] - Vector_2[i];
}   /* Subtract_vectors */



/*--------------------------------------------------------------------------
Scales a vector.
Arguments
    float Scalar                Factor, input
    float Vector_in[]           Vector, input
    float Vector_out[]          Factor*Vector_in, returned
--------------------------------------------------------------------------*/
void Scalar_times_vector(float Scalar, Vector Vector_in, Vector Vector_out) {
    int i;      /* loop index */

    for(i=0; i<Dimensionality; i++)
	Vector_out[i]=Scalar*Vector_in[i];
}   /* Subtract_vectors */



/*--------------------------------------------------------------------------
Forms the inner product of two vectors.
Arguments
    float Vector_1[]            One vector, input
    float Vector_2[]            Other vector, input
Returns
    float Inner_product         Vector_1 . Vector_2
--------------------------------------------------------------------------*/
float Inner_product(float Vector_1[], float Vector_2[]) {
    float Sum_of_products;      /* Name says it all */
    int i;                      /* loop index */

    Sum_of_products=0;
    for(i=0; i<Dimensionality; i++)
	Sum_of_products += Vector_1[i]*Vector_2[i];
    return Sum_of_products;
}   /* Inner_product */


/*--------------------------------------------------------------------------
Returns the length of a vector.
Argument
    float Vector_1[]            Vector, input
Returns
    float length                Magnitude of Vector_1
--------------------------------------------------------------------------*/
float Vector_length(float Vector_1[]) {
    float Sum_of_Squares;       /* Name says it all */
    int i;                      /* loop index */

    Sum_of_Squares=0;
    for(i=0; i<Dimensionality; i++)
	Sum_of_Squares += Vector_1[i]*Vector_1[i];
    return (float)sqrt(Sum_of_Squares);
}   /* Vector_length */



/*--------------------------------------------------------------------------
Scales the input vector to unit length.  This is not possible if the input
vector has zero length, and in that case an error message is printed, and
the output vector is set to zero.

This routine may be called with both arguments pointing to the same array.

Arguments
    float Vector_in[]           Vector, input
    float Vector_out[]          Normalized vector, output
--------------------------------------------------------------------------*/
void Normalize(float Vector_in[], float Vector_out[]) {
    float Length_of_vector_in;  /* Name says it all */
    int i;                      /* loop index */

    Length_of_vector_in=Vector_length(Vector_in);
    if(Length_of_vector_in==0) {
	printf("Normalize: Input vector has zero length!\n");
	for(i=0; i<Dimensionality; i++)
	    Vector_out[i] = 0;
    } else {
	for(i=0; i<Dimensionality; i++)
	    Vector_out[i] = Vector_in[i]/Length_of_vector_in;
    }
}   /* Normalize */



/*-------------------------------------------------------------------------
This function forms the product Af = g where both f and g are column
vectors.
-------------------------------------------------------------------------*/
void Matrix_times_vector(Matrix A, Vector F, Vector G) {
    int i,j;    /* loop indices */

    /* Zero vector G */
    for(i=0; i<Dimensionality; i++)
	G[i]=0;
    for(i=0; i<Dimensionality; i++)
	for(j=0; j<Dimensionality; j++)
	    G[i] += A[i][j]*F[j];
}   /* Matrix_times_vector */


/*-------------------------------------------------------------------------
Calculates the sum of 2 matrices dimensioned [Dimensionality][Dimensionality]
Arguments
    Matrix Mx_1[]            One matrix, input
    Matrix Mx_2[]            Other matrix, input
Returns
    Matrix Matrix_sum        Mx_1 + Mx_2
-------------------------------------------------------------------------*/
void Add_matrices(Matrix Mx_1, Matrix Mx_2, Matrix Matrix_sum) {
    int i,j;    /* loop indices */

    for(i=0; i<Dimensionality; i++)
	for(j=0; j<Dimensionality; j++)
	    Matrix_sum[i][j]=Mx_1[i][j]+Mx_2[i][j];
}   /* Add_matrices */


/*-------------------------------------------------------------------------
We want to generate an outer product matrix A so that
	    T
     A = G F
-------------------------------------------------------------------------*/
void Outer_product(Vector G, Vector F, Matrix A) {
    int i,j;    /* loop indices */

    for(i=0; i<Dimensionality; i++)
	for(j=0; j<Dimensionality; j++)
	    A[i][j] = G[i]*F[j];
}   /* Outer_product */



/*------------------------------------------------------------------------
		Interpret
This routine takes a vector of floats, applies a threshold to so that
each element is classifed as +1) probably set; -1) probably clear; and 0)
ambiguous.  These intermediate values are stored in int IV[].  Then this
array is turned into a character string and returned as <Interpretation>.
The string will in general contain both ascii characters that have been
identified, and the character '_' if it is still ambiguous.

This routine is used in Test_whole_set() to turn the raw association
output from Matrix_times_vector() into a string.  It is also used there
to recover the input and required output strings from F_Set[] and G_Set[].
The arrays F_Set[] and G_Set[] consist of float values that are either
zero or greater than zero, and so for this second application it is just
a matter of setting Threshold to zero and of supplying a vector from
F_Set[] or G_Set[].

Uses void Make_Binary_Vector()
     int Count()
     char Byte_Interpret()
     void Read_IVector_as_ASCII()
Arguments
    float Threshold          Value between 0 and 1
    Vector V                 array to be interpreted
Returns
    String Interpretation    string of Number_of_characters, plus null
------------------------------------------------------------------------*/
void Interpret(float Threshold,   /* Determines interpretation threshold. */
	       Vector V,                /* Input state vector. */
	       String Interpretation) { /* Output interpreted string. */
    IVector IV;

    Make_Binary_Vector(Threshold, V, IV);
    Read_IVector_as_ASCII(IV, Interpretation);
}   /* Interpret */


/*------------------------------------------------------------------------
Applies a threshold to the values in <V>, and writes result to <IV>.
Each value in these arrays is considered to correspond to a bit of
a character string.  If the value is >Threshold, that bit is considered
set; if it is < -Threshold, that bit is considered cleared; otherwise it
is considered ambiguous.  'Set' is indicated by +1 in IV[], 'Cleared' is
indicated by -1 in IV[], and 'Ambiguous' is indicated by 0.

This routine is also used to recover the bit pattern of the input strings
from the float arrays F_Set[] and G_Set[].  These arrays consist of values
that are either less than zero or greater than zero.  Hence Threshold
is set to zero, and the if statements below should then classify each
float value either as +1 or -1 and never 0.

To make the behaviour when Threshold=0 clearer, I have changed the
inequality below from >= to >.

Used by Interpret()
Arguments
    float Threshold          Value between 0 and 1
    Vector V                 array to be interpreted
Returns
    IVector IV               result of applying threshold: array of -1,0 +1
------------------------------------------------------------------------*/
static void Make_Binary_Vector(float Threshold, Vector V, IVector IV) {
    int j;      /* loop index */

    for(j=0; j<Dimensionality; j++) {
	IV[j]=0;
	if(V[j] >   (float)fabs(Threshold)) IV[j]= 1;
	if(V[j] <= -(float)fabs(Threshold)) IV[j]=-1;
    }
}   /* Make_Binary_Vector */


/*------------------------------------------------------------------------
Counts the number of ambiguous bits (B[i]==0) indicated by the array <B>.
Used by Interpret()
------------------------------------------------------------------------*/
static int Count(Byte B) {
    int i;      /* loop index */
    int sum;    /* number of zeros in byte */

    sum=0;
    for(i=0; i<7; i++)          /* Ignore parity bit. */
	if(B[i]==0) sum++;
    return sum;
}   /* Count */


/*------------------------------------------------------------------------
B is an array of eight integers each equal to +1, 0 or -1.  This routine
converts it to an ascii character.  If a group of 8 contains any 0 values,
the corresponding character is considered ambiguous and a '_' results.
Otherwise +1 causes the corresponding bit to be set, and -1 causes the
corresponding bit to be cleared.

Used by Interpret()
------------------------------------------------------------------------*/
static char Byte_Interpret(Byte B) {
    int i;      /* loop index */
    int Charnr; /* ASCII code of character, as an integer */
    Byte C;     /* copy of input array of eight integers */

    /* First check to see if all byte elements + 1 or - 1. */
    if(Count(B)>0)
	return '_';     /* character is still ambiguous */
    else {
        /* Convert B[] to an array of zero and +1. */
        for(i=0; i<8; i++)
            C[i]=(B[i]==-1)? 0: 1;

	Charnr=64*C[6] + 32*C[5] + 16*C[4] + 8*C[3] + 4*C[2] + 2*C[1] + C[0];
	if(Charnr==0 || Charnr==127)    /* Zero/delete convention is '_' */
	    return '_';
	else if(Charnr<32)              /* Non-printing character. */
	    return '#';
	else                            /* Print character. */
	    return (char)Charnr;
    }
}   /* Byte_Interpret */


/*------------------------------------------------------------------------
Takes the vector <IV>, consisting of 8*Number_of_characters integers, and
converts it to an ascii string.

Used by Interpret()
Uses Byte_Interpret()
Arguments
    IVector IV               result of applying threshold: array of -1,0 +1
Returns
    String Interpretation    ascii string corresponding to IV
------------------------------------------------------------------------*/
static void Read_IVector_as_ASCII(IVector IV, String Interpretation) {
    int i,j;    /* loop indices */
    Byte B;     /* eight consecutive values copied from IV[] */

    for(j=0; j<Number_of_characters; j++) {
	/* Next 'for loop' flips order of vector elements. */
	for(i=0; i<8; i++)
	    B[7-i]=IV[j*8+i];
	Interpretation[j]=Byte_Interpret(B);
    }
    Interpretation[Number_of_characters]='\0';  /* Add null termination */
}   /* Read_IVector_as_ASCII */



/*------------------------------------------------------------------------
		Make_stimulus
This routine is called by Read_stimuli.  It takes a string containing
<Number_of_characters> ascii characters, and for each of the
8*Number_of_characters bits sets the corresponding element of
<Output_vector> to +1 (if bit is set), or to 0 (if bit is clear).
Subsequently in Initialize(), the output vectors will be 'normalized'
by dividing each element by sqrt(Dimensionality).
Uses void Zero_byte()
     void Nr_to_byte()
     void Add_Byte_to_Vector()
Arguments
    String Input_string      string of Number_of_characters, plus null
Returns
    Vector Output_vector     8*Number_of_characters integers, each 1 or 0
------------------------------------------------------------------------*/
void Make_stimulus(String Input_string, Vector Output_vector) {
    int i;      /* loop index */
    Byte B;     /* One character, expanded to an array of 8 integers */

    for(i=0; i<Number_of_characters; i++) {
	Nr_to_byte((int)Input_string[i],B);
	if(Input_string[i]=='_') Zero_byte(B);
	Add_Byte_to_Vector(Output_vector, i ,B);
    }
}   /* Make_stimulus */


/*------------------------------------------------------------------------
Sets the array of 8 integers to zero.
Called by Make_stimulus()
------------------------------------------------------------------------*/
static void Zero_byte(Byte B) {
    int i;      /* loop index */

    for(i=0; i<8; i++)
	B[i]=0;
}   /* Zero_byte */


/*------------------------------------------------------------------------
Copies each of the seven least significant bits of <Nr> to the elements
of B[].
Called by Make_stimulus()
------------------------------------------------------------------------*/
static void Nr_to_byte(int Nr, Byte B) {
    int i;              /* loop index */
    int Parity=0;       /* sum of all bits */
    int mask;           /* bit mask */

    if(Nr>=128) {
	printf("Too large. ASCII conversion fails.\n"); return;
    } else if(Nr<0) {
	printf("Negative number. ASCII conversion fails.\n"); return;
    }

    /* Copy bits 6..0 of Nr to the corresponding elements of B[] */
    for(mask=0x40,i=6; i>=0; mask>>=1,i--) {
	if(Nr&mask) {
	    B[i]=1; Parity++;
	} else
	    B[i]=-1;
    }
    /* Set bit 7 such that the sum of all bits mod 2 is equal to 0 */
    B[7]=(Parity&1)? 1: -1;
}   /* Nr_to_byte */


/*------------------------------------------------------------------------
Copies 8 integers from B to sequential elements of V, starting from
element 8*i.
Called by Make_stimulus()
------------------------------------------------------------------------*/
static void Add_Byte_to_Vector(Vector V, int i, Byte B) {
    int j;      /* loop index */

    if(i<0 || i>=Number_of_characters)
        printf("Add_Byte_to_Vector(): i=%d is out of range\n",i);
    else
        for(j=0; j<8; j++)
            V[i*8+j]=B[7-j];
}   /* Add_Byte_to_Vector */



/*-------------------------------------------------------------------------
Prints the specified string to screen.
-------------------------------------------------------------------------*/
void Write_string(String S) {
    printf("%s",S);
}   /* Write_string */



/*-------------------------------------------------------------------------
Reads one line of input from stream IPFile, and copies to the string
Input_string.
Up to 80 characters can be read, but only the first 25 will be copied.
If there are less than 25 (even 0) characters read, then the contents
of Input_string will be padded with blank characters to a length of 25.
-------------------------------------------------------------------------*/
void Read_input_line(FILE* IPFile, String Input_string) {
    char Input_line[80+2];      /* maximum of 80 characters+newline+null */

    /* read one line from input, and remove newline character, if present */
    fgets(Input_line,sizeof(Input_line),IPFile);
    if(strchr(Input_line,'\n'))
	*strchr(Input_line,'\n')='\0';

    /* Copy at most <Number_of_characters> characters from input line */
    strncpy(Input_string,Input_line,sizeof(String));
    Input_string[sizeof(String)-1]='\0';        /* Ensure null termination */

    /* Pad input with blanks, if necessary */
    while(strlen(Input_string)<Number_of_characters)
	strcat(Input_string," ");
}   /* Read_input_line */


/*-------------------------------------------------------------------------
Read the contents of the file specified by <File_name>, which should
consist of groups of three lines.  The first line can contain anything,
the second will be copied to F_Set[n], and the third to G_Set[n].  The
total number of groups read successfully is stored as Number_of_associations.
-------------------------------------------------------------------------*/
void Read_stimuli(void) {
    String Input_string;        /* 25 characters read from input */
    FILE *Stimuli;              /* pointer to data file */

    Stimuli=fopen(File_name,"rt");
    Number_of_associations=0;
    do {
	/* Read 1st line, but this is ignored */
	Read_input_line(Stimuli,Input_string);
	if(feof(Stimuli)) break;

	/* Read 2nd line.  Second line is the F stimulus. */
	Read_input_line(Stimuli,Input_string);
	if(feof(Stimuli)) break;
	Make_stimulus(Input_string, F_Set[Number_of_associations]);

	/* Read 3rd line.  Third line is the G stimulus. */
	Read_input_line(Stimuli,Input_string);
	if(feof(Stimuli)) {printf("Incomplete input\n"); break;}
	Make_stimulus(Input_string, G_Set[Number_of_associations]);

	Number_of_associations++;
    } while(!feof(Stimuli) && Number_of_associations<=Max_no_associations);
    fclose(Stimuli);
}   /* Read_stimuli */


/*-------------------------------------------------------------------------
Form the association of a pair of F and G column vectors (calculate the
outer product G F^T), and add this into the array A.
-------------------------------------------------------------------------*/
void Associate(Vector F, Vector G) {
   Outer_product(G, F, Delta_A);
   Add_matrices(A, Delta_A, A);
}   /* Associate */


 
/*-------------------------------------------------------------------------
-------------------------------------------------------------------------*/
void Learn(void) {
    int i;      /* trial number */
    int j;      /* loop index */
    int Random_pair;            /* random number: 0..Pairs_to_learn-1 */
    Vector Error_signal;        /* correction: ideal-approx output */
    Vector Output_vector;       /* approx output */

    for(i=0; i<Total_learning_trials; i++) {
	Random_pair=(int)(Random() * Pairs_to_learn); /* Chooses random pair */
	Matrix_times_vector(A, F_Set[Random_pair], Output_vector);
	for(j=0; j<Dimensionality; j++)
	    Error_signal[j]=
		(G_Set[Random_pair][j]-Output_vector[j])/Eta_prime;

        /* Form outer product of the ideal response (F_Set[]) and the */
        /* error (Error_signal), and add this to the matrix A         */
	Associate(F_Set[Random_pair], Error_signal);
	if(i%10==0) printf("%3d  Learning: %4d\n",i,Random_pair);
    }
    printf("Completion of specified learning.\n");
}   /* Learn */


/*-------------------------------------------------------------------------
-------------------------------------------------------------------------*/
void BSB_Iteration(int *pNumber_limited,
		   Vector Original_vector,      /* Starting vector. */
		   Vector V_in,                 /* Input vector. */
		   Vector V_out) {              /* Output vector. */
    int i;      /* loop index */
    float Upper_limit= 1;       /* */
    float Lower_limit=-1;       /* */
    float Alpha=0.5f;           /* Multiplies matrix feedback term */
    float Gamma=0.9f;           /* Decay term. */
    float Delta=1.0f;           /* Maintains original input. */
    Vector Matrix_output;       /* Terms in BSB expression. */

    Matrix_times_vector(A, V_in, Matrix_output);
    Scalar_times_vector(Alpha, Matrix_output, Matrix_output);

    /* Put them all together. */
    for(i=0; i<Dimensionality; i++)
	V_out[i]=Gamma*V_in[i] + Matrix_output[i] + Delta*Original_vector[i];

    (*pNumber_limited)=0;
    for(i=0; i<Dimensionality; i++) {
	if(V_out[i]>=Upper_limit) {
	    V_out[i]=Upper_limit; (*pNumber_limited)++;
	} else if(V_out[i]<=Lower_limit) {
	    V_out[i]=Lower_limit; (*pNumber_limited)++;
	}
    }
}   /* BSB_Iteration */


/*-------------------------------------------------------------------------
-------------------------------------------------------------------------*/
void Write_output(int Iteration_number,
		  int Number_limited,
		  Vector V_in) {
    String Interpretation;

    Interpret(Threshold, V_in, Interpretation);
    printf("  %3d.  ",Iteration_number);
    Write_string(Interpretation);
    printf("   Limited:%4d\n",Number_limited);
}   /* Write_output */


/*-------------------------------------------------------------------------
This routine gets the name of the file containing association strings,
and a threshold value for use by Interpret().  It calls the routine
Read_stimuli() to fill the arrays F_Set[] and G_Set[], and finally
zeros the association arrays A[][] and Delta_A[][].
-------------------------------------------------------------------------*/
void Initialize(void) {
    int i,j;            /* loop indices */
    unsigned int Seed;  /* seed for srand() */
    float Eta;          /* learning constant */

    printf("\n\nDemonstration of BSB model.  Uses Widrow-Hoff learning.\n");
    printf("Reads character vectors from a file.  Learns and recalls\n");
    printf("the pairs of associations.  Uses BSB retrieval network.\n");
    printf("See p515 et seq. of An Introduction to Neural Networks\n");
    printf("\n");

    /* Get file containing association strings, and threshold */
    do {
	printf("File name of stimuli : "); scanf("%39s",File_name);
    } while(access(File_name,4)!=0);
    Read_stimuli();     /* Read data file to F_Set[][] and G_Set[][] */

    printf("Seed for random number generator (0..65535): ");
    scanf("%u",&Seed); srand(Seed);

    do {
	printf("Threshold (0..1; about 0.5): "); scanf("%f",&Threshold);
    } while(Threshold<0 || Threshold>1);

    do {
	printf("Different pairs to learn   (1..%d): ",Number_of_associations);
	scanf("%d",&Pairs_to_learn);
    } while(Pairs_to_learn<1 && Pairs_to_learn>Number_of_associations);

    do {
	printf("Total learning trials      (1..): ");
	scanf("%d",&Total_learning_trials);
    } while(Total_learning_trials<1);

    /* This is length of a vector completely composed of +1 and -1 */
    Eta=(float)(sqrt((float)Dimensionality));
    Eta_prime=10*Eta;   /* Learning constant in Widrow-Hoff Procedure. */

    /* Normalize input vectors, and zero A[][] and Delta_A[][] */
    /*
    for(i=0; i<Number_of_associations; i++)
	for(j=0; j<Dimensionality; j++)
	    F_Set[i][j]/=Eta;
    */
    for(i=0; i<Dimensionality; i++)
	for(j=0; j<Dimensionality; j++) {
	    A[i][j]*=0;
	    Delta_A[i][j]*=0;
	}
}   /* Initialize */


/*--------------------------------------------------------------------------
Returns a float number >=0 and < 1.0
--------------------------------------------------------------------------*/
float Random(void) {
    return ((float)rand())/((float)RAND_MAX+1);
}   /* Random */



/*--------------------------------------------------------------------------
--------------------------------------------------------------------------*/
void main(void) {
    int i;                         /* loop index */
    int j;                         /* loop index */
    int Number_limited;            /* Number of vector elements at limit. */
    int Iteration_number;          /* Number of BSB Iterations. */
    int Max_number_of_iterations;  /* Maximum number of BSB iterations. */
    Vector Original_state_vector;  /* Starting vector for BSB iterations. */
    Vector Current_state_vector;   /* Current system state. */
    Vector New_state_vector;       /* Next system state after 1 BSB iteration */

    Initialize();
    Learn();    /* Widrow-Hoff Learning. */

    printf("Start of BSB procedure.\n");
    printf("Threshold: %6.2f\n",Threshold);  /* 0.5 is a good value. */

    for(j=0; j<Number_of_associations; j++) {
	/* Initialize first iteration. */
	Iteration_number=0;  Number_limited=0; Max_number_of_iterations=25;
	for(i=0; i<Dimensionality; i++) {
	    Original_state_vector[i]=F_Set[j][i];
	    Current_state_vector[i]=F_Set[j][i];
	}
	printf("\nAssociation: %2d\n",j+1);
	Write_output(Iteration_number, Number_limited, Original_state_vector);

	/* Perform iterations until finish. */
	do {
	    BSB_Iteration(&Number_limited,
			 Original_state_vector,
			 Current_state_vector,
			 New_state_vector);
	    Iteration_number++;
	    Write_output(Iteration_number, Number_limited, New_state_vector);
	    for(i=0; i<Dimensionality; i++)
		Current_state_vector[i]=New_state_vector[i];
	} while(Iteration_number<=Max_number_of_iterations &&
		Number_limited<Dimensionality);
    }
}   /* main */




#if 0
/*
--------------------------------------------------------------------
ASCII Input File of State Names and Capitals.  This is the same input
file, states.txt, as used in Appendix J.
--------------------------------------------------------------------
1234567890123456789012345
Boston      Massachusetts
Boston      Massachusetts
2
Providence   Rhode Island
Providence   Rhode Island
3
Sacramento     California
Sacramento     California
4
Bismarck     North Dakota
Bismarck     North Dakota
5
Oklahoma City    Oklahoma
Oklahoma City    Oklahoma
6
Harrisburg   Pennsylvania
Harrisburg   Pennsylvania
7
Charleston  West Virginia
Charleston  West Virginia
8
Baton Rouge     Louisiana
Baton Rouge     Louisiana
9
Concord     New Hampshire
Concord     New Hampshire
10
Columbia   South Carolina 
Columbia   South Carolina 
11
Albany           New York
Albany           New York
12
Trenton        New Jersey
Trenton        New Jersey
13
Juneau             Alaska
Juneau             Alaska
14
Talahassee        Florida
Talahassee        Florida
15
Austin              Texas
Austin              Texas
1234567890123456789012345
Boston___________________
Boston      Massachusetts
17
Providence_______________
Providence   Rhode Island
18
Sacramento_______________
Sacramento     California
19
Bismarck_________________
Bismarck     North Dakota
20
Oklahoma City____________
Oklahoma City    Oklahoma
21
Harrisburg_______________
Harrisburg   Pennsylvania
22
Charleston_______________
Charleston  West Virginia
23
Baton Rouge______________
Baton Rouge     Louisiana
24
Concord__________________
Concord     New Hampshire
25
Columbia_________________ 
Columbia   South Carolina
26
Albany___________________
Albany           New York
27
Trenton__________________
Trenton        New Jersey
28
Juneau___________________
Juneau             Alaska
29
Talahassee_______________
Talahassee        Florida
30
Austin___________________
Austin              Texas
1234567890123456789012345
____________Massachusetts
Boston      Massachusetts
32
_____________Rhode Island
Providence   Rhode Island
33
_______________California
Sacramento     California
34
_____________North Dakota
Bismarck     North Dakota
35
_________________Oklahoma
Oklahoma City    Oklahoma
36
_____________Pennsylvania
Harrisburg   Pennsylvania
37
____________West Virginia
Charleston  West Virginia
38
________________Louisiana
Baton Rouge     Louisiana
39
____________New Hampshire
Concord     New Hampshire
40
___________South Carolina 
Columbia   South Carolina 
41
_________________New York
Albany           New York
42
_______________New Jersey
Trenton        New Jersey
43
___________________Alaska
Juneau             Alaska
44
__________________Florida
Talahassee        Florida
45
____________________Texas
Austin              Texas
*/
#endif
