/* File bsb_int.c */

#include <math.h>	/* fabs */

/* Number_of_characters is related to Dimensionality of vectors */
#define Number_of_characters (int)(25)
#define Dimensionality       (int)(8*Number_of_characters)
/* String_length refers to strings in the struct Stimulus.Name */
#define String_length  (int)(60)	/* Must be > Number_of_characters */

typedef float Vector[Dimensionality];
typedef char String[String_length+1];
typedef int Byte[8];
typedef int IVector[Dimensionality];

/* Defined in bsb_int.c */
void Interpret(float Threshold,   /* Determines interpretation threshold.*/
               Vector V,                /* Input state vector. */
               String Interpretation);  /* Output interpreted string. */
static void Make_Binary_Vector(float Threshold, Vector V, IVector IV);
static int Count(Byte B);
static char Byte_Interpret(Byte B);
static void Read_IVector_as_ASCII(IVector IV, String Interpretation);


/*------------------------------------------------------------------------
                Interpret
This routine takes a vector of floats, applies a threshold to so that
each element is classifed as +1) probably set; -1) probably clear; and 0)
ambiguous.  These intermediate values are stored in int IV[].  Then this
array is turned into a character string and returned as <Interpretation>.
The string will in general contain both ascii characters that have been
identified, and the character '_' if it is still ambiguous.

This routine is used in Test_whole_set() to turn the raw association
output from Matrix_times_vector() into a string.  It is also used there
to recover the input and required output strings from F_Set[] and G_Set[].
The arrays F_Set[] and G_Set[] consist of float values that are either
zero or greater than zero, and so for this second application it is just
a matter of setting Threshold to zero and of supplying a vector from
F_Set[] or G_Set[].

Uses void Make_Binary_Vector()
     int Count()
     char Byte_Interpret()
     void Read_IVector_as_ASCII()
Arguments
    float Threshold          Value between 0 and 1
    Vector V                 array to be interpreted
Returns
    String Interpretation    string of Number_of_characters, plus null
------------------------------------------------------------------------*/
void Interpret(float Threshold,   /* Determines interpretation threshold. */
               Vector V,                /* Input state vector. */
               String Interpretation) { /* Output interpreted string. */
    IVector IV;

    Make_Binary_Vector(Threshold, V, IV);
    Read_IVector_as_ASCII(IV, Interpretation);
}   /* Interpret */

/*------------------------------------------------------------------------
Applies a threshold to the values in <V>, and writes result to <IV>.
Each value in these arrays is considered to correspond to a bit of
a character string.  If the value is >Threshold, that bit is considered
set; if it is < -Threshold, that bit is considered cleared; otherwise it
is considered ambiguous.  'Set' is indicated by +1 in IV[], 'Cleared' is
indicated by -1 in IV[], and 'Ambiguous' is indicated by 0.

This routine is also used to recover the bit pattern of the input strings
from the float arrays F_Set[] and G_Set[].  These arrays consist of values
that are either less than zero or greater than zero.  Hence Threshold
is set to zero, and the if statements below should then classify each
float value either as +1 or -1 and never 0.

To make the behaviour when Threshold=0 clearer, I have changed the
inequality below from >= to >.

Used by Interpret()
Arguments
    float Threshold          Value between 0 and 1
    Vector V                 array to be interpreted
Returns
    IVector IV               result of applying threshold: array of -1,0 +1
------------------------------------------------------------------------*/
static void Make_Binary_Vector(float Threshold, Vector V, IVector IV) {
    int j;      /* loop index */

    for(j=0; j<Dimensionality; j++) {
        IV[j]=0;
        if(V[j] >   (float)fabs(Threshold)) IV[j]= 1;
        if(V[j] <= -(float)fabs(Threshold)) IV[j]=-1;
    }
}   /* Make_Binary_Vector */


/*------------------------------------------------------------------------
Counts the number of ambiguous bits (B[i]==0) indicated by the array <B>.
Used by Interpret()
------------------------------------------------------------------------*/
static int Count(Byte B) {
    int i;      /* loop index */
    int sum;    /* number of zeros in byte */

    sum=0;
    for(i=0; i<7; i++)          /* Ignore parity bit. */
        if(B[i]==0) sum++;
    return sum;
}   /* Count */

/*------------------------------------------------------------------------
B is an array of eight integers each equal to +1, 0 or -1.  This routine
converts it to an ascii character.  If a group of 8 contains any 0 values,
the corresponding character is considered ambiguous and a '_' results.
Otherwise +1 causes the corresponding bit to be set, and -1 causes the
corresponding bit to be cleared.

Used by Interpret()
------------------------------------------------------------------------*/
static char Byte_Interpret(Byte B) {
    int i;      /* loop index */
    int Charnr; /* ASCII code of character, as an integer */
    Byte C;     /* copy of B */

    /* First check to see if all byte elements + 1 or - 1. */
    if(Count(B)>0)
        return '_';     /* character is still ambiguous */
    else {
        /* Convert B[] to an byte of zero and +1. */
        for(i=0; i<8; i++)
            C[i]=(B[i]==-1)? 0: 1;
        Charnr=64*C[6] + 32*C[5] + 16*C[4] + 8*C[3] + 4*C[2] + 2*C[1] + C[0];
        if(Charnr==0 || Charnr==127)    /* Zero/delete convention is '_' */
            return '_';
        else if(Charnr<32)              /* Non-printing character. */
            return '#';
        else                            /* Print character. */
            return (char)Charnr;
    }
}   /* Byte_Interpret */


/*------------------------------------------------------------------------
Takes the vector <IV>, consisting of 8*Number_of_characters integers, and
converts it to an ascii string Interpretation[].  The string is padded
with blanks to its full length of <String_length> characters, and then
null terminated.

Used by Interpret()
Uses Byte_Interpret()
Arguments
    IVector IV               result of applying threshold: array of -1,0 +1
Returns
    String Interpretation    ascii string corresponding to IV
------------------------------------------------------------------------*/
static void Read_IVector_as_ASCII(IVector IV, String Interpretation) {
    int i,j;    /* loop indices */
    Byte B;     /* eight consecutive values copied from IV[] */

    /* Check for inappropriate defines */
    if(Number_of_characters>String_length) return;

    for(j=0; j<Number_of_characters; j++) {
        /* Next 'for loop' flips order of vector elements. */
        for(i=0; i<8; i++)
            B[7-i]=IV[j*8+i];
        Interpretation[j]=Byte_Interpret(B);
    }
    for(; j<String_length; j++) 
        Interpretation[j]=' ';
    Interpretation[String_length]='\0';  /* Add null termination */
}   /* Read_IVector_as_ASCII */

