/* File bsb_make.c */

/*------------------------------------------------------------------------
This long module makes stimulus files from character strings.
It can also do some element by element manipulation and can add
stimuli together, as well as allowing inspection of single elements
in the vectors.  These facilities are rarely used, but are provided
anyway.  The various options available in this procedure are
discussed at length in the Manual.
------------------------------------------------------------------------*/
#include <stdio.h>      /* fgets fputc getchar printf scanf */
#include <string.h>     /* memcpy strncpy strlen */

/* Number_of_characters is related to Dimensionality of vectors */
#define Number_of_characters (int)(25)
#define Dimensionality       (int)(8*Number_of_characters)
/* Maximum_set_size is the size of each training set */
#define Maximum_set_size     (int)(100)
/* String_length refers to strings in the struct Stimulus.Name */
#define String_length   (int)(60)       /* Must be > Number_of_characters */

typedef float Vector[Dimensionality];
typedef char String[String_length+1];
typedef struct {
    String Name;
    Vector Val;
} Stimulus;
typedef int Byte[8];

/* defined in bsb.c */
/* Element [0] does not appear to be used anywhere */
Stimulus F_set[Maximum_set_size+1];
Stimulus G_set[Maximum_set_size+1];
Stimulus T_set[Maximum_set_size+1];
int FG_mode;
float Threshold;

/* Defined in bsb_int.c */
void Interpret(float Threshold,   /* Determines interpretation threshold.*/
               Vector V,                /* Input state vector. */
               String Interpretation);  /* Output interpreted string. */

/* Defined in bsb_list.c */
void List_test_set(void);
                
/* Defined in bsb_make.c */
static Stimulus Stim;
static String Template;
static void Write_box(char Stimulus_set, int Stimulus_number);
static void Display_stimulus(char Stimulus_set, int Stimulus_number);
static void New_value(char Stimulus_set, int Stimulus_number);
static void Choose_stimulus(char Stimulus_set, int Stimulus_number);
void Showvals(void);
static void Zero_byte(Byte B);
static void W(String S);
static void Blank_string(String S);
static void Zero_stimulus(Stimulus St);
static void Nr_to_byte(int Nr, Byte B);
static void Add_Byte_to_Vector(Vector V, int i, Byte B);
static void Edit_stimulus(void);
static void Replace(void);
void Init_template(void);
static void Make_template(void);
static void Add_stimuli(Stimulus *S1, Stimulus *S2, Stimulus *Sum);
static void Sum_stimuli(void);
static void Change_stimulus_name(void);
static void Copy_stimulus(void);
void Make_stimulus(void);

/* Defined in bsb_vt100.c */
void Clear_box(void);
void Scroll_box(void);
void Clear_block(void);
void Erase_line(void);
void Scroll_block(void);
void Save_cursor(void);
void Restore_cursor(void);
void Bottom_line(void);

/* Defined in bsb_writef.c */
void Write_file(char Which_file);

/*------------------------------------------------------------------------
------------------------------------------------------------------------*/
static void Write_box(char Stimulus_set, int Stimulus_number) {
    Clear_box();
    printf("Threshold       : %8.3f\n",Threshold);
    printf("Current Stimulus: %c %2d\n",Stimulus_set,Stimulus_number);
}   /* Write_box */

/*------------------------------------------------------------------------
------------------------------------------------------------------------*/
static void Display_stimulus(char Stimulus_set, int Stimulus_number) { 
    int i;		/* loop index, 0..Dimensionality-1 */
    int Row_nr;		/* line number of list, 0.. */
    const int Elements_per_line=8;	/* one byte; stim elements per line */
    String Stim_interpretation;	/* char version of stimulus */

    Interpret(Threshold, Stim.Val, Stim_interpretation);

    Write_box(Stimulus_set,Stimulus_number);
    printf("            Name: %s\n", Stim.Name);
    printf("  Interpretation: %s\n", Stim_interpretation);
    Clear_block();
    Scroll_block();
    for(i=0; i<Dimensionality; i++) {
        if(i%Elements_per_line==0) {
            Row_nr=i/Elements_per_line;
            printf("%4d %c  ",i,Stim_interpretation[Row_nr]);
        }
        printf("%8.4f",Stim.Val[i]);
        if(i%Elements_per_line==Elements_per_line-1)
            printf("%4d\n",i); 
    }
}   /* Display_stimulus */


/*------------------------------------------------------------------------
------------------------------------------------------------------------*/
static void New_value(char Stimulus_set, int Stimulus_number) { 
    int Element_number;		/* 0..Dimensionality-1 */
    float New_element_value;	/* Input value */

    Write_box(Stimulus_set,Stimulus_number);
    do {
        printf("Which element (0..%d) : ",Dimensionality-1);
        scanf("%d",&Element_number);
        getchar();
    } while(Element_number<0 || Element_number>=Dimensionality);
    printf("Old Value       : %f8.4\n", Stim.Val[Element_number]);
    printf("New Value       : "); scanf("%f",&New_element_value); 
    getchar();

    Stim.Val[Element_number]=New_element_value;
    if(Stimulus_set=='F' || Stimulus_set=='f') 
        F_set[Stimulus_number]=Stim;
    else if(Stimulus_set=='G' || Stimulus_set=='g')
        G_set[Stimulus_number]=Stim;
    else if(Stimulus_set=='T' || Stimulus_set=='t')
        T_set[Stimulus_number]=Stim;
    else
        printf("\nNew_value(): Illegal value of Stimulus_set\n");
}   /* New_value */


/*------------------------------------------------------------------------
------------------------------------------------------------------------*/
static void Choose_stimulus(char Stimulus_set, int Stimulus_number) { 
    int Stimulus_set_OK;	/* Set only if == 'f', 'g' or 't' */
    int Stimulus_number_OK;	/* Set only if == 1..Maximum_set_size */

    Write_box(Stimulus_set,Stimulus_number);
    do {
        printf("New stimulus & number (F|G|T 1..%d) : ",Maximum_set_size);
        scanf("%c %d",&Stimulus_set,&Stimulus_number);
        getchar();
        Stimulus_set_OK=(Stimulus_set=='F' || Stimulus_set=='f' ||
                         Stimulus_set=='G' || Stimulus_set=='g' ||
                         Stimulus_set=='T' || Stimulus_set=='t');
        Stimulus_number_OK=(Stimulus_number>=1 && 
                            Stimulus_number<=Maximum_set_size);
    } while(!Stimulus_set_OK && !Stimulus_number_OK);

    Write_box(Stimulus_set,Stimulus_number);
    if(Stimulus_set=='F' || Stimulus_set=='f') 
        Stim=F_set[Stimulus_number];
    else if(Stimulus_set=='G' || Stimulus_set=='g')
        Stim=G_set[Stimulus_number];
    else if(Stimulus_set=='T' || Stimulus_set=='t')
        Stim=T_set[Stimulus_number];
    else
        printf("\nChoose_stimulus(): Illegal value of Stimulus_set\n");
}   /* Choose_stimulus */


/*------------------------------------------------------------------------
------------------------------------------------------------------------*/
void Showvals(void) {
    char Stimulus_set='F';
    int Stimulus_number=1;
    char Input_char;

    Save_cursor();
    Stim=F_set[Stimulus_number];
    Write_box(Stimulus_set,Stimulus_number);
    
    while(1) {
        Bottom_line();
        printf("%s Shows Values  ",FG_mode? "FG": "TF");
        printf("C)hoose D)ispl L)ist M)ode Q)uit T)hresh X)chng > ");
        scanf("%c",&Input_char);
        getchar();
        if(Input_char=='C' || Input_char=='c')
            Choose_stimulus(Stimulus_set,Stimulus_number);
        else if(Input_char=='D' || Input_char=='d')
            Display_stimulus(Stimulus_set, Stimulus_number);
        else if(Input_char=='L' || Input_char=='l')
            List_test_set();
        else if(Input_char=='M' || Input_char=='m')
            FG_mode=!FG_mode;
        else if(Input_char=='Q' || Input_char=='q')
            break;
        else if(Input_char=='T' || Input_char=='t') {
            Bottom_line();
            printf("New Threshold : "); scanf("%f",&Threshold);
            getchar();
            Write_box(Stimulus_set, Stimulus_number);
        } else if(Input_char=='X' || Input_char=='x')
            New_value(Stimulus_set, Stimulus_number);
    }   
    Restore_cursor();
}   /* Showvals */



/*------------------------------------------------------------------------
Sets the array of 8 integers to zero.
------------------------------------------------------------------------*/
static void Zero_byte(Byte B) {
    int i;      /* loop index */

    for(i=0; i<8; i++)
        B[i]=0;
}   /* Zero_byte */



/*------------------------------------------------------------------------
Prints the string S.  Just <Nr_of_characters> characters are printed,
whereas a string can in principle contain as many as <String_length>
characters.  (That is why I don't use printf("%s",S);)
------------------------------------------------------------------------*/
static void W(String S) {
    int i;		/* loop index */

    for(i=0; i<Number_of_characters; i++)
        fputc(S[i],stdout);
}   /* W */

/*------------------------------------------------------------------------
Set all <String_length> characters of S to ' ', and null terminate it.
------------------------------------------------------------------------*/
static void Blank_string(String S) {
    int i;		/* loop index */

    for(i=0; i<String_length; i++)
        S[i]=' ';
    S[String_length]='\0';	/* null termination */
}   /* Blank_string */


/*------------------------------------------------------------------------
------------------------------------------------------------------------*/
static void Zero_stimulus(Stimulus St) {
    int i;		/* loop index */

    Blank_string(St.Name);
    for(i=0; i<Dimensionality; i++)
        St.Val[i]=0;
}   /* Zero_stimulus */

/*------------------------------------------------------------------------
Copies each of the seven least significant bits of <Nr> to the elements
of B[].
------------------------------------------------------------------------*/
static void Nr_to_byte(int Nr, Byte B) {
    int i;              /* loop index */
    int Parity=0;       /* sum of all bits */
    int mask;           /* bit mask */

    if(Nr>=128) {
        printf("Too large. ASCII conversion fails.\n"); return;
    } else if(Nr<0) {
        printf("Negative number. ASCII conversion fails.\n"); return;
    }

    /* Copy bits 6..0 of Nr to the corresponding elements of B[] */
    for(mask=0x40,i=6; i>=0; mask>>=1,i--) {
        if(Nr&mask) {
            B[i]=1; Parity++;
        } else
            B[i]=-1;
    }
    /* Set bit 7 such that the sum of all bits mod 2 is equal to 0 */
    B[7]=(Parity&1)? 1: -1;
}   /* Nr_to_byte */

/*------------------------------------------------------------------------
Copies 8 integers from B to sequential elements of V, starting from
element 8*i.
------------------------------------------------------------------------*/
static void Add_Byte_to_Vector(Vector V, int i, Byte B) {
    int j;      /* loop index */

    if(i<0 || i>=Number_of_characters)
        printf("Add_Byte_to_Vector(): i=%d is out of range\n",i);
    else
        for(j=0; j<8; j++)
            V[i*8+j]=B[7-j];
}   /* Add_Byte_to_Vector */

 

/*------------------------------------------------------------------------
Replace a selected stimulus, displaying the existing string.  It first 
asks the question
    Edit which stimulus (set, number) :
to which the user responds with something like `G 10'.  Then a
new string may be entered.
------------------------------------------------------------------------*/
static void Edit_stimulus(void) {
    char Read_string[132];	/* stimulus string: just 25 chars used */
    String Input_string;	/* copy of Read_string limited to 60 chars */
    Stimulus New_stimulus;	/* new stimulus */
    Stimulus Old_stimulus;	/* old stimulus */
    char Stimulus_set;		/* 'f', 'g' or 't' */
    int Stimulus_number;	/* stimulus number to edit */
    int Stimulus_set_OK;	/* set only if == 'f', 'g' or 't' */
    int Stimulus_number_OK;	/* set only if == 1..Maximum_set_size */
    int i;			/* loop index */
    Byte B;			/* byte version of character */

    Bottom_line();
    Erase_line();
    do {
        printf("Edit which stimulus (set, number) : ");
        scanf("%c %d",&Stimulus_set, &Stimulus_number);
        getchar();
        Stimulus_set_OK=(Stimulus_set=='F' || Stimulus_set=='f' ||
                         Stimulus_set=='G' || Stimulus_set=='g' ||
                         Stimulus_set=='T' || Stimulus_set=='t');
        Stimulus_number_OK=(Stimulus_number>=1 && 
                            Stimulus_number<=Maximum_set_size);
    } while(!Stimulus_set_OK && !Stimulus_number_OK);

    if(Stimulus_set=='F' || Stimulus_set=='f') 
        Old_stimulus=F_set[Stimulus_number];
    else if(Stimulus_set=='G' || Stimulus_set=='g')
        Old_stimulus=G_set[Stimulus_number];
    else if(Stimulus_set=='T' || Stimulus_set=='t')
        Old_stimulus=T_set[Stimulus_number];
    else
        printf("\nEdit_stimulus(): Illegal value of Stimulus_set\n");

    /* Get new stimulus string: Read_string */
    Clear_box();
    Scroll_box();
    Blank_string(New_stimulus.Name);
    printf("Edit    : %c %d\n",Stimulus_set,Stimulus_number);
    printf("Old     : %s\n",Old_stimulus.Name);
    printf("New     : "); fgets(Read_string,sizeof(Read_string),stdin);
    *strchr(Read_string,'\n')='\0';

    /* Copy Read_string to a fixed-length string, '_' padded if necessary */
    strncpy(Input_string,Read_string,Number_of_characters);
    Input_string[Number_of_characters]='\0';
    for(i=strlen(Input_string); i<Number_of_characters; i++) 
        Input_string[i]='_';

    /* Copy Input_string to New_stimulus.Name and New_stimulus.Val */
    /* When (as is typical) String_length > Number_of_characters the */
    /* following will copy with null padding, but still works if */
    /* String_length <= Number_of_characters. */
    strncpy(New_stimulus.Name,Input_string,String_length);
    New_stimulus.Name[String_length]='\0';
    for(i=0; i<Number_of_characters; i++) {
        if(Input_string[i]=='_') Zero_byte(B);
        else Nr_to_byte(Input_string[i],B);
        Add_Byte_to_Vector(New_stimulus.Val,i,B);
    }

    /* Copy New_stimulus to F_set[], G_set[] or T_set[] */
    if(Stimulus_set=='F' || Stimulus_set=='f') 
        F_set[Stimulus_number]=New_stimulus;
    else if(Stimulus_set=='G' || Stimulus_set=='g')
        G_set[Stimulus_number]=New_stimulus;
    else if(Stimulus_set=='T' || Stimulus_set=='t')
        T_set[Stimulus_number]=New_stimulus;
    else
        printf("\nEdit_stimulus(): Illegal value of Stimulus_set\n");

}   /* Edit_stimulus */


/*------------------------------------------------------------------------
Replace a selected stimulus, displaying the template string.  It first 
asks the question
    Replace which stimulus (set, number) :
to which the user responds with something like `G 10'.  Then a
new string may be entered.
------------------------------------------------------------------------*/
static void Replace(void) {
    char Read_string[132];	/* stimulus string: just 25 chars used */
    String Input_string;	/* copy of Read_string limited to 60 chars */
    Stimulus New_stimulus;	/* new stimulus */
    char Stimulus_set;		/* 'f', 'g' or 't' */
    int Stimulus_number;	/* stimulus number to edit */
    int Stimulus_set_OK;	/* set only if == 'f', 'g' or 't' */
    int Stimulus_number_OK;	/* set only if == 1..Maximum_set_size */
    int i;			/* loop index */
    Byte B;			/* byte version of character */

    Bottom_line();
    Erase_line();
    do {
        printf("Replace which stimulus (set, number) : ");
        scanf("%c %d",&Stimulus_set, &Stimulus_number);
        getchar();	/* remove newline from input buffer */
        Stimulus_set_OK=(Stimulus_set=='F' || Stimulus_set=='f' ||
                         Stimulus_set=='G' || Stimulus_set=='g' ||
                         Stimulus_set=='T' || Stimulus_set=='t');
        Stimulus_number_OK=(Stimulus_number>=1 && 
                            Stimulus_number<=Maximum_set_size);
    } while(!Stimulus_set_OK && !Stimulus_number_OK);

    /* Get new stimulus string: Read_string */
    Clear_box();
    Scroll_box();
    Blank_string(New_stimulus.Name);
    printf("Replace : %c %d\n",Stimulus_set,Stimulus_number);
    printf("Template: %s\n",Template);
    printf("New     : "); fgets(Read_string,sizeof(Read_string),stdin);
    *strchr(Read_string,'\n')='\0';

    /* Copy Read_string to a fixed-length string, '_' padded if necessary */
    strncpy(Input_string,Read_string,Number_of_characters);
    Input_string[Number_of_characters]='\0';
    for(i=strlen(Input_string); i<Number_of_characters; i++) 
        Input_string[i]='_';

    /* Copy Input_string to New_stimulus.Name and New_stimulus.Val */
    /* When (as is typical) String_length > Number_of_characters the */
    /* following will copy with null padding, but still works if */
    /* String_length <= Number_of_characters. */
    strncpy(New_stimulus.Name,Input_string,String_length);
    New_stimulus.Name[String_length]='\0';
    for(i=0; i<Number_of_characters; i++) {
        if(Input_string[i]=='_') Zero_byte(B);
        else Nr_to_byte(Input_string[i],B);
        Add_Byte_to_Vector(New_stimulus.Val,i,B);
    }

    /* Copy New_stimulus to F_set[], G_set[] or T_set[] */
    if(Stimulus_set=='F' || Stimulus_set=='f') 
        F_set[Stimulus_number]=New_stimulus;
    else if(Stimulus_set=='G' || Stimulus_set=='g')
        G_set[Stimulus_number]=New_stimulus;
    else if(Stimulus_set=='T' || Stimulus_set=='t')
        T_set[Stimulus_number]=New_stimulus;
    else
        printf("\nEdit_stimulus(): Illegal value of Stimulus_set\n");
}   /* Replace */


/*------------------------------------------------------------------------
Initialize Template to "1234567890123456789012345" (25 characters)
Other templates can be constructed using Make_template().
------------------------------------------------------------------------*/
void Init_template(void) {
    int i;	/* loop index */
 
    for(i=0; i<Number_of_characters; i++)
        Template[i]='0'+(i+1)%10;
    Template[i]='\0';
}   /* Init_template */

/*------------------------------------------------------------------------
------------------------------------------------------------------------*/
static void Make_template(void) {
    char InputLine[80];		/* Keyboard input */
    char Model_char;		/* Y/N */
    String Read_string;		/* template string: just 25 chars used */
    char Stimulus_set;		/* 'f', 'g' or 't' */
    int Stimulus_number;	/* stimulus number to edit */
    int Stimulus_set_OK;	/* set only if == 'f', 'g' or 't' */
    int Stimulus_number_OK;	/* set only if == 1..Maximum_set_size */
    int i;			/* loop index */

    Clear_box();
    Scroll_box();
    printf("New stimulus template.  Use Model? Y/N : ");
    fgets(InputLine,sizeof(InputLine),stdin);
    Model_char=InputLine[0];

    /* Display either existing stimulus or default template */
    if(Model_char=='Y' || Model_char=='y') {
        do {
            printf("Which stimulus as model (set, number) : ");
            scanf("%c %d",&Stimulus_set, &Stimulus_number);
            getchar();
            Stimulus_set_OK=(Stimulus_set=='F' || Stimulus_set=='f' ||
                             Stimulus_set=='G' || Stimulus_set=='g' ||
                             Stimulus_set=='T' || Stimulus_set=='t');
            Stimulus_number_OK=(Stimulus_number>=1 && 
                                Stimulus_number<=Maximum_set_size);
        } while(!Stimulus_set_OK && !Stimulus_number_OK);

        /* Print selected stimulus from F_set[], G_set[] or T_set[] */
        printf("          ");
        if(Stimulus_set=='F' || Stimulus_set=='f') 
            W(F_set[Stimulus_number].Name);
        else if(Stimulus_set=='G' || Stimulus_set=='g')
            W(G_set[Stimulus_number].Name);
        else if(Stimulus_set=='T' || Stimulus_set=='t')
            W(T_set[Stimulus_number].Name);
        else
            printf("\nMake_template(): Illegal value of Stimulus_set\n");
    } else {
        printf("          ");
        for(i=0; i<Number_of_characters; i++)
            printf("%c",(char)('0'+(i+1)%10));
    }

    /* Get new template */
    printf("\nTemplate: ");
    fgets(Read_string,sizeof(Read_string),stdin);
    *strchr(Read_string,'\n')='\0';

    /* Copy Read_string to a fixed-length string, '_' padded if necessary */
    strncpy(Template,Read_string,Number_of_characters);
    Template[Number_of_characters]='\0';
    for(i=strlen(Template); i<Number_of_characters; i++) 
        Template[i]='_';
}   /* Make_template */

/*------------------------------------------------------------------------
------------------------------------------------------------------------*/
static void Add_stimuli(Stimulus *S1, Stimulus *S2, Stimulus *Sum) {
    int i;	/* loop index */

    for(i=0; i<Dimensionality; i++)
        Sum->Val[i] = S1->Val[i] + S2->Val[i];
}   /* Add_stimuli */


/*------------------------------------------------------------------------
Sometimes it is convenient to sum stimuli together.
------------------------------------------------------------------------*/
static void Sum_stimuli(void) {
    int i;		/* loop index */
    Stimulus Sum;	/* Sum of stimuli */
    int How_many_stimuli;
    char Stimulus_set;		/* 'f', 'g' or 't' */
    int Stimulus_number;	/* stimulus number to edit */
    int Stimulus_set_OK;	/* set only if == 'f', 'g' or 't' */
    int Stimulus_number_OK;	/* set only if == 1..Maximum_set_size */

    Zero_stimulus(Sum);
    memcpy(Sum.Name,"Sum ",4);

    /* Get how many to sum */
    Bottom_line();
    Erase_line();
    do {
        printf("How many stimuli? (Averages sum): ");
        scanf("%d",&How_many_stimuli);
        getchar();
    } while(How_many_stimuli<1 || How_many_stimuli>Maximum_set_size);

    /* Add selected stimuli */
    for(i=1; i<=How_many_stimuli; i++) {
        Erase_line();
        do {
            printf("Add stimulus number %d of %d stimuli (set, number) : ",
                i,How_many_stimuli);
            scanf("%c %d",&Stimulus_set, &Stimulus_number);
            getchar();
            Stimulus_set_OK=(Stimulus_set=='F' || Stimulus_set=='f' ||
                             Stimulus_set=='G' || Stimulus_set=='g' ||
                             Stimulus_set=='T' || Stimulus_set=='t');
            Stimulus_number_OK=(Stimulus_number>=1 && 
                                Stimulus_number<=Maximum_set_size);
        } while(!Stimulus_set_OK && !Stimulus_number_OK);

        /* Add to Sum.Val */
        if(Stimulus_set=='F' || Stimulus_set=='f') 
            Add_stimuli(&Sum, &F_set[Stimulus_number], &Sum);
        else if(Stimulus_set=='G' || Stimulus_set=='g')
            Add_stimuli(&Sum, &G_set[Stimulus_number], &Sum);
        else if(Stimulus_set=='T' || Stimulus_set=='t')
            Add_stimuli(&Sum, &T_set[Stimulus_number], &Sum);
        else
            printf("\nSum_stimuli(): Illegal value of Stimulus_set\n");
       
        /* Augment Sum.Name */
        if(i<14) {
             Sum.Name[4*i+0]=Stimulus_set; 
             Sum.Name[4*i+1]=(char)('0'+Stimulus_number/10); 
             Sum.Name[4*i+2]=(char)('0'+Stimulus_number%10); 
             Sum.Name[4*i+3]=' '; 
        }
    }

    /* Normalize sum */
    for(i=0; i<Dimensionality; i++)
        Sum.Val[i] /= How_many_stimuli;

    /* Get destination for sum */
    Erase_line();
    do {
        printf("Copy to   (set, number) : ");
        scanf("%c %d",&Stimulus_set, &Stimulus_number);
        getchar();
        Stimulus_set_OK=(Stimulus_set=='F' || Stimulus_set=='f' ||
                         Stimulus_set=='G' || Stimulus_set=='g' ||
                         Stimulus_set=='T' || Stimulus_set=='t');
        Stimulus_number_OK=(Stimulus_number>=1 && 
                            Stimulus_number<=Maximum_set_size);
    } while(!Stimulus_set_OK && !Stimulus_number_OK);

    /* Write Sum to specified destination */
    Erase_line();
    printf("Name of sum: %s\n",Sum.Name);
    if(Stimulus_set=='F' || Stimulus_set=='f') 
        F_set[Stimulus_number]=Sum;
    else if(Stimulus_set=='G' || Stimulus_set=='g')
        G_set[Stimulus_number]=Sum;
    else if(Stimulus_set=='T' || Stimulus_set=='t')
        T_set[Stimulus_number]=Sum;
    else
        printf("\nSum_stimuli(): Illegal value of Stimulus_set\n");
}   /* Sum_stimuli */


/*------------------------------------------------------------------------
------------------------------------------------------------------------*/
static void Change_stimulus_name(void) {
    char Stimulus_set;		/* 'f', 'g' or 't' */
    int Stimulus_number;	/* stimulus number to edit */
    int Stimulus_set_OK;	/* set only if == 'f', 'g' or 't' */
    int Stimulus_number_OK;	/* set only if == 1..Maximum_set_size */
    String Old_name;		/* copy of original Stim.Name */
    String New_name;		/* replacement for old name */

    Bottom_line();
    Erase_line();
    printf("Change the name of a stimulus.\n");
    do {
        printf("Change name of (set, number) : ");
        scanf("%c %d",&Stimulus_set, &Stimulus_number);
        getchar();
        Stimulus_set_OK=(Stimulus_set=='F' || Stimulus_set=='f' ||
                         Stimulus_set=='G' || Stimulus_set=='g' ||
                         Stimulus_set=='T' || Stimulus_set=='t');
        Stimulus_number_OK=(Stimulus_number>=1 && 
                            Stimulus_number<=Maximum_set_size);
    } while(!Stimulus_set_OK && !Stimulus_number_OK);

    /* Get copy of old name */
    if(Stimulus_set=='F' || Stimulus_set=='f') 
        strncpy(Old_name,F_set[Stimulus_number].Name,String_length+1);
    else if(Stimulus_set=='G' || Stimulus_set=='g')
        strncpy(Old_name,G_set[Stimulus_number].Name,String_length+1);
    else if(Stimulus_set=='T' || Stimulus_set=='t')
        strncpy(Old_name,T_set[Stimulus_number].Name,String_length+1);
    else
        printf("\nChange_stimulus_name(): Illegal value of Stimulus_set\n");

    /* Get new name */
    Clear_box();
    Scroll_box();
    printf("Template: %s\n",Template);
    printf("Old name: %s\n",Old_name);
    printf("New name: "); fgets(New_name,sizeof(New_name),stdin);
    *strchr(New_name,'\n')='\0';

    /* Write new name */
    if(Stimulus_set=='F' || Stimulus_set=='f') 
        strncpy(F_set[Stimulus_number].Name,New_name,String_length+1);
    else if(Stimulus_set=='G' || Stimulus_set=='g')
        strncpy(G_set[Stimulus_number].Name,New_name,String_length+1);
    else if(Stimulus_set=='T' || Stimulus_set=='t')
        strncpy(T_set[Stimulus_number].Name,New_name,String_length+1);
    else
        printf("\nChange_stimulus_name(): Illegal value of Stimulus_set\n");
}   /* Change_stimulus_name */

/*------------------------------------------------------------------------
------------------------------------------------------------------------*/
static void Copy_stimulus(void) {
    char Stimulus_set;		/* 'f', 'g' or 't' */
    int Stimulus_number;	/* stimulus number to copy */
    int Stimulus_set_OK;	/* set only if == 'f', 'g' or 't' */
    int Stimulus_number_OK;	/* set only if == 1..Maximum_set_size */
    char Source_set;		/* 'f', 'g' or 't' */
    int Source_number;		/* stimulus number to copy */
    Stimulus Stim;		/* Copy of copied stimulus */
    
    Bottom_line();
    Erase_line();
    printf("Copy a stimulus.\n");
    do {
        printf("Copy from (set, number) : ");
        scanf("%c %d",&Stimulus_set, &Stimulus_number);
        getchar();
        Stimulus_set_OK=(Stimulus_set=='F' || Stimulus_set=='f' ||
                         Stimulus_set=='G' || Stimulus_set=='g' ||
                         Stimulus_set=='T' || Stimulus_set=='t');
        Stimulus_number_OK=(Stimulus_number>=1 && 
                            Stimulus_number<=Maximum_set_size);
    } while(!Stimulus_set_OK && !Stimulus_number_OK);
    Source_set=Stimulus_set; Source_number=Stimulus_number;

    /* Copy from specified source */
    if(Stimulus_set=='F' || Stimulus_set=='f') 
        Stim=F_set[Stimulus_number];
    else if(Stimulus_set=='G' || Stimulus_set=='g')
        Stim=G_set[Stimulus_number];
    else if(Stimulus_set=='T' || Stimulus_set=='t')
        Stim=T_set[Stimulus_number];
    else
        printf("\nCopy_stimulus(): Illegal value of Stimulus_set\n");

    /* Get destimation */
    do {
        printf("Copy from %c %d. ",Source_set,Source_number);
        printf("Copy to (set, number) : ");
        scanf("%c %d",&Stimulus_set, &Stimulus_number);
        getchar();
        Stimulus_set_OK=(Stimulus_set=='F' || Stimulus_set=='f' ||
                         Stimulus_set=='G' || Stimulus_set=='g' ||
                         Stimulus_set=='T' || Stimulus_set=='t');
        Stimulus_number_OK=(Stimulus_number>=1 && 
                            Stimulus_number<=Maximum_set_size);
    } while(!Stimulus_set_OK && !Stimulus_number_OK);

    /* Copy to specified destination */
    if(Stimulus_set=='F' || Stimulus_set=='f') 
        F_set[Stimulus_number]=Stim;
    else if(Stimulus_set=='G' || Stimulus_set=='g')
        G_set[Stimulus_number]=Stim;
    else if(Stimulus_set=='T' || Stimulus_set=='t')
        T_set[Stimulus_number]=Stim;
    else
        printf("\nCopy_stimulus(): Illegal value of Stimulus_set\n");
}   /* Copy_stimulus */

/*------------------------------------------------------------------------
------------------------------------------------------------------------*/
void Make_stimulus(void) {
    char Which_file;	/* F, G or T */
    char Command;	/* a, c, e, i, m, n, l, q, r,s or t */
    char InputLine[80];

    Clear_box();
    Scroll_box();

    while(1) {
        Bottom_line();
        printf("%s ",FG_mode? "FG": "T ");
        printf("A)dd C)opy E)dit I)ndiv. L)ist M)ode ");
        printf("N)ame Q)uit R)epl. S)ave T)empl. > ");
        fgets(InputLine,sizeof(InputLine),stdin);
        Command=InputLine[0];

        Bottom_line();
        Erase_line();
        if(Command=='A' || Command=='a')
            Sum_stimuli();
        else if(Command=='C' || Command=='c')
            Copy_stimulus();
        else if(Command=='E' || Command=='e')
            Edit_stimulus();
        else if(Command=='I' || Command=='i')
            Showvals();
        else if(Command=='M' || Command=='m') {
            printf("Toggle Mode"); FG_mode=!FG_mode;
        } else if(Command=='N' || Command=='n')
            Change_stimulus_name();
        else if(Command=='L' || Command=='l')
            List_test_set();
        else if(Command=='Q' || Command=='q')
            break;
        else if(Command=='R' || Command=='r')
            Replace();
        else if(Command=='S' || Command=='s') {
            Bottom_line();
            printf("Write to which file F, G or T? ");
            scanf("%c",&Which_file);
            getchar();
            Write_file(Which_file);
        } else if(Command=='T' || Command=='t')
            Make_template();
    }
}   /* Make_stimulus */

