/* bsb_readnf.c */

/*---------------------------------------------------------------------
Reads neuron file in from disk.  These files are composed of structures of
type struct Neuron.    If the neurons are partially connected, then some
synapses need not be used.  The files store data as connection lists,
so a partially connected system will compute faster than one that is
fully connected.  This is a useful technique on a machine which does
not vectorize.  If these programs are run on many supercomputer or a
machine with an array processor, more traditional matrix storage would
be faster, in spite of a number of zero entries.

Our experience has been that partial connectivity usually works
better than full connectivity.  In a fully connected system, every
element sees the same enviroment, which seems to lead to problems
with metastable and symmetric computational states.
---------------------------------------------------------------------*/
#include <stdio.h>	/* fopen fclose fread printf */

/* Number_of_characters is related to Dimensionality of vectors */
#define Number_of_characters (int)(25)
#define Dimensionality       (int)(8*Number_of_characters)

typedef struct {
    float Strength;
    int From;
} Synapse;
typedef struct {
    float Activity;
    Synapse Synapses[Dimensionality];
    float Neuron_output;
    float Learning;
    float Neuron_upper_limit;
    float Neuron_lower_limit;
} Neuron;

/* Defined in bsb.c */
Neuron Neurons[Dimensionality];
int Neuron_file_present;
extern char NFileName[];

/* Defined in bsb_readnf.c */
int Number_of_synapses[Dimensionality];
int Avg_number_of_synapses;
static int Synapse_number(Neuron *N);
static void Read_Nfile(void);
void Read_neuron_file(void);

/* Defined in bsb_vt100.c */
void Erase_line(void);
void Scroll_block(void);
void Bottom_line(void);
void Restore_scrolling(void);

#if 0
/*---------------------------------------------------------------------
This routine is no longer required
---------------------------------------------------------------------*/
static int Found_Zero(Neuron *N) {
    int i;	/* loop index */

    for(i=0; i<Dimensionality; i++)
        if(N->Synapses[i].From==-1) return 1;
    return 0;
}   /* Found_Zero */
#endif


/*---------------------------------------------------------------------
Returns the index of the synapse preceding the first -1, which
signifies the end of the list.  Possible values returned are 0 (for 
a neuron with no connections) to Dimensionality (for a fully connected 
neuron).
---------------------------------------------------------------------*/
static int Synapse_number(Neuron *N) {
    int i;	/* loop index */

    for(i=0; i<Dimensionality; i++)
        if(N->Synapses[i].From==-1) return i;
    return i;
}   /* Synapse_number */

    
/*---------------------------------------------------------------------
There is a possible bug in the pascal version of this routine, in that 
*all* loops run from 1 to Dimensionality, when some should run from 1
to Nr_of_neurons.  Also there was no limit on the number of records
read from the file, so that the array Neurons[] could potentially
overflow.  It is although the neuron file *must* contain exactly
<Dimensionality> neurons.
---------------------------------------------------------------------*/
static void Read_Nfile(void) {
    int Nr_of_neurons;	/* size of Nfile */
    int i;		/* loop index */
    FILE *Nfile;
    long Total_number_of_synapses;

    Nfile=fopen(NFileName,"r");
    if(Nfile==NULL) {
        Neuron_file_present=0;
        printf("Neuron file could not be opened.\n");
        return;
    }

    /* Read file */
    Neuron_file_present=1;
    for(Nr_of_neurons=0; Nr_of_neurons<Dimensionality; Nr_of_neurons++)
        if(fread(&Neurons[Nr_of_neurons],sizeof(Neuron),1,Nfile)==0) break;
    printf("The neuron file contains %d neurons.\n",Nr_of_neurons);
    fclose(Nfile);
    Erase_line();

    /* Fill Number_of_synapses[] */
    /* Compute average number of synapses in neuron file. */
    Total_number_of_synapses=0;
    for(i=0; i<Nr_of_neurons; i++) {
        Number_of_synapses[i]=Synapse_number(&Neurons[i]);
        Total_number_of_synapses += Number_of_synapses[i];
    }
    /* Fill remainder of array with zeros, just in case */
    for(; i<Dimensionality; i++)
        Number_of_synapses[i]=0;

    /* Calculate average number of synapses */
    if(Nr_of_neurons==0)
        Avg_number_of_synapses=0;
    else
        Avg_number_of_synapses = (int)(Total_number_of_synapses/Nr_of_neurons);
    printf("Average number of synapses in neuron file: %d\n",
        Avg_number_of_synapses);
    Erase_line();
}   /* Read_Nfile */


/*---------------------------------------------------------------------
---------------------------------------------------------------------*/
void Read_neuron_file(void) {
    printf("\n");
    Scroll_block();
    Bottom_line();
    printf("The program is reading the neuron file.\n");
    Erase_line();
    printf("This is a file of type Neuron.\n");
    Erase_line();
    Read_Nfile();
    Restore_scrolling();
    printf("\n");
}   /* Read_neuron_file */
